/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_config_config_hpp_included
#define cool_config_config_hpp_included

/**
 * \file config.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::config_parser, cool::basic_config, cool::config
 *
 * In module \ref cool_config.
 */

#include <cool/cool.hpp>
#include <cool/exception/exception.hpp>
#include <cool/util/noncopyable.hpp>
#include <cool/util/less_str.hpp>
#include <cool/util/convert.hpp>

#if defined(istream_header)
#  include istream_header
#else
#  include iostream_header
#endif
#include map_header
#include algorithm_header


namespace cool {
/** \addtogroup cool_config */
/** \{ */


/**
 * \brief Abstract base class for configuration data parsers.
 *
 * All configuration parsers must be inherited from this class, which defines
 * interface for parsing configuration stream.
 *
 * Following virtual abstract methods must be overridden:
 *  - next(): read next key/value pair from stream and store these values into
 *    internal buffer. In case of parsing error cool::format_error should be
 *    thrown
 *  - key(): fetch last stored key name
 *  - value(): fetch last stored value corresponding to key
 */
class COOL_API config_parser
{
  public:

      /** \brief Construct new configuration parser object */
      config_parser () {}


      /** \brief Destroy configuration parser object */
      virtual ~config_parser () {}


      /**
       * \brief Read next key/value pair from configuration data stream \a is.
       *
       * Read values should be stored into internal buffer. If there is
       * parsing error, cool::format_error should be thrown.
       */
      virtual bool next (STD_ istream &is) = 0;


      /**
       * \brief Fetch last extracted name.
       *
       * If \a len is specified, store length of key into memory area pointed
       * by \a len.
       */
      virtual const char *key (int *len = 0) const = 0;


      /**
       * \brief Query last extracted value corresponding to key().
       *
       * If \a len is specified, store length of key into memory area pointed
       * by \a len.
       */
      virtual const char *value (int *len = 0) const = 0;
};


/**
 * \brief Configuration data storage
 *
 * This class is defined as template class i.e. programmer can specify
 * different containers for storing real key/value pairs. This container
 * should be associative array. This means that is should support storing and
 * retrieving elements based on key. Both key_type and value_type must be
 * \c 'const char *', corresponding to configuration variable name and value.
 *
 * \see cool::config
 */
template <class container>
class basic_config: public container, private noncopyable
{
  public:

      /**
       * \brief Parse configuration stream \a is using \a parser
       *
       * Stream is read using cool::config_parser::next() until it returns
       * \c false. All read key/value pairs are stored into \a container and
       * can be fetched or iterated later using standard container methods.
       */
      basic_config (config_parser &parser, STD_ istream &is);


      /** \brief Release internal resources */
      virtual ~basic_config () {
          STD_ for_each(container::begin(), container::end(), free_content());
      }


      /** \brief Return \c true, if \a key exists in configuration data */
      bool exists (const char *key) const {
          return container::find(key) != container::end();
      }


      /**
       * \brief Return pointer to value corresponding to \a key.
       *
       * Returned pointer points to internal buffer i.e. caller shouldn't
       * modify it's content. If \a key doesn't exist, \a def is returned.
       */
      const char *get (const char *key, const char *def = 0) const {
          typename container::const_iterator x = container::find(key);
          return (x != container::end()) ? x->second : def;
      }


      /** \brief Convenience function to get typed variable */
      template <typename T>
      T get (const char *key, T def) const {
          const char *v = get(key, (const char *)0);
          if (v == 0)
              return def;
          return as<T>(v);
      }


  private:

      struct free_content {
          void operator() (typename container::value_type &key_value) {
              free((void *)key_value.first);
          }
      };
};


/**
 * \brief Convenience type for cool::basic_config, specifying container as
 * \c "map<const char *, const char *, cool::less_strcmp>".
 *
 * \see cool::less_strcmp, cool::less_strcasecmp
 */
typedef basic_config<STD_ map<const char *, const char *, cool::less_strcmp> >
    config;


/** \} */
}   // namespace cool


#endif  /* cool_config_config_hpp_included */
