/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_config_ini_parser_hpp_included
#define cool_config_ini_parser_hpp_included

/**
 * \file ini_parser.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::ini_parser
 *
 * In module \ref cool_config.
 */

#include <cool/cool.hpp>
#include <cool/config/config.hpp>
#include <cool/util/noncopyable.hpp>


namespace cool {
/** \addtogroup cool_config */
/** \{ */


/**
 * @brief INI-file style configuration data parser.
 *
 * INI-file format is as series of simple assignments. For example:
 * @code
 * #
 * # Lines starting with '#' are comments as you can see. Comments and empty
 * # lines are ignored by parser.
 * #
 *
 * # get("msg") returns 'hello, world'
 * msg = hello, world
 *
 * # get("files") returns 'file1.cpp file2.cpp'
 * files = \
 *    file1.cpp \
 *    file2.cpp
 *
 * # get("file") returns 'file.cpp'
 * file = file\
 *    .\
 *    cpp
 * @endcode
 *
 * Multiline values are handled depending how escape char is placed:
 *  - whitespaces after value and before newline (not included) are preserved
 *  - whitespaces before following line line are removed
 *
 * Escape characters can be used to insert non-printable characters into
 * configuration value (but not in key names). For example, if escape char is
 * '\\', then known escape characters are converted into their corresponding
 * binary value ('\\n' as newline, '\\t' as tabulation character, etc).
 * Unknown escape combinations are ignored and escape character itself isn't
 * stored into value i.e. 'a\\xb' is stored as 'axb'. To insert escape
 * character itself, it must be written as double escape char,
 * for example '\\\\'.
 */
class COOL_API ini_parser: public config_parser, private noncopyable
{
  public:

      /**
       * @brief Constrct and initialize new parser object
       *
       * Allocates internal buffer for storing current key and it's
       * corresponding value. Sets also special characters.
       *
       * @param a assignment character
       * @param c comment character
       * @param e escape character. It is also used as continuation character
       * i.e. if it is last character in line then next line is parsed as part
       * of current value.
       */
      ini_parser (char a = '=', char c = '#', char e = '\\')
          : assign(a), comment(c), escape(e), line(0), size(0) {
          do_alloc(128);
          reset();
      }


      /** @brief Destroy parser object and release internal resources */
      ~ini_parser () {
          free(line);
      }


      /**
       * @brief Read next key/value pair from stream @a is and store their
       * values into internal buffer.
       *
       * Stream @a is must be readable.
       *
       * @throw cool::format_error if line can't be parsed.
       */
      virtual bool next (STD_ istream &is);


      /**
       * @brief Return pointer to content of last read key.
       *
       * Return pointer points to internal buffer i.e. caller shouldn't modify
       * it's content. If @a len is specified, then length of key content is
       * stored to @a *len
       */
      virtual const char *key (int *len = 0) const {
          if (len)
              *len = key_size;
          return line;
      }


      /**
       * @brief Return pointer to content of last read value.
       *
       * Return pointer points to internal buffer i.e. caller shouldn't modify
       * it's content. If @a len is specified, then length of value content is
       * stored to @a *len
       */
      virtual const char *value (int *len = 0) const {
          if (len)
              *len = value_size;
          return value_ptr;
      }


  protected:

      /**
       * @brief Additional hook for parsing: called in method next() after
       * line is read from stream but before parse_line().
       *
       * If good() returns true, then parser will continue, otherwise parsing
       * is cancelled just like stream is ended. Inherited class can override
       * this method to modify read @a data or to cancel reading early.
       *
       * @a data points to line just read.
       */
      virtual bool good (char *data) {
          return true;
      }


  private:

      void reset () {
          value_ptr = 0;
          key_size = value_size = 0;
      }

      void do_alloc (int new_size);
      void skip (STD_ istream &is);
      void parse_line ();

      char assign;
      char comment;
      char escape;

      char *line;
      int size;
      int key_size;
      char *value_ptr;
      int value_size;
};


/** \} */
}   // namespace cool


#endif  /* cool_config_ini_parser_hpp_included */
