/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_exception_exception_hpp_included
#define cool_exception_exception_hpp_included

/**
 * \file exception.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es):
 * cool::exception,
 * cool::format_error,
 * cool::interrupted,
 * cool::invalid_value_error,
 * cool::limit_error,
 * cool::not_found_error,
 * cool::not_supported_error,
 * cool::permission_error,
 * cool::resource_error,
 * cool::stopped,
 * cool::timeout_error
 *
 * In module \ref cool_exception.
 */

#include <cool/cool.hpp>
#include exception_header
#if defined(ostream_header)
#  include ostream_header
#else
#  include iostream_header
#endif


namespace cool {
/**
 * \addtogroup cool_exception
 * \{
 */


/** \brief Throw an exception of type \a ex */
template <class ex>
inline void
raise (ex e, const char *locus = 0)
{
  e.locus = locus;
  throw e;
}


/** \brief Throw an exception of type \a ex */
template <class ex>
inline void
raise (const char *locus = 0)
{
  raise(ex(), locus);
}


/** \brief Base exception class for all COOL library exceptions */
class COOL_API exception: public STD_ exception
{
  public:

      /**
       * \brief Source location where exception was thrown.
       *
       * This field is not set in constructor. It can be changed during
       * multi-stage exception simply by changing it's value. Actually it can
       * be any value.
       *
       * \see COOL_LOCUS
       */
      const char *locus;


      /** \brief Default constructor. Initializes locus */
      exception (): locus(0) {}


      /**
       * \brief Return pointer to a NUL-terminated string that has name of
       * current exception type.
       */
      virtual const char *what () const throw()
      { return "exception"; }
};


/**
 * \brief Macro for creating new exception classes.
 *
 * \arg name Name for new exception class. Method what() will return this
 * value
 * \arg parem Parent class for created exception class
 */
#define EXCEPTION_CLASS(name,parent) \
    class name: public parent { \
      public: \
          virtual const char *what () const throw() \
          { return #name; } \
    }


/**
 * \class format_error
 * \brief Thrown whenever data format error happens.
 */
EXCEPTION_CLASS(format_error, exception);

/**
 * \class interrupted
 * \brief Thrown whenever ongoing operation is interrupted somehow
 */
EXCEPTION_CLASS(interrupted, exception);

/**
 * \class invalid_value_error
 * \brief Thrown whenever invalid value is used
 */
EXCEPTION_CLASS(invalid_value_error, exception);

/**
 * \class limit_error
 * \brief Thrown whenever specified value falls outside of enabled range.
 */
EXCEPTION_CLASS(limit_error, exception);

/**
 * \class not_found_error
 * \brief Thrown whenever specified item is not found.
 */
EXCEPTION_CLASS(not_found_error, exception);

/**
 * \class not_supported
 * \brief Thrown whenever some operation is not available on current platform.
 *
 * Whenever any method or function throws this kind of exception, it is
 * explicitly documented.
 */
EXCEPTION_CLASS(not_supported_error, exception);

/**
 * \class permission_error
 * \brief Thrown whenever application has no enough rights to finish current
 * operation.
 */
EXCEPTION_CLASS(permission_error, exception);

/**
 * \class resource_error
 * \brief Thrown whenever resource allocation fails (memory allocation, file
 * opening, etc).
 */
EXCEPTION_CLASS(resource_error, exception);

/**
 * \class stopped
 * \brief Thrown whenever ongoing operation is suspended somehow.
 */
EXCEPTION_CLASS(stopped, exception);

/**
 * \class timeout_error
 * \brief Thrown whenever current operation couldn't be finished during
 * specified time-frame.
 */
EXCEPTION_CLASS(timeout_error, exception);


/**
 * \brief Store error reason \a e.what() with locus (<i>file</i>:<i>line</i>)
 * string into stream \a os.
 */
inline STD_ ostream &
operator<< (STD_ ostream &os, const exception &e)
{
  os << e.what();
  if (e.locus)
      os << " (" << e.locus << ')';
  return os;
}


/** \} */
}   // namespace cool


#endif  /* cool_exception_exception_hpp_included */
