/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_os_error_hpp_included
#define cool_os_error_hpp_included

/**
 * \file os_error.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::os_error
 *
 * In module \ref cool_exception.
 */

#include <cool/cool.hpp>
#include <cool/exception/exception.hpp>
#include "bits/os_error.hpp"


namespace cool {
/**
 * \addtogroup cool_exception
 * \{
 */


/** \brief Simple exception class wrapper around OS system error code */
class COOL_API os_error: public exception
{
  public:

      /** \brief OS low-level error type */
      typedef bits::exception::error_type error_type;


      /**
       * \brief Create exception object that holds OS's error code returned by
       * os_error::last().
       *
       * System error code depends on platform where COOL is compiled. On
       * Posix platforms, it is \a errno.
       */
      os_error (const char *func = 0): e_(last()), func_(func) { }


      /**
       * \brief Create exception object that holds OS's error code \a e.
       *
       * \copydoc os_error(const char *)
       */
      os_error (error_type e, const char *func = 0): e_(e), func_(func) { }


      /** \brief Return a C-style character string "os_error" */
      virtual const char *what () const throw() {
          return "os_error";
      }


      /** \brief Return function name that caused error */
      const char *func () const throw() {
          return func_;
      }


      /**
       * \brief Return reference to internal error variable i.e. make it to
       * behave as ordinary \c errno variable.
       */
      operator error_type& () throw() {
          return e_;
      }


      /** \brief Return error message corresponding to error code */
      const char *detail () const throw() {
          return string(e_);
      }


      error_type code () const throw() {
          return e_;
      }


      /** \brief Return current OS's error code i.e. result of last operation */
      static error_type last () {
          return bits::exception::last();
      }


      /** \brief Set OS's error code i.e. result of last operation */
      static void last (error_type e) {
          bits::exception::last(e);
      }


      /** \brief Return pointer to error message corresponding to code \a e */
      static const char *string (error_type e) {
          return bits::exception::string(e);
      }


  private:

      error_type e_;
      const char *func_;
};


/** \brief Store error reason \a e into stream \a os */
inline STD_ ostream &
operator<< (STD_ ostream &os, const os_error &e)
{
  os << (e.func() ? e.func() : e.what()) << ' ';
  if (e.locus) os << '(' << e.locus << ") ";
  os << e.detail();

  return os;
}


/** \} */
}   // namespace cool


#endif  /* cool_os_error_hpp_included */
