/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_log_log_hpp_included
#define cool_log_log_hpp_included

/**
 * \file log.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::log
 *
 * In module \ref cool_log.
 */

#include <cool/cool.hpp>
#include <cool/time/time.hpp>
#include <cstdarg>
#include list_header
#include sstream_header
#include algorithm_header
#include iostream_header


namespace cool {
/** \addtogroup cool_log */
/** \{ */


/**
 * \brief
 */
class COOL_API log
{
  public:

      enum level_type {
          level_emerg
          , level_alert
          , level_crit
          , level_err
          , level_warn
          , level_notice
          , level_info
          , level_debug
          , level_unset
      };


      static const char *level_string (level_type level) {
          static const char *strings[] = {
              "EMERG", "ALERT", "CRIT", "ERR", "WARN", "NOTICE", "INFO", "DEBUG"
          };
          assert(level >= level_emerg);
          assert(level < level_unset);
          return strings[level];
      }


      struct event_type
      {
          cool::time::time time;
          level_type level;
          const char *message;

          event_type (level_type l, const char *m): level(l), message(m) {
              assert(level != level_unset);
              assert(m != 0);
          }
      };


  private:

      class basic_writer
      {
        public:

            virtual ~basic_writer () {}

            virtual bool reopen () { return true; }
            virtual void close () { }
            virtual void operator() (const event_type &ev) = 0;
      };


  public:

      log () {}


      log &add (basic_writer *w) {
          assert(w != 0);
          writers_.push_back(w);
          return *this;
      }


      log &remove (basic_writer *w) {
          assert(w != 0);
          writers_.remove(w);
          return *this;
      }


      log &reopen () {
          STD_ for_each(writers_.begin(), writers_.end(), do_reopen());
          return *this;
      }


      log &close () {
          STD_ for_each(writers_.begin(), writers_.end(), do_close());
          return *this;
      }


      log &write (const event_type &ev) {
          STD_ for_each(writers_.begin(), writers_.end(), do_write(ev));
          return *this;
      }


      log &writeva (level_type level, const char *fmt, va_list va) {
          assert(fmt != 0);
          char line[1024];
          vsnprintf(line, sizeof(line), fmt, va);
          return write(event_type(level, line));
      }


      log &write (level_type level, const char *fmt, ...);
      log &emerg (const char *fmt, ...);
      log &alert (const char *fmt, ...);
      log &crit (const char *fmt, ...);
      log &err (const char *fmt, ...);
      log &warn (const char *fmt, ...);
      log &notice (const char *fmt, ...);
      log &info (const char *fmt, ...);
      log &debug (const char *fmt, ...);


  private:

      typedef STD_ list<basic_writer *> writer_list;

      struct do_write {
          const event_type &ev_;

          do_write (const event_type &ev): ev_(ev) {}

          void operator() (writer_list::value_type &log_writer) {
              log_writer->operator()(ev_);
          }
      };


      struct do_reopen {
          void operator() (writer_list::value_type &log_writer) {
              log_writer->reopen();
          }
      };


      struct do_close {
          void operator() (writer_list::value_type &log_writer) {
              log_writer->close();
          }
      };


      writer_list writers_;
};


/** \} */
}   // namespace cool


#endif  /* cool_log_log_hpp_included */
