/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_bits_poller_epoll_included
#define cool_net_bits_poller_epoll_included


#if defined(HAVE_SYS_EPOLL_H)
#  include <sys/epoll.h>
#endif
#if defined(HAVE_UNISTD_H)
#  include <unistd.h>
#endif


namespace cool {
namespace bits {
namespace net {


class poller_type
{
  public:

      // initial size of events_
      enum { ini_size = 16 };


      // number of maximum opened file/socket descriptors specified as hint
      // for epoll_create()
      static int max_fd () {
          return sysconf(_SC_OPEN_MAX);
      }


      // initialize poller
      poller_type (): epollfd_(-1), events_(0), allocated_(0), size_(0) {
          if ((epollfd_ = ::epoll_create(max_fd())) == -1)
              raise<resource_error>(COOL_LOCUS);
          if ((events_ = (epoll_event *)malloc(ini_size*sizeof(*events_))) == 0)
              raise<resource_error>(COOL_LOCUS);
          allocated_ = ini_size;
      }


      // free used resources
      ~poller_type () {
          ::close(epollfd_);
          if (allocated_ > 0) {
              assert(events_ != 0);
              free(events_);
          }
      }


      // return number of registered polled sockets
      int size () const {
          return size_;
      }


      // register new socket to be polled
      void insert (socket *s, socket::event_type ev);


      // update socket events to be watched
      void modify (socket *s, socket::event_type ev) {
          mod_fd(s, ev, false);
      }


      // unregister socket
      void erase (socket *s) {
          del_fd(s);
          size_--;
      }


      // wait for any event on registered sockets
      int wait (const cool::time *timeout = 0);


      // return next socket that has pending event(s)
      socket *pending (socket::event_type &ev) {
          socket *sock = 0;
          ev = socket::no_events;
          if (last_ < 1 || pending_ == last_)
              // no active sockets
              return 0;

          // next active socket
          if (events_[pending_].events & EPOLLIN) ev |= socket::in;
          if (events_[pending_].events & EPOLLOUT) ev |= socket::out;
          if (events_[pending_].events & EPOLLPRI) ev |= socket::except;
          sock = (socket *)events_[pending_++].data.ptr;

          return sock;
      }


  private:

      // insert or modify socket in internal poller
      void mod_fd (socket *s, socket::event_type ev, bool add);


      // remove socket from internal poller
      void del_fd (socket *s) {
          epoll_event pev;
          if (::epoll_ctl(epollfd_, EPOLL_CTL_DEL, *s, &pev) == -1) {
              os_error::error_type e = os_error::last();
              if (e == ENOENT)
                  return;

              // EBADF probably is not an error as epoll_wait() automatically
              // erases closed socket
              if (e != EBADF)
                  throw_last("epoll_ctl", COOL_LOCUS);
          }
      }


      // epoll file descriptor
      int epollfd_;

      // pointer to array of returned sockets and corresponding events
      epoll_event *events_;

      // number of allocated items
      int allocated_;

      // number of inserted items
      int size_;

      // result of last wait
      int last_;

      // last iterated socket
      int pending_;
};


}  // namespace net
}  // namespace bits
}  // namespace cool


#endif  // cool_net_bits_poller_epoll_included
