/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_bits_poller_poll_included
#define cool_net_bits_poller_poll_included


#if defined(HAVE_SYS_POLL_H)
#  include <sys/poll.h>
#endif
#include <map>


namespace cool {
namespace bits {
namespace net {


struct socket_data_type
{
  socket *socket_;
  int ndx_;

  socket_data_type ()
      : socket_(0), ndx_(-1) {}

  socket_data_type (socket *socket, int ndx)
      : socket_(socket), ndx_(ndx) {}

  socket_data_type (const socket_data_type &o)
      : socket_(o.socket_), ndx_(o.ndx_) {}
};


typedef STD_ map<socket::socket_type, socket_data_type> socket_map;


class poller_type
{
  public:

      enum { ini_size = 16 };

      // initialize poller
      poller_type (): fds_(0), begin_(0), end_(0), size_(0) {
          if ((fds_ = (pollfd *)malloc(ini_size*sizeof(*fds_))) == 0)
              raise<resource_error>(COOL_LOCUS);
          size_ = ini_size;
      }


      ~poller_type () {
          if (size_ > 0) {
              assert(fds_ != 0);
              free(fds_);
          }
      }


      // return number of registered polled sockets
      int size () const {
          return sockets_.size();
      }


      // register new socket to be polled
      void insert (socket *s, socket::event_type ev);


      // update socket events to be watched
      void modify (socket *s, socket::event_type ev) {
          socket_map::iterator i = sockets_.find(*s);
          assert(i != sockets_.end());
          fds_[i->second.ndx_].events = cool_ev_to_sys_ev(ev);
      }


      // unregister socket
      void erase (socket *s);


      // wait for any event on registered sockets
      int wait (const cool::time *timeout = 0);


      // return next socket that has pending event(s)
      socket *pending (socket::event_type &ev) {
          socket *socket = 0;
          ev = socket::no_events;
          if (last_ < 1 || pending_ == sockets_.end())
              // no active sockets
              return 0;

          // search for active socket
          while (ev == socket::no_events && pending_ != sockets_.end()) {
              short events = fds_[pending_->second.ndx_].revents;

              if (events & POLLIN) ev |= socket::in;
              if (events & POLLOUT) ev |= socket::out;
              if (events & POLLPRI) ev |= socket::except;

              if (ev != socket::no_events) {
                  // found
                  socket = pending_->second.socket_;
                  --last_;
              }

              // next
              ++pending_;
          }

          return socket;
      }


  private:

      short cool_ev_to_sys_ev (socket::event_type ev) {
          short r = 0;
          if (ev & socket::in) r |= POLLIN;
          if (ev & socket::out) r |= POLLOUT;
          if (ev & socket::except) r |= POLLPRI;
          return r;
      }

      // map of sockets and corresponding index to array of fds_
      socket_map sockets_;

      // pointer to array of pollfd
      pollfd *fds_;

      // index of first valid item in fds_
      int begin_;

      // index to next of last valid item in fds_
      int end_;

      // number of allocated items in fds_
      int size_;

      // result of last poll
      int last_;
      socket_map::const_iterator pending_;
};


}  // namespace net
}  // namespace bits
}  // namespace cool


#endif  // cool_net_bits_poller_poll_included
