/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_bits_poller_select_included
#define cool_net_bits_poller_select_included


#if defined(HAVE_SYS_SELECT_H)
#  include <sys/select.h>
#else
#  if defined(HAVE_SYS_TIME_H)
#    include <sys/time.h>
#  endif
#  if defined(HAVE_UNISTD_H)
#    include <sys/unistd.h>
#  endif
#endif
#include <map>


namespace cool {
namespace bits {
namespace net {


struct socket_data_type
{
  socket *socket_;
  socket::event_type event_;

  socket_data_type ()
      : socket_(0), event_(socket::no_events) {}

  socket_data_type (socket *socket, socket::event_type event)
      : socket_(socket), event_(event) {}

  socket_data_type (const socket_data_type &o)
      : socket_(o.socket_), event_(o.event_) {}
};


typedef STD_ map<socket::socket_type, socket_data_type> socket_map;


class poller_type
{
  public:

      // initialize poller
      poller_type (): last_(0), pending_(sockets_.end()) {
          FD_ZERO(&rd);
          FD_ZERO(&wr);
          FD_ZERO(&ex);
      }


      // return number of registered polled sockets
      int size () const {
          return sockets_.size();
      }


      // register new socket to be polled
      void insert (socket *s, socket::event_type ev);


      // update socket events to be watched
      void modify (socket *s, socket::event_type ev);


      // unregister socket
      void erase (socket *s);


      // wait for any event on registered sockets
      int wait (const cool::time *timeout = 0);


      // return next socket that has pending event(s)
      socket *pending (socket::event_type &ev) {
          socket *s = 0;
          ev = socket::no_events;
          if (last_ < 1 || pending_ == sockets_.end())
              // no active sockets
              return 0;

          // search for active socket
          while (ev == socket::no_events && pending_ != sockets_.end()) {
              if (FD_ISSET(pending_->first, &rrd)) {
                  ev |= socket::in;
                  --last_;
              }
              if (FD_ISSET(pending_->first, &rwr)) {
                  ev |= socket::out;
                  --last_;
              }
              if (FD_ISSET(pending_->first, &rex)) {
                  ev |= socket::except;
                  --last_;
              }

              if (ev != socket::no_events)
                  // found
                  s = pending_->second.socket_;

              // next
              ++pending_;
          }

          return s;
      }


  private:

      // return maximum socket descriptor value
      socket_map::key_type max_handle () const {
          socket_map::const_reverse_iterator i = sockets_.rbegin();
          return i->first;
      }


      // update fd_set
      void set (socket::event_type f,
              socket::event_type old_ev,
              socket::event_type new_ev,
              socket *s, fd_set *fds)
      {
          socket::event_type ev = new_ev & f;
          if ((old_ev & f) != ev) {
              if (ev == socket::no_events)
                  FD_CLR(*s, fds);
              else
                  FD_SET(*s, fds);
          }
      }


      // map of sockets and corresponding events to be wait
      socket_map sockets_;

      // result of last wait
      int last_;
      socket_map::const_iterator pending_;

      // OS-level socket bitmap
      fd_set rd, wr, ex;
      fd_set rrd, rwr, rex;
};


}  // namespace net
}  // namespace bits
}  // namespace cool


#endif  // cool_net_bits_poller_select_included
