/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_bits_socket_hpp_included
#define cool_net_bits_socket_hpp_included


#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif
#ifdef HAVE_FCNTL_H
#  include <fcntl.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#  include <sys/ioctl.h>
#endif
#ifdef HAVE_SYS_FILIO_H
#  include <sys/filio.h>
#endif
#ifdef HAVE_SYS_SOCKIO_H
#  include <sys/sockio.h>
#endif


namespace cool {
namespace bits {
namespace net {


void throw_last (const char *func_name, const char *locus);
void throw_last_netio (os_error::error_type e,
        const char *func_name, const char *locus);
void bind (socket_base::socket_type s,
        const sockaddr *a, socket_base::socklen_type l); // ugly :(
socket_base::event_type wait (socket_base::socket_type s,
        socket_base::event_type ev, const cool::time *timeout);


inline socket_base::socket_type
open (socket_base::family_type family, socket_base::style_type style)
{
  socket_base::socket_type s_ = ::socket(family, style, 0);
  if (s_ == -1)
      throw_last("socket", COOL_LOCUS);

  return s_;
}


inline void
close (socket_base::socket_type s)
{
  if (::close(s) == -1)
      ; // XXX What can we do?
}


inline socket_base::socket_type
clone (socket_base::socket_type s)
{
  socket_base::socket_type s_ = ::dup(s);
  if (s_ == -1)
      throw_last("dup", COOL_LOCUS);

  return s_;
}


inline int
getopt (const socket_base::socket_type &s, int l, int op, void *buf, int len)
{
  if (::getsockopt(s, l, op, buf, (socket_base::socklen_type *)&len) == -1)
      throw_last("getsockopt", COOL_LOCUS);
  return len;
}


inline void
setopt (const socket_base::socket_type &s, int l, int op, void *buf, int len)
{
  if (::setsockopt(s, l, op, buf, len) == -1)
      throw_last("setsockopt", COOL_LOCUS);
}


#undef NO_DELAY_FLAG
#if defined(O_NONBLOCK)
#  define NO_DELAY_FLAG O_NONBLOCK
#elif defined(O_NDELAY)
#  define NO_DELAY_FLAG O_NDELAY
#elif defined(O_FNDELAY)
#  define NO_DELAY_FLAG O_FNDELAY
#else
// XXX What can we do?
#  define NO_DELAY_FLAG 0
#endif


inline bool
getblock (const socket_base::socket_type &s)
{
  int f = ::fcntl(s, F_GETFL, 0);
  if (f == -1)
      throw_last("fcntl", COOL_LOCUS);
  return (f & NO_DELAY_FLAG) == 0;
}


inline void
setblock (const socket_base::socket_type &s, bool on)
{
  int old = ::fcntl(s, F_GETFL, 0);
  int flags = on ? old & ~NO_DELAY_FLAG : old | NO_DELAY_FLAG;
  if (old != flags && ::fcntl(s, F_SETFL, flags) == -1)
      throw_last("fcntl", COOL_LOCUS);
}


inline bool
atmark (const socket_base::socket_type &s)
{
  int arg;
  if (::ioctl(s, SIOCATMARK, &arg) == -1)
      throw_last("ioctl", COOL_LOCUS);
  return arg;
}


inline int
getpgrp (const socket_base::socket_type &s)
{
  int arg = 0;
  if (::ioctl(s, SIOCGPGRP, &arg) == -1)
      throw_last("ioctl", COOL_LOCUS);
  return arg;
}


inline void
setpgrp (const socket_base::socket_type &s, int pgrp)
{
  if (::ioctl(s, SIOCSPGRP, &pgrp) == -1)
      throw_last("ioctl", COOL_LOCUS);
}


inline long
in_avail (const socket_base::socket_type &s)
{
  long arg = 0;
  if (::ioctl(s, FIONREAD, &arg) == -1)
      throw_last("ioctl", COOL_LOCUS);
  return arg;
}


inline long
async (const socket_base::socket_type &s, bool on)
{
  int arg = on;
  if (::ioctl(s, FIOASYNC, &arg) == -1)
      throw_last("ioctl", COOL_LOCUS);
  return arg;
}


}   // namespace net
}   // namespace bits
}   // namespace cool


#endif  /* cool_net_bits_socket_hpp_included */
