/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_client_socket_hpp_included
#define cool_net_client_socket_hpp_included

/**
 * \file client_socket.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::client_socket
 *
 * In module \ref cool_net.
 */

#include <cool/cool.hpp>
#include <cool/net/socket.hpp>
#include "bits/client_socket.hpp"


namespace cool {
/** \addtogroup cool_net */
/** \{ */


/**
 * \brief Base class for client side sockets.
 *
 * This base class is not meant to be used directly but provides only
 * properties and methods common to client sockets. Applications should use
 * cool::inet::socket or cool::unix::socket instead.
 */
class COOL_API client_socket: public socket
{
  public:

      /** \brief Create new client side socket with invalid descriptor */
      client_socket (): socket() { }


      /**
       * \brief Create object from already existing socket descriptor
       *
       * \note Specified socket descriptor \a s is not duplicated i.e.
       * ownership of descriptor goes to new object.
       */
      client_socket (socket_type s): socket(s) { }


      /** \brief Create new socket descriptor */
      client_socket (family_type family, style_type style = def::style)
          : socket(family, style) { }


      /** \brief Create copy of \a o by duplicating it's socket descriptor */
      client_socket (const client_socket &o): socket(o) { }


      /**
       * \brief Take ownership of socket \a s i.e. caller shouldn't use it any
       * more.
       */
      client_socket &operator = (socket_type s) {
          socket::operator=(s);
          return *this;
      }


      /**
       * \brief Connect socket to address \a a.
       *
       * Returns true if establishing connection succeeded.
       *
       * - Non-blocking sockets: if false is returned, it just means that you
       *   can call connect() again after little delay and it'll probably
       *   succeeds. If still no luck after delay, application can retry or
       *   check for error reason code by calling socket::pending_error().
       * - Blocking sockets: if connection can't be established, exception is
       *   thrown. If not, then true is always returned.
       */
      bool connect (const address &a) {
          return bits::net::connect(s_, a);
      }


      /** \brief Return true if socket is connected */
      bool is_connected () const {
          return bits::net::is_connected(s_);
      }


      /**
       * \brief Close part of full-duplex communication channel depending on
       * \a how
       */
      client_socket &shutdown (shutmode_type how = rw) {
          bits::net::shutdown(s_, how);
          return *this;
      }


      /**
       * \brief Read from socket.
       *
       * Attempts to read up to \a len byte(s) from socket into memory area
       * \a buf. If \a len is zero, method returns zero and has no other
       * results
       */
      int recv (char *buf, int len, io_flags_type f = def::io_flags) {
          return bits::net::recv(s_, buf, len, f);
      }


      /**
       * \brief Read from socket.
       *
       * \copydoc client_socket::recv()
       */
      int recvfrom (address &a, char *buf, int len,
              io_flags_type f = def::io_flags) {
          return bits::net::recvfrom(s_, a, buf, len, f);
      }


      /**
       * \brief Send to socket.
       *
       * Write up to \a len byte(s) to socket from data area \a buf. On
       * success number of byte(s) written is returned (zero indicates nothing
       * has written).
       */
      int send (char *buf, int len, io_flags_type f = def::io_flags) {
          return bits::net::send(s_, buf, len, f);
      }


      /**
       * \brief Send to socket.
       *
       * \copydoc client_socket::send()
       */
      int sendto (const address &a, char *buf, int len,
              io_flags_type f = def::io_flags) {
          return bits::net::sendto(s_, a, buf, len, f);
      }
};


/** \} */
}   // namespace cool


#endif  /* cool_net_client_socket_hpp_included */
