/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_inet_address_hpp_included
#define cool_net_inet_address_hpp_included

/**
 * \file inet_address.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::inet_address
 *
 * \todo Add support for IPv6 addresses
 *
 * In module \ref cool_net.
 */

#include <cool/cool.hpp>
#include <cool/net/exception.hpp>
#include "bits/inet_address.hpp"

#if defined(ostream_header)
#  include ostream_header
#else
#  include iostream_header
#endif


namespace cool {
/** \addtogroup cool_net */
/** \{ */


/**
 * \brief IPv4 host address
 *
 * \todo In configure.ac check for reentrant functions (with _r suffix) and
 * use these if available.
 */
class COOL_API inet_address: public in_addr
{
  public:

      /** \brief Low level IP address representation */
      typedef bits::net::ip_type ip_type;


      /** \brief Well-known IP addresses */
      enum {
          none = bits::net::none           /**< Invalid address */
          , any = bits::net::any           /**< Any address */
          , loopback = bits::net::loopback /**< Loopback address */
      };


      /**
       * \brief Maximum neccessary buffer sizes for getting textual
       * representations
       */
      enum {
          host_length = 128
          , ip_length = sizeof("255.255.255.255")
      };


      /** \brief Construct an address object with numeric IP address \a ip_ */
      inet_address (ip_type ip_ = none) {
          ip(ip_);
      }


      /**
       * \brief Construct an address with IP corresponding to \a host.
       *
       * \see host(char*)
       */
      inet_address (const char *host_) {
          host(host_);
      }


      /** \brief Destruct the address object */
      virtual ~inet_address () {}


      /** \brief Set IP address and return reference to this object */
      inet_address &ip (ip_type ip_) {
          s_addr = htonl(ip_);
          return *this;
      }


      /** \brief Return IP address */
      ip_type ip () const {
          return ntohl(s_addr);
      }


      /**
       * \brief Set IP address to address of networking interface (lo, eth0
       * etc).
       */
      inet_address &iface (const char *iface_) {
          if ((s_addr = bits::net::lookup_iface(iface_)) == none)
              raise<iface_not_found>(COOL_LOCUS);
          return *this;
      }


      /**
       * \brief Set IP address by hostname (can be textual representation of
       * IP address).
       *
       * If \a host_ is 0 or emtpy or first character equal to '*', then IP is
       * set to inet_address::any
       *
       * \a host_ can also be network interface name (lo, eth0 etc). If
       * resolving fails, this method tries to resolve interface name to IP
       * address.
       *
       * On failure, cool::host_not_found is thrown.
       */
      inet_address &host (const char *host_);


      /**
       * \brief Return hostname.
       *
       * If IP is inet_address::any then name of local machine is returned.
       *
       * Name is copied into \a buf not exceeding \a bufsz byte(s), truncating
       * name if neccessary. \a buf must have room at least for
       * inet_address::host_length byte(s).
       *
       * Returne pointer points to gethostbyaddr() internal buffer that will be
       * overwritten by subsequent calls (even when other objects or functions
       * call it).
       */
      const char *host () const {
          return bits::net::hoststr(*this);
      }


      /**
       * \brief Return textual representation of IP address.
       *
       * Returne pointer points to inet_ntoa() internal buffer that will be
       * overwritten by subsequent calls (even when other objects or functions
       * call it).
       */
      const char *c_str () const {
          return bits::net::ipstr(*this);
      }


      /**
       * \brief Return local machine's host name
       *
       * Return pointer points to internal buffer that will be overwritten by
       * subsequent calls.
       */
      static const char *hostname () {
          return bits::net::hostname();
      }
};


/** \brief Return true, if \a a1 and \a a2 are equal */
inline bool
operator == (const inet_address &a1, const inet_address &a2)
{
  return a1.s_addr == a2.s_addr;
}


/** \brief Return true, if \a a1 and \a a2 are not equal */
inline bool
operator != (const inet_address &a1, const inet_address &a2)
{
  return !(operator==(a1, a2));
}


/**
 * \brief Format and store IP-part of \a a into stream os as quad-dotted
 * numbers
 */
inline STD_ ostream &
operator << (STD_ ostream &os, const inet_address &a)
{
  os << a.c_str();
  return os;
}


/** \} */
}   // namespace cool


#endif  /* cool_net_inet_address_hpp_included */
