/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_inet_socket_hpp_included
#define cool_net_inet_socket_hpp_included

/**
 * \file inet_socket.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::inet::address, cool::inet::socket,
 * cool::inet::listener
 *
 * In module \ref cool_net.
 */

#include <cool/cool.hpp>
#include <cool/net/inet_address.hpp>
#include <cool/net/client_socket.hpp>
#include <cool/net/server_socket.hpp>
#include <cool/util/convert.hpp>
#include "bits/inet_socket.hpp"

#if defined(ostream_header)
#  include ostream_header
#else
#  include iostream_header
#endif


namespace cool {
/** \addtogroup cool_net */
/** \{ */


/**
 * \brief Wrapper class containing classes inet::address, inet::socket,
 * inet::listener
 */
class COOL_API inet
{
  public:

      /** \brief Family value */
      static const cool::socket::family_type family;


      /**
       * \brief socket_base::inet family socket address
       *
       * \todo Check and use reentrant functions \c gethostbyname_r() etc.
       */
      class address: public client_socket::address
      {
        public:

            /** \brief Port number type */
            typedef bits::net::port_type port_type;


            /**
             * \brief Construct an address object with interface
             * inet_address::any and port 0.
             */
            address () {
                sa_.sin_family = socket::inet;
                sa_.sin_addr.s_addr = htons(inet_address::any);
                sa_.sin_port = htons(0);
            }


            /**
             * \brief Construct an address object with inet_address and port
             * number.
             */
            address (const inet_address &a, port_type p) {
                sa_.sin_family = socket::inet;
                host(a);
                port(p);
            }


            /**
             * \brief Construct an address from textual host:port pair
             *
             * \see operator=(char*)
             */
            address (const char *as_string) {
                sa_.sin_family = socket::inet;
                operator=(as_string);
            }


            /**
             * \brief Parse host and port pair, specified as string.
             *
             * \a as_string must have pattern [host][:port], where
             *  - \e host name. Same rules apply as to
             *    inet_address::host(char*). If not specified,
             *    inet_address::any is used
             *  - \e port number or service name. If not specified, 0 is used.
             *
             * If \a as_string is 0 or empty, then
             * inet_socket::address(inet_address::any, 0) is returned.
             */
            address &operator = (const char *as_string);


            /** \brief Return object as \c "void *" */
            operator void * () const {
                return addr_in();
            }


            /** \brief Return socket_base::inet specific address structure */
            sockaddr_in *addr_in () const {
                return (sockaddr_in *)&sa_;
            }


            /** \brief Return pointer to standard socket address data */
            sockaddr *addr () const {
                return reinterpret_cast<struct sockaddr *>(addr_in());
            }


            /** \brief Return size of sockaddr_in part of address object */
            socket::socklen_type size () const {
                return sizeof(sa_);
            }


            /** \brief Return address family */
            socket::family_type family () const {
                return socket::family_type(sa_.sin_family);
            }


            /** \brief Set new host */
            address &host (const inet_address &a) {
                sa_.sin_addr.s_addr = a.s_addr;
                return *this;
            }


            /** \brief Return host */
            inet_address host () const {
                inet_address a;
                a.s_addr = sa_.sin_addr.s_addr;
                return a;
            }


            /** \brief Return textual representation of address */
            const char *c_str () const {
                return host().c_str();
            }


            /** \brief Set new port number */
            address &port (port_type p) {
                sa_.sin_port = htons(p);
                return *this;
            }


            /**
             * \brief Return port number
             *
             * \see service()
             */
            port_type port () const {
                return ntohs(sa_.sin_port);
            }


            /**
             * \brief Return well-known port number for service \a servname
             * and protocol \a proto (defaults to \c "tcp").
             */
            static port_type service (const char *servname,
                    const char *proto = "tcp");

        protected:

            friend bool
            operator== (const inet::address &a1, const inet::address &a2);

            sockaddr_in sa_;
      };


      /**
       * \brief Simple socket_base::inet client socket's API wrapper.
       *
       * This class provides only client side functions for connecting,
       * send/receive. It also has methods for altering socket options. See
       * documentation for parent about these methods.
       */
      class socket: public client_socket
      {
        public:

            /** \brief Create new socket descriptor */
            socket (style_type style = def::style)
                : client_socket(inet, style) {}


            /**
             * \brief Create client object from already existing socket
             * descriptor
             *
             * \note Specified socket descriptor \a s is not duplicated i.e.
             * ownership of descriptor goes to new object.
             */
            socket (socket_type s): client_socket(s) {}


            /**
             * \brief Create copy of \a o by duplicating it's socket
             * descriptor
             */
            socket (const socket &o): client_socket(o) {}


            /**
             * \brief Take ownership of socket \a s i.e. caller shouldn't use
             * it anymore.
             */
            socket &operator = (socket_type s) {
                client_socket::operator=(s);
                return *this;
            }


            /** \brief Return local communication endpoint address */
            inet::address local () const;


            /** \brief Return peer's communication endpoint address */
            inet::address peer () const;


            /** \brief Return true if Nagle's algorithm is disabled */
            bool tcpnodelay ();


            /**
             * \brief Disable Nagle's algorithm if \a on == \c true and enable
             * otherwise
             */
            socket &tcpnodelay (bool on);
      };


      /**
       * \brief Listener (or acceptor) socket.
       *
       * A listener socket accepts incoming connection requests, returning
       * OS-level socket descriptor that represents communication channel.
       */
      class listener: public server_socket
      {
        public:

            /** \brief Create new server socket descriptor */
            listener (): server_socket(inet) { }


            /**
             * \brief Create server socket object from already existing socket
             * descriptor.
             *
             * \note Specified socket descriptor \a s is not duplicated i.e.
             * ownership of descriptor goes to new object.
             */
            listener (socket_type s): server_socket(s) {}


            /**
             * \brief Create copy of \a o by duplicating it's socket
             * descriptor
             */
            listener (const listener &o): server_socket(o) {}


            /**
             * \brief Create new server socket object that is bound to \a a
             * and has \a backlog queue length
             *
             * This constructor also enables reuse of address.
             */
            listener (const address &a, int backlog = somaxconn)
                : server_socket(inet) {
                reuseaddr(true);
                bind(a);
                listen(backlog);
            }


            /**
             * \brief Take ownership of socket \a s i.e. caller shouldn't use
             * it anymore.
             */
            listener &operator = (socket_type s) {
                server_socket::operator=(s);
                return *this;
            }
      };
};


/** \brief Return true, if \a a1 and \a a2 are equal */
inline bool
operator == (const inet::address &a1, const inet::address &a2)
{
  return a1.sa_.sin_family == a2.sa_.sin_family &&
      a1.sa_.sin_addr.s_addr == a2.sa_.sin_addr.s_addr &&
      a1.sa_.sin_port == a2.sa_.sin_port;
}


/** \brief Return true, if \a a1 and \a a2 are not equal */
inline bool
operator != (const inet::address &a1, const inet::address &a2)
{
  return !(operator==(a1, a2));
}


/**
 * \brief Store IP-address and port number pair to stream \a os separated with
 * colon.
 */
inline STD_ ostream &
operator << (STD_ ostream &os, const inet::address &a)
{
  os << a.host().c_str() << ':' << a.port();
  return os;
}


/** \brief Parse textual inet::address */
template<> inline inet::address
as<inet::address> (const char *as_string, char ** /*unused*/)
{
  return inet::address(as_string);
}


/** \} */
}   // namespace cool


#endif  /* cool_net_inet_socket_hpp_included */
