/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_local_socket_hpp_included
#define cool_net_local_socket_hpp_included

/**
 * \file local_socket.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::local::address, cool::local::socket,
 * cool::local::listener
 *
 * In module \ref cool_net.
 */

#include <cool/cool.hpp>
#include <cool/net/client_socket.hpp>
#include <cool/net/server_socket.hpp>
#include <cool/util/convert.hpp>
#include "bits/local_socket.hpp"

#if defined(ostream_header)
#  include ostream_header
#else
#  include iostream_header
#endif


namespace cool {
/** \addtogroup cool_net */
/** \{ */


/**
 * \brief Wrapper class containing classes local::address, local::socket,
 * local::listener
 */
class COOL_API local
{
  public:

      /** \brief Family value */
      static const cool::socket::family_type family;


      /** \brief socket_base::local family socket address */
      class address: public client_socket::address
      {
        public:

            /** \brief Construct an empty address */
            address () {
                sa_.sun_family = socket::local;
                sa_.sun_path[0] = '\0';
            }


            /** \brief Construct an address object from \a path */
            address (const char *as_string) {
                sa_.sun_family = socket::local;
                operator=(as_string);
            }


            /** \brief Construct an address object from other address. */
            address (const address &a) {
                sa_.sun_family = socket::local;
                operator=(a.sa_.sun_path);
            }


            /** \brief Set unix address path */
            address &operator = (const char *as_string) {
                assert(as_string != 0);
                strncpy(sa_.sun_path, as_string, sizeof(sa_.sun_path));
                sa_.sun_path[sizeof(sa_.sun_path) - 1] = '\0';
                return *this;
            }


            /** \brief Return object as \c "void *" */
            operator void * () const {
                return addr_un();
            }


            /** \brief Return socket_base::local specific address structure */
            sockaddr_un *addr_un () const {
                return (sockaddr_un *)&sa_;
            }


            /** \brief Return pointer to standard socket address data */
            sockaddr *addr () const {
                return reinterpret_cast<struct sockaddr *>(addr_un());
            }


            /** \brief Return size of sockaddr_in part of address object */
            socket::socklen_type size () const {
                return sizeof(sa_);
            }


            /** \brief Return address family */
            socket::family_type family () const {
                return socket::family_type(sa_.sun_family);
            }


            /** \brief Return textual representation of address */
            const char *c_str () const {
                return sa_.sun_path;
            }


        protected:

            friend bool
            operator== (const local::address &a1, const local::address &a2);

            sockaddr_un sa_;
      };


      /**
       * \brief Simple socket_base::local client socket's API wrapper.
       *
       * This class provides only client side methods for connecting,
       * send/receive. It also has methods for altering socket options. See
       * documentation for parent about these mehtods.
       */
      class socket: public client_socket
      {
        public:

            /** \brief Create new socket descriptor */
            socket (style_type style = def::style)
                : client_socket(local, style) {}


            /**
             * \brief Create socket object from already existing socket
             * descriptor
             *
             * \note Specified socket descriptor \a s is not duplicated i.e.
             * ownership of descriptor goes to new object.
             */
            socket (socket_type s): client_socket(s) {}


            /**
             * \brief Create copy of \a o by duplicating it's socket
             * descriptor
             */
            socket (const socket &o): client_socket(o) {}


            /**
             * \brief Take ownership of socket \a s i.e. caller shouldn't use
             * it anymore.
             */
            socket &operator = (socket_type s) {
                client_socket::operator=(s);
                return *this;
            }
      };


      /**
       * \brief Listener (or acceptor) socket.
       *
       * A listener socket accepts incoming connection requests, returning
       * OS-level socket descriptor that represents communication channel.
       */
      class listener: public server_socket
      {
        public:

            /** \brief Create new server socket descriptor */
            listener (): server_socket(local) { }


            /**
             * \brief Create server socket object from already existing socket
             * descriptor.
             *
             * \note Specified socket descriptor \a s is not duplicated i.e.
             * ownership of descriptor goes to new object.
             */
            listener (socket_type s): server_socket(s) {}


            /**
             * \brief Create copy of \a o by duplicating it's socket
             * descriptor
             */
            listener (const listener &o): server_socket(o) {}


            /**
             * \brief Create new server socket object that is bound to \a a
             * and has \a backlog queue length
             *
             * This constructor also enables reuse of address.
             */
            listener (const address &a, int backlog = somaxconn)
                : server_socket(local) {
                reuseaddr(true);
                bind(a);
                listen(backlog);
            }


            /**
             * \brief Take ownership of socket \a s i.e. caller shouldn't use
             * it anymore.
             */
            listener &operator = (socket_type s) {
                server_socket::operator=(s);
                return *this;
            }
      };
};


/** \brief Return true, if \a a1 and \a a2 are equal */
inline bool
operator == (const local::address &a1, const local::address &a2)
{
  return a1.sa_.sun_family == a2.sa_.sun_family &&
      strncmp(a1.sa_.sun_path, a2.sa_.sun_path, sizeof(a1.sa_.sun_path)) == 0;
}


/** \brief Return true, if \a a1 and \a a2 are not equal */
inline bool
operator != (const local::address &a1, const local::address &a2)
{
  return !(operator==(a1, a2));
}


/** \brief Store Unix address to stream by writing out path name */
inline STD_ ostream &
operator << (STD_ ostream &os, const local::address &a)
{
  os << a.c_str();
  return os;
}


/** \brief Parse local address textual into object */
template<> inline local::address
as<local::address> (const char *as_string, char ** /*unused*/)
{
  return local::address(as_string);
}


/** \} */
}   // namespace cool


#endif  /* cool_net_local_socket_hpp_included */
