/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_poller_hpp_included
#define cool_net_poller_hpp_included

/**
 * \file poller.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::poller, cool::poller::iterator
 *
 * In module \ref cool_net.
 */

#include <cool/cool.hpp>
#include <cool/net/socket.hpp>
#include "bits/poller.hpp"


namespace cool {
/** \addtogroup cool_net */
/** \{ */


/**
 * \brief Container class for cool::socket that can watch status change for
 * each descriptor
 *
 * There are three independent event on descriptor that can be watched:
 *  - descriptor becomes readable
 *  - descriptor becomes writable
 *  - descriptor has pending exception (OOB data or smth).
 *
 * Typical usage of poller:
 * \code
 * cool::poller p;
 * p.insert(...);
 * p.insert(...);
 *
 * while (true) {
 *     cool::poller::iterator i = p.begin(cool::time(3));
 *     if (i == p.end())
 *         // timeout -- continue or do some idle stuff
 *         continue;
 *
 *     for ( ;  i != p.end();  ++i) {
 *         if (i->readable()) {
 *             // *i is now readable
 *             // if recv() returns 0, then peer has closed connection
 *         }
 *
 *         if (i->writable()) {
 *             // *i is now writable
 *         }
 *
 *         if (i->has_exception()) {
 *             // *i has OOB data or smth
 *         }
 *     }
 * }
 * \endcode
 *
 * For more information about non-blocking readable sockets, see
 * cool::socket::readable().
 *
 * \note On POSIX platforms any descriptor can be watched (files, sockets
 * etc). On Windows platform only sockets are allowed.
 */
class COOL_API poller: private noncopyable
{
  public:

      /** \brief Create new empty poller */
      poller () {}


      /**
       * \brief Destroy poller object.
       *
       * Destructor does nothing with previously added sockets. It is
       * application's responsibility to keep track of them and close each (or
       * do whatever is neccessary).
       */
      virtual ~poller () {}


      /** \brief Return number of sockets added to poller. */
      size_t size () const {
          return poller_.size();
      }


      /**
       * \brief Add new socket to be watched for events \a ev.
       *
       * \a ev is OR-ed value of flags cool::socket::in, cool::socket::out and
       * cool::socket::except. Only specified events are watched for this
       * socket.
       *
       * For efficiency, do not add sockets that you do not have indent to
       * check it's result after wait() and respond appropriately. Any socket
       * that is cool::socket::writable() must be written to, etc.
       *
       * Adding socket multiple times causes assertion thrown. If you need to
       * modify event flags to be watched, use modify().
       *
       * If socket is already registered in poller, assertion is thrown (not
       * exception) because it is rather programming error than runtime error.
       */
      void insert (socket *s, socket::event_type ev = socket::in) {
          assert(s != 0);
          assert(*s > -1);
          //assert(ev != socket::no_events);

          poller_.insert(s, ev);
      }


      /**
       * \brief Change socket event flags.
       *
       * \see insert()
       *
       * If \a s is not found in internal sockets set, assertion is thrown.
       */
      void modify (socket *s, socket::event_type ev) {
          assert(s != 0);
          assert(*s > -1);
          //assert(ev != socket::no_events);

          poller_.modify(s, ev);
      }


      /** \brief Remove \a s from sockets set. */
      void erase (socket *s) {
          assert(s != 0);
          assert(*s > -1);

          poller_.erase(s);
      }


      /**
       * \brief Socket set iterator that iterates only descriptors that have
       * status changed (i.e. become readable etc).
       *
       * Programmer should make no assumptions about ordering of iterated
       * sockets as it depends on low-level multiplexor function.
       */
      class iterator
      {
        public:

            /**
             * \brief Create new invalid iterator.
             *
             * Use of invalid iterator is undefined behaviour. Iterator is
             * correctly initialized only using copy constructor (other
             * iterator must be valid, of course) or returned by
             * cool::poller::wait().
             */
            iterator (): owner_(0), current_(0) {}


            /**
             * \brief Return pointer to current socket that have status
             * changed
             */
            operator socket * () const {
                return current_;
            }


            /**
             * \brief Return pointer to current socket that have status
             * changed
             */
            socket *operator -> () const {
                return current_;
            }


            /**
             * \brief Increment iterator to point to next descriptor that has
             * changed status.
             *
             * \return Iterator pointing to next socket or poller::end()
             * if no more sockets.
             */
            iterator &operator ++ () {
                next_pending();
                return *this;
            }


            /**
             * \brief Increment iterator to point to next descriptor that has
             * changed status.
             *
             * \return Iterator pointing to socket that was active before
             * incrementing.
             */
            iterator operator ++ (int) {
                iterator tmp = *this;
                next_pending();
                return tmp;
            }


            /** \brief Return true if iterator is still valid */
            operator bool () const {
                return current_ != 0;
            }


            /**
             * \brief Return true if this iterator points to same socket
             * as \a o
             */
            bool operator == (const iterator &o) const {
                return current_ == o.current_;
            }


            /**
             * \brief Return true if this iterator points to different socket
             * as \a o
             */
            bool operator != (const iterator &o) const {
                return !(operator==(o));
            }


        private:

            iterator (poller *owner): owner_(owner), current_(0) {
                next_pending();
            }

            void next_pending () {
                if (!owner_)
                    return;

                socket::event_type ev;
                if ((current_ = owner_->poller_.pending(ev)) != 0)
                    // set socket internal last_wait_
                    owner_->update_socket_event(current_, ev);
                else
                    // hit the end of socket collection
                    owner_ = 0;
            }

            poller *owner_;
            socket *current_;

            friend class poller;
      };


      /**
       * \brief Wait indefinitely until any of watched sockets changes it's
       * state i.e. event happens.
       *
       * \return Iterator that points to first active socket or end() if there
       * is no such socket.
       */
      iterator begin () {
          if (poller_.wait() > 0)
              return iterator(this);
          return iterator(0);
      }


      /**
       * \brief Wait up to timeout until any of watched sockets changes it's
       * state i.e. event happens.
       *
       * \return Iterator that points to first active socket or end() if there
       * is no such socket.
       */
      iterator begin (const time::time &timeout) {
          if (poller_.wait(&timeout) > 0)
              return iterator(this);
          return 0;
      }


      /** \brief Return past of sockets collection i.e. invalid iterator */
      iterator end () const {
          return iterator(0);
      }


  private:

      friend class iterator;

      void update_socket_event (socket *s, socket::event_type e) {
          s->last_wait_ = e;
      }

      bits::net::poller_type poller_;
};


/** \} */
}   // namespace cool


#endif  /* cool_net_poller_hpp_included */
