/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_server_socket_hpp_included
#define cool_net_server_socket_hpp_included

/**
 * \file server_socket.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::server_socket
 *
 * In module \ref cool_net.
 */

#include <cool/cool.hpp>
#include <cool/net/socket.hpp>
#include "bits/server_socket.hpp"


namespace cool {
/** \addtogroup cool_net */
/** \{ */


/**
 * \brief Base class for server side sockets.
 *
 * This base class is not meant to be used directly but provides only
 * properties and methods common to server sockets. Applications should use
 * cool::inet::listener or cool::local::listener instead.
 */
class COOL_API server_socket: public socket
{
  public:

      /**
       * \brief Create object from already existing socket descriptor.
       *
       * \note Specified socket descriptor \a s is not duplicated i.e.
       * ownership of descriptor goes to new object which's destructor closes
       * this socket descriptor.
       */
      server_socket (socket_type s): socket(s) { }


      /** \brief Create new socket descriptor */
      server_socket (family_type family)
          : socket(family, stream) { }


      /** \brief Create copy of \a o by duplicating it's socket descriptor */
      server_socket (const server_socket &o): socket(o) { }


      /**
       * \brief Take ownership of socket \a s i.e. caller shouldn't use it any
       * more.
       */
      server_socket &operator = (socket_type s) {
          socket::operator=(s);
          return *this;
      }


      /**
       * \brief Mark this socket as listening socket that accepts incoming
       * connection requests.
       *
       * \a backlog specifies how many pending connections can be queued by
       * operating system without accepting before refusing to accept more
       * incoming connection requests.
       */
      server_socket &listen (int backlog = somaxconn) {
          bits::net::listen(s_, backlog);
          return *this;
      }


      /** \brief Accept new incoming connection request */
      socket_type accept () {
          return bits::net::accept(s_, 0);
      }


      /**
       * \brief Accept new incoming connection request.
       *
       * Store address of other communication endpoint to \a a.
       */
      socket_type accept (address &a) {
          return bits::net::accept(s_, &a);
      }
};


/** \} */
}   // namespace cool


#endif  /* cool_net_server_socket_hpp_included */
