/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_socket_hpp_included
#define cool_net_socket_hpp_included

/**
 * \file socket.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::socket, cool::socket::address
 *
 * In module \ref cool_net.
 */

#include <cool/cool.hpp>
#include <cool/util/noncopyable.hpp>
#include <cool/net/socket_base.hpp>
#include <cool/time/time.hpp>
#include "bits/socket.hpp"


namespace cool {
/** \addtogroup cool_net */
/** \{ */


/**
 * \brief Class socket is used as base for all types of socket and it
 * has a system socket handle (i.e. descriptor).
 *
 * This base class is not meant to be used directly but provides only
 * properties common to inherited socket classes. Actually it is very thin
 * wrapper around OS-level socket routines.
 *
 * \note Destructor does not close socket descriptor. It must be done by
 * caller or higher level object.
 *
 * \todo multicast sockets if needed
 */
class COOL_API socket: public socket_base
{
  public:


      /** \brief Abstract base class for specific domain address classes */
      class address
      {
        public:

            /** \brief Non-op */
            virtual ~address () {}


            /** \brief Return pointer to socket address structure */
            virtual operator void * () const = 0;


            /** \brief Return pointer to specific socket address structure */
            operator sockaddr * () const {
                return addr();
            }


            /** \brief Return pointer to specific socket address structure */
            virtual sockaddr *addr () const = 0;


            /** \brief Return size of socket address structure */
            virtual socklen_type size () const = 0;


            /** \brief Return family of socket address */
            virtual family_type family () const = 0;
      };


      /** \brief Create new object with invalid socket */
      socket (): s_(-1), last_wait_(no_events) { }


      /**
       * \brief Create object from already existing socket descriptor
       *
       * \note Specified socket descriptor \a s is not duplicated i.e.
       * ownership of descriptor goes to new object.
       */
      socket (socket_type s): s_(s), last_wait_(no_events) { }


      /** \brief Create new socket descriptor */
      socket (family_type family, style_type style = def::style)
          : s_(-1), last_wait_(no_events) {
          // family parameter not default to create difference with
          // parameterless constructor
          open(family, style);
      }


      /** \brief Create copy of \a o */
      socket (const socket &o): s_(o.s_), last_wait_(o.last_wait_) { }


      /**
       * \brief Destroy socket object
       *
       * Destructor does not close socket descriptor. It must be done by
       * caller or higher level object.
       */
      virtual ~socket () throw() { }


      /** \brief Set new socket descriptor, closing old first if neccessary */
      socket &operator = (socket_type s) {
          s_ = s;
          return *this;
      }


      /** \brief Return OS socket descriptor */
      operator socket_type () const {
          return handle();
      }


      /** \brief Return OS socket descriptor */
      socket_type handle () const {
          return s_;
      }


      /** \brief Return duplicate of OS socket descriptor */
      socket_type clone () const {
          if (s_ == -1)
              return -1;
          return bits::net::clone(s_);
      }


      /** \brief Return duplicate of external socket \a s */
      static socket_type clone (socket_type s) {
          return bits::net::clone(s);
      }


      /** \brief Return true if socket descriptor is invalid */
      bool operator ! () const {
          return s_ == -1;
      }


      /** \brief Create new socket descriptor */
      socket &open (family_type f = def::family, style_type s = def::style) {
          close();
          s_ = bits::net::open(f, s);
          last_wait_ = no_events;
          return *this;
      }


      /** \brief Release socket descriptor */
      socket &close () {
          if (s_ != -1) {
              bits::net::close(s_);
              s_ = -1;
          }
          last_wait_ = no_events;
          return *this;
      }


      /** \brief Bind socket to specific network interface */
      socket &bind (const address &a) {
          assert(s_ != -1);
          bits::net::bind(s_, a.addr(), a.size());
          return *this;
      }


      /**
       * \brief Wait indefinitely until at least one of events specified in
       * \a ev happens on socket.
       *
       * Result of last wait() can be checked with method pending_events() or
       * readable() | writable() | has_exception()
       */
      socket &wait (event_type ev = in | out) {
          assert(s_ != -1);
          last_wait_ = bits::net::wait(s_, ev, 0);
          return *this;
      }


      /**
       * \brief Wait up to \a timeout until at least one of events specified
       * in \a ev happens on socket.
       *
       * Result of last wait() can be checked with method pending_events() or
       * readable() | writable() | has_exception()
       */
      socket &wait (const time::time &timeout, event_type ev = in | out) {
          assert(s_ != -1);
          last_wait_ = bits::net::wait(s_, ev, &timeout);
          return *this;
      }


      /** \brief Return result of last wait() */
      event_type pending_events () const {
          return last_wait_;
      }


      /** \brief Return true if socket is readable after last wait() */
      bool readable () const {
          return last_wait_ & in;
      }


      /** \brief Return true if socket is writable after last wait() */
      bool writable () const {
          return last_wait_ & out;
      }


      /** \brief Return true if socket has exception after last wait() */
      bool has_exception () const {
          return last_wait_ & except;
      }


      /** \brief Return type of socket */
      style_type style () const {
          int s = 0;
          getopt(sock_style, &s, sizeof(s));
          return style_type(s);
      }


      /** \brief Get and clear then pending socket error */
      os_error::error_type pending_error () {
          os_error::error_type e = 0;
          getopt(error, &e, sizeof(e));
          return e;
      }


      /** \brief Get socket_base::debug setting */
      bool debug () const {
          int d;
          getopt(socket_base::debug, &d, sizeof(d));
          return d;
      }


      /** \brief Set socket_base::debug setting */
      socket &debug (bool on) {
          int d = on;
          return setopt(socket_base::debug, &d, sizeof(d));
      }


      /** \brief Get socket_base::reuseaddr setting */
      bool reuseaddr () const {
          int d;
          getopt(socket_base::reuseaddr, &d, sizeof(d));
          return d;
      }


      /** \brief Set socket_base::reuseaddr setting */
      socket &reuseaddr (bool on) {
          int d = on;
          return setopt(socket_base::reuseaddr, &d, sizeof(d));
      }


      /** \brief Get socket_base::keepalive setting */
      bool keepalive () const {
          int d;
          getopt(socket_base::keepalive, &d, sizeof(d));
          return d;
      }


      /** \brief Set socket_base::keepalive setting */
      socket &keepalive (bool on) {
          int d = on;
          return setopt(socket_base::keepalive, &d, sizeof(d));
      }


      /** \brief Get socket_base::dontroute setting */
      bool dontroute () const {
          int d;
          getopt(socket_base::dontroute, &d, sizeof(d));
          return d;
      }


      /** \brief Set socket_base::keepalive setting */
      socket &dontroute (bool on) {
          int d = on;
          return setopt(socket_base::dontroute, &d, sizeof(d));
      }


      /** \brief Get socket_base::broadcast setting */
      bool broadcast () const {
          int d;
          getopt(socket_base::broadcast, &d, sizeof(d));
          return d;
      }


      /** \brief Set socket_base::broadcast setting */
      socket &broadcast (bool on) {
          int d = on;
          return setopt(socket_base::broadcast, &d, sizeof(d));
      }


      /** \brief Get socket_base::oobinline setting */
      bool oobinline () const {
          int d;
          getopt(socket_base::oobinline, &d, sizeof(d));
          return d;
      }


      /** \brief Set socket_base::oobinline setting */
      socket &oobinline (bool on) {
          int d = on;
          return setopt(socket_base::oobinline, &d, sizeof(d));
      }


      /** \brief Return socket_base::linger setting */
      linger_type linger () const {
          linger_type d(0, 0);
          getopt(socket_base::linger, &d, sizeof(d));
          return d;
      }


      /** \brief Set socket_base::oobinline setting */
      socket &linger (linger_type d) {
          return setopt(socket_base::linger, &d, sizeof(d));
      }


      /** \brief Return socket_base::sndbuf setting */
      int sendbufsz () const {
          int d = 0;
          getopt(sndbuf, &d, sizeof(d));
          return d;
      }


      /** \brief Set socket_base::sndbuf setting */
      socket &sendbufsz (int d) {
          return setopt(sndbuf, &d, sizeof(d));
      }


      /** \brief Return socket_base::rcvbuf setting */
      int recvbufsz () const {
          int d = 0;
          getopt(rcvbuf, &d, sizeof(d));
          return d;
      }


      /** \brief Set socket_base::rcvbuf setting */
      socket &recvbufsz (int d) {
          return setopt(rcvbuf, &d, sizeof(d));
      }


      /** \brief Return true if socket operations could block caller */
      bool blocking () const {
          assert(s_ != -1);
          return bits::net::getblock(s_);
      }


      /** \brief If \a on is false, set socket to non-blocking I/O mode */
      socket &blocking (bool on) {
          assert(s_ != -1);
          bits::net::setblock(s_, on);
          return *this;
      }


      /**
       * \brief Return true if there is out-of-band data waiting in
       * communication channel
       */
      bool atmark () const {
          assert(s_ != -1);
          return bits::net::atmark(s_);
      }


      /**
       * \brief If \a on is true, set socket for asynchronous I/O mode.
       *
       * If read/write is possible on socket, the signal \c SIGIO will be
       * delivered to process group which id is set with pgrp(int).
       */
      socket &async (bool on) {
          assert(s_ != -1);
          bits::net::async(s_, on);
          return *this;
      }


      /**
       * \brief Return process group id that would receive SIGIO and SIGURG
       * signals.
       */
      int pgrp () const {
          assert(s_ != -1);
          return bits::net::getpgrp(s_);
      }


      /**
       * \brief Set process group id that would receive SIGIO and SIGURG
       * signals.
       */
      socket &pgrp (int grp) {
          assert(s_ != -1);
          bits::net::setpgrp(s_, grp);
          return *this;
      }


      /**
       * \brief Return number of bytes in socket communication channel ready
       * for reading
       */
      long in_avail () const {
          assert(s_ != -1);
          return bits::net::in_avail(s_);
      }


      /**
       * \brief Scoped socket close.
       *
       * Socket class's destructor does not close automatically underlying
       * descriptor. Automatic close can be done through socket::scoped_close
       * which's destructor does this.
       *
       * Example usage:
       * \code
       * cool::socket s = ...;
       *
       * {
       *   cool::socket::scoped_close socket_guard(s);
       *   ...
       *
       *   // s is closed when variable socket_guard goes out of scope.
       * }
       *
       * // socket s is closed
       * \endcode
       */
      class scoped_close: private noncopyable
      {
        public:

            /** \brief Does nothing except initializing internal fields */
            scoped_close (socket &s): s_(s) {}

            /** \brief Close socket descriptor when leaving scope */
            ~scoped_close () { s_.close(); }

        private:

            socket &s_;
      };


  protected:

      /** \brief Set socket option value */
      int getopt (option_type o, void *buf, int len,
              level_type l = socket_level) const {
          assert(s_ != -1);
          return bits::net::getopt(s_, l, o, buf, len);
      }


      /** \brief Set socket option value */
      socket &setopt (option_type o, void *buf, int len,
              level_type l = socket_level) {
          assert(s_ != -1);
          bits::net::setopt(s_, l, o, buf, len);
          return *this;
      }

      /** \brief Low-level OS socket descriptor */
      socket_type s_;


  private:

      // Result of last wait()
      event_type last_wait_;

      // give poller access to last_wait_
      friend class poller;
};


/** \} */
}   // namespace cool


#endif  /* cool_net_socket_hpp_included */
