/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_socket_base_hpp_included
#define cool_net_socket_base_hpp_included

/**
 * \file socket_base.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::socket_base
 *
 * In module \ref cool_net.
 */

#include <cool/cool.hpp>
#include "bits/socket_base.hpp"


namespace cool {
/** \addtogroup cool_net */
/** \{ */


/** \brief Common stuff for all socket related classes */
class COOL_API socket_base
{
  public:

      /** \brief OS socket descriptor type */
      typedef bits::net::socket_type socket_type;

      /** \brief Socket address length */
      typedef bits::net::socklen_type socklen_type;

      /** \brief Enumerates protocol (or address) families */
      typedef bits::net::family_type family_type;

      /** \brief Enumerates socket styles */
      typedef bits::net::style_type style_type;

      /** \brief Enumerates shutdown modes */
      typedef bits::net::shutmode_type shutmode_type;

      /** \brief Enumerates socket I/O flags */
      typedef bits::net::io_flags_type io_flags_type;

      /** \brief Enumerates socket options */
      typedef bits::net::option_type option_type;

      /** \brief Enumerates socket option levels */
      typedef bits::net::level_type level_type;

      /** \brief Enumerates socket event types */
      typedef bits::net::event_type event_type;

      /** \brief Socket lingering option data */
      typedef bits::net::linger_type linger_type;


      /** \{ */
      /** \brief UNIX socket based on local file system */
      static family_type local;

      /** \brief TCP/IP socket */
      static family_type inet;
      /** \} */


      /** \{ */
      /**
       * \brief Provide sequenced, reliable, two-way, connection-based byte
       * streams with support of out-of-band data.
       *
       * In the socket_base::local (UNIX) domain, this type of socket works
       * like pipe. In socket_base::inet (IP) domain, type is implemented
       * on the TCP/IP protocol.
       */
      static style_type stream;

      /**
       * \brief Provides datagrams, which are connectionless messages of a
       * fixed maximum length.
       *
       * In the socket_base::local (UNIX) domain, this type is similar to a
       * message queue. In socket_base::inet (IP) domain, type is implemented
       * on the UDP/IP protocol.
       *
       * Datagram socket support bidirectional flow of data that is
       * unreliable, not sequenced and can be duplicated.
       */
      static style_type datagram;

      /**
       * \brief Sequenced reliable, two-way connection-based data transmission
       * for datagrams with fixed maximum length.
       *
       * A consumer is required to read an entire packet with each read()
       * system call.
       */
      static style_type seqpacket;

      /**
       * \brief Provides raw network protocol access.
       *
       * Available only with root privileges.
       */
      static style_type raw;

      /**
       * \brief Provides reliable datagram layer that does not guarantee
       * ordering.
       */
      static style_type rdm;
      /** \} */


      /** \{ */
      /** \brief Shutdown socket for reading */
      static shutmode_type r;

      /** \brief Shutdown socket for writing */
      static shutmode_type w;

      /** \brief Shutdown socket for reading and writing */
      static shutmode_type rw;
      /** \} */


      /** \{ */
      /** \brief Enable socket debugging. */
      static option_type debug;

      /** \brief Allow reuse of local address. */
      static option_type reuseaddr;

      /**
       * \brief Enable sending keep-alive messages for connection-oriented
       * sockets.
       */
      static option_type keepalive;

      /**
       * \brief Do not send via gateway, only send to hosts connected directly
       * to network.
       */
      static option_type dontroute;

      /**
       * \brief When enabled, datagram sockets receive packets sent to
       * broadcast address and they are allowed to send packets to broadcast
       * addresses.
       */
      static option_type broadcast;

      /**
       * \brief Socket lingering options.
       *
       * When enabled, close() or shutdown() blocks caller until all queued
       * messages have been successfully sent or the linger timeout has been
       * reached.
       *
       * Otherwise, the call will return immediately and the closing is done
       * in the background.
       */
      static option_type linger;

      /**
       * \brief When enabled, out-of-band data is directly placed into the
       * receive data stream.
       *
       * Otherwise OOB data is only passed when oob io_flag is set during
       * receiving.
       */
      static option_type oobinline;

      /** \brief Max socket send buffer size in bytes. */
      static option_type sndbuf;

      /** \brief Max socket receive buffer size in bytes. */
      static option_type rcvbuf;

      /** \brief Used to get pending error. */
      static option_type error;

      /** \brief Used to query socket style */
      static option_type sock_style;
      /** \} */


      /** \{ */
      /** \brief Socket level option */
      static level_type socket_level;

      /** \brief TCP protocol level option */
      static level_type tcp_level;
      /** \} */


      /** \{ */
      /** \brief Request receipt or transmission of out-of-band data */
      static io_flags_type oob;

      /**
       * \brief Causes the receive operation to return data from beginning if
       * the receive buffer without removing that data from buffer.
       */
      static io_flags_type peek;

      /**
       * \brief Request that operation blocks until the full request is
       * satisfied.
       *
       * However, the call may still return less data than requested if signal
       * is caught, an error or disconnect occurs, or the next data to be
       * received is a different type than requested.
       */
      static io_flags_type waitall;

      /**
       * \brief Return the real length of the packet, even when it was longer
       * than passed buffer.
       *
       * Only valid for packet sockets.
       */
      static io_flags_type trunc;
      /** \} */


      /** \{ */
      /** \brief No events on socket */
      static event_type no_events;
      /** \brief Socket can be read without blocking */
      static event_type in;
      /** \brief Socket can be written without blocking */
      static event_type out;
      /** \brief Socket has pending exceptions (i.e. out-of-band data) */
      static event_type except;
      /** \} */


      /** \brief Values for method's default parameters */
      struct def {
          static socket_base::family_type    family;    /**< Default family */
          static socket_base::style_type     style;     /**< Default style */
          static socket_base::io_flags_type  io_flags;  /**< Default io_flags */
      };


      /** \brief Number of maximum pending connections in accept() queue */
      static int somaxconn;
};


/** \} */
}   // namespace cool


#endif  /* cool_net_socket_base_hpp_included */
