/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_socketbuf_hpp_included
#define cool_net_socketbuf_hpp_included

/**
 * \file socketbuf.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::socketbuf
 *
 * In module \ref cool_netstream.
 */

#include <cool/cool.hpp>
#include <cool/net/socket.hpp>
#include <cool/net/exception.hpp>
#include "bits/socketbuf.hpp"


namespace cool {
/** \addtogroup cool_netstream */
/** \{ */


/**
 * \brief Base device class for socket streams.
 *
 * This class does actual work of socket input and output. It is inherited
 * from standard iostream class \c streambuf, which manages two byte sequences
 * -- input and output buffers.
 *
 * This class is designed to 'is-a' socket not 'has-a' socket, so it can be
 * used with cool::poller. Although socket is a parent with public access,
 * clean application should not use it's methods directly as it can mess up
 * buffering.
 *
 * Template parameter family is either cool::local or cool::inet.
 */
template <typename family>
class COOL_API socketbuf: public STD_ streambuf, public family::socket
{
  public:

      /** \{ */
      /** \brief Instantiated socket types */
      typedef typename family::address address_type;    /**< Address */
      typedef typename family::socket socket_type;      /**< Socket */
      typedef typename family::listener listener_type;  /**< Server socket */
      /** \} */


      /** \{ */
      /**
       * \brief Standard types.
       *
       * Right now they have little use until socket streams evolve to be more
       * standard compliant i.e. templatized character types etc.
       */
#if defined(SS_HAVE_STREAMBUF_CHARTYPE)
      typedef STD_ streambuf::char_type char_type;  /**< Character */
      typedef STD_ streambuf::int_type int_type;    /**< Integer */
#else
      typedef char char_type;
      typedef int int_type;
#endif
      /** \} */


      /** \brief Creates new socketbuf with low-level socket */
      socketbuf ()
          : STD_ streambuf()
          , socket_type(socket::stream), io_flags_(socket::def::io_flags)
          , throw_timeout_(false), send_timeout_(-1), recv_timeout_(-1)
          , buf_(0), buf_size_(buffer_size), allocated_(false) { }


      /** \brief Constructor that uses existing low-level socket */
      socketbuf (const socket_type &s)
          : STD_ streambuf()
          , socket_type(s), io_flags_(socket::def::io_flags)
          , throw_timeout_(false), send_timeout_(-1), recv_timeout_(-1)
          , buf_(0), buf_size_(buffer_size), allocated_(false) {
          if (is_open())
              // already opened, create buffer
              create_buf();
      }


      /** \brief Destructor: close underlying socket */
      virtual ~socketbuf () throw () {
          close();
          destroy_buf();
      }


      /**
       * \brief Open connection to remote address.
       *
       * Returns pointer to self if establishing connection succeeded.
       * For more information about differences of connecting blocking and
       * non-blocking sockets, see cool::client_socket::connect(). Short
       * version: for non-blocking sockets, call this method multiple times
       * until it succeeds or timeout happens.
       */
      STD_ streambuf *open (const address_type &a) {
          if (static_cast<socket_type &>(*this) == -1)
              // socket not created yet, do it
              socket_type::open(family::family, socket::stream);
          if (socket_type::connect(a) == true) {
              if (!buf_)
                  // buffer not created yet, do it
                  create_buf();
              return this;
          }
          return 0;
      }


      /**
       * \brief Close connection to remote address.
       *
       * If there is data in put buffer, it is flushed first before closing.
       */
      STD_ streambuf *close () throw() {
          try {
              if (!is_open())
                  return 0;

              // flush send buffer if there is anything
              if (pptr() && pbase() < pptr())
                  sync();
          }
          catch (...) {}

          socket_type::close();
          return this;
      }


      /** \brief Return true if socket is open */
      bool is_open () const throw () {
          return static_cast<socket_type>(*this) != -1 &&
              socket_type::is_connected();
      }


      /**
       * \brief Close part of full-duplex communication channel.
       *
       * This method fails immediately if connection is not established,
       * returning \c NULL. On success \c this is returned.
       */
      socketbuf *shutdown (socket::shutmode_type m) {
          if (is_open()) {
              socket_type::shutdown(m);
              return this;
          }
          return 0;
      }


      /** \brief Set default I/O flags and return old flags */
      socket::io_flags_type setf (socket::io_flags_type flags) {
          socket::io_flags_type old = io_flags_;
          io_flags_ = flags;
          return old;
      }


      /**
       * \brief Get socket data sending timeout in milliseconds
       *
       * Possible values:
       *    - -1: depends on socket blocking option
       *    - 0: poll
       *    - >0: timeout in milliseconds
       *
       * If synchronizing methods can't send or receive data during timeout
       * and throw_timeout() == true, then exception is thrown
       * (cool::timeout_error)
       */
      int send_timeout () const {
          return send_timeout_;
      }


      /**
       * \brief Set socket data sending timeout.
       *
       * \copydoc send_timeout()
       */
      socketbuf &send_timeout (int msecs) {
          send_timeout_ = msecs;
          return *this;
      }


      /**
       * \brief Get socket data receiving timeout in milliseconds
       *
       * \copydoc send_timeout()
       */
      int recv_timeout () const {
          return recv_timeout_;
      }


      /**
       * \brief Set socket data receiving timeout.
       *
       * \copydoc send_timeout()
       */
      socketbuf &recv_timeout (int msecs) {
          recv_timeout_ = msecs;
          return *this;
      }


      /** \brief Returns true if cool::timeout_error is thrown on timeout */
      bool throw_timeout () const {
          return throw_timeout_;
      }


      /** \brief Set flag to throw cool::timeout_error on I/O timeout */
      socketbuf &throw_timeout (bool do_throw) {
          throw_timeout_ = do_throw;
          return *this;
      }


  protected:

      /** \brief Return number of byte(s) available in buffer */
      virtual streamsize showmanyc () SS_SHOWMANYC_CONST {
          return is_open() ? egptr() - gptr() : -1;
      }


      /**
       * \brief Fetch data from input buffer.
       *
       * If input buffer is exhausted, fill buffer with data received from
       * socket.
       *
       * Returns next available character or EOF data is no available in
       * socket input space.
       *
       * If other end closed connection, connection_error is thrown.
       */
      virtual int_type underflow () {
          char_type *cur = gptr();
          if (!cur)
              return eof_;

          if (cur < egptr())
              return (char_type)*cur;

          char_type *beg = eback();
          int rlen = 0;
          if (check_recv())
              if ((rlen = socket_type::recv(beg, gend_ - beg, io_flags_)) == 0)
                  raise<connection_error>(COOL_LOCUS);

          if (rlen < 1)
              return eof_;

          setg(beg, beg, beg + rlen);
          return (char_type)*beg;
      }


      /**
       * \brief Fetch data from input buffer and return next available
       * character.
       *
       * \see underflow()
       */
      virtual int_type uflow () {
          int_type r = underflow();
          if (r != eof_)
              gbump(1);
          return r;
      }


      /**
       * \brief Consume data from the buffer. If buffer is full, send data to
       * socket.
       */
      virtual int_type overflow (int_type c = eof_) {
          if (pbase() == 0)
              return eof_;

          if (c == eof_)
              return sync();

          if (pptr() == epptr())
              sync();

          *pptr() = (char_type)c;
          pbump(1);

          return c;
      }


      /** \brief Synchronize send buffer with socket */
      virtual int sync () {
          char_type *cur = pptr();
          char_type *beg;

          if (cur && cur > (beg = pbase())) {
              int wrlen = 0;
              int len = cur - beg;
              while (len > 0) {
                  try {
                      int slen = 0;
                      if (check_send())
                          slen = socket_type::send(beg + wrlen, len, io_flags_);

                      if (slen < 1)
                          return -1;

                      len -= slen;
                      wrlen += slen;
                  }
                  catch (exception &e) {
                      if (wrlen) {
                          register char *pto = beg;
                          register char *pfrom = pto + wrlen;
                          len = cur - beg - wrlen;
                          while (pfrom < cur)
                              *pto++ = *pfrom++;
                          setp(beg, epptr());
                          pbump(len);
                      }

                      // rethrow error
                      e.locus = COOL_LOCUS;
                      throw;
                  }

                  setp(beg, epptr());
              }
          }

          return 0;
      }


      /**
       * \brief Read data from input buffer to \a s not exceeding \a n byte(s).
       *
       * If input buffer is exhausted, receive data from socket (if
       * available).
       */
      virtual streamsize xsgetn (char_type *s, streamsize n) {
          int to_read = n;

          while (to_read > 0) {
              streamsize av = showmanyc();
              int r = av < to_read ? av : to_read;

              memcpy(s, gptr(), r*sizeof(char_type));
              gbump(r);
              to_read -= r;
              s += r;

              if (to_read && underflow() == eof_)
                  break;
          }

          return n - to_read;
      }


      /**
       * \brief Write \a n byte(s) from \a s to output buffer.
       *
       * If buffer is full, send data to socket.
       */
      virtual streamsize xsputn (char_type *s, streamsize n) {
          int to_write = n;

          while (to_write > 0) {
              char_type *cur = pptr();
              int av = epptr() - cur;
              int w = av < to_write ? av : to_write;

              memcpy(cur, s, w*sizeof(char_type));
              pbump(w);
              to_write -= w;
              s += w;

              if (to_write && overflow() == eof_)
                  break;
          }

          return n - to_write;
      }


      /**
       * \brief Set application defined I/O buffer.
       *
       * Buffer \a buf is divided to halves -- equally to both buffer.
       *    - If \a buf is not NULL and \a bufsz > 2 then specified buffer is
       *      used for buffering.
       *    - If \a buf is NULL and \a bufsz > 2 then new buffer with \a bufsz
       *      byte(s) is allocated using new.
       *    - If \a buf is NULL or bufsz < 2 then I/O is not buffered.
       */
      virtual STD_ streambuf *setbuf (char_type *buf, streamsize bufsz) {
          destroy_buf();
          buf_ = buf;
          buf_size_ = bufsz;
          create_buf();

          return this;
      }


  private:

      enum { eof_ = EOF };
      enum { buffer_size = 1024 };


      void create_buf () {
          if (!buf_ || (unsigned)buf_size_ < sizeof(buf_static_)) {
              if ((unsigned)buf_size_ > sizeof(buf_static_)) {
                  // allocate buffer
                  buf_ = new char_type[buf_size_];
                  allocated_ = true;
              }
              else {
                  // use internal static buffer
                  buf_ = buf_static_;
                  buf_size_ = sizeof(buf_static_);
              }
          }

          // first half for put buffer
          char_type *pend_ = buf_ + (buf_size_ >> 1);
          setp(buf_, pend_);

          // other half for get buffer
          gend_ = buf_ + buf_size_;;
          setg(pend_, gend_, gend_);
      }


      void destroy_buf () {
          if (allocated_) {
              delete [] pbase();
              allocated_ = false;
          }
          buf_ = 0;
          buf_size_ = buffer_size;
          setg(0, 0, 0);
          setp(0, 0);
      }


      bool check_send () {
          if (send_timeout_ != -1) {
              if (send_timeout_) {
                  time::sec_type sec = send_timeout_/time::msec_per_sec;
                  time::nsec_type nsec =
                      (send_timeout_%time::msec_per_sec)*time::nsec_per_msec;
                  socket_type::wait(time::time(sec, nsec), socket::out);
              }
              else
                  socket_type::wait(socket::out);

              if (!socket_type::writable()) {
                  if (throw_timeout_)
                      raise<timeout_error>(COOL_LOCUS);
                  return false;
              }
          }

          return true;
      }


      bool check_recv () {
          if (recv_timeout_ != -1) {
              if (recv_timeout_) {
                  time::sec_type sec = recv_timeout_/time::msec_per_sec;
                  time::nsec_type nsec =
                      (recv_timeout_%time::msec_per_sec)*time::nsec_per_msec;
                  socket_type::wait(time::time(sec, nsec), socket::in);
              }
              else
                  socket_type::wait(socket::in);

              if (!socket_type::readable()) {
                  if (throw_timeout_)
                      raise<timeout_error>(COOL_LOCUS);
                  return false;
              }
          }

          return true;
      }


      // socket I/O stuff
      socket::io_flags_type io_flags_;
      bool throw_timeout_;
      int send_timeout_;
      int recv_timeout_;

      // buffering stuff
      char_type *buf_;
      streamsize buf_size_;
      char_type buf_static_[2];
      bool allocated_;

      // can't use egptr() as it can change with every underflow()
      char_type *gend_;
};


/** \} */
}   // namespace cool


#endif  /* cool_net_socketbuf_hpp_included */
