/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_socketstream_hpp_included
#define cool_net_socketstream_hpp_included

/**
 * \file socketstream.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::socketstream, cool::isocketstream,
 * cool::osocketstream
 *
 * In module \ref cool_netstream.
 */

#include <cool/cool.hpp>
#include <cool/net/socketbuf.hpp>
#include iostream_header


namespace cool {
/** \addtogroup cool_netstream */
/** \{ */


/**
 * \brief Input stream class.
 *
 * Provides member functions to assist in reading and interpreting input from
 * a socket stream buffer.
 */
template <typename family>
class COOL_API isocketstream: public STD_ istream
{
  public:

      /** \brief Family socket buffering device type */
      typedef socketbuf<family> socketbuf_type;

      /** \brief Family address type */
      typedef typename socketbuf_type::address_type address_type;

      /** \brief Family socket type */
      typedef typename socketbuf_type::socket_type socket_type;

      /** \brief Family listener type */
      typedef typename socketbuf_type::listener_type listener_type;


      /** \brief Create new socket stream */
      isocketstream (): STD_ istream(0), buffer_() {
          init(&buffer_);
      }


      /**
       * \brief Construct input socket stream object that automatically
       * opens connection to \a.
       */
      isocketstream (const address_type &a): STD_ istream(0), buffer_() {
          init(&buffer_);
          open(a);
      }


      /**
       * \brief Construct input socket stream object that uses low-level
       * socket \a s for underlying socket input operations
       */
      isocketstream (const socket_type &s): STD_ istream(0), buffer_(s) {
          init(&buffer_);
      }


      /**
       * \brief Construct input socket stream object from low-level socket
       * accepted by \a server object.
       */
      isocketstream (listener_type &server)
          : STD_ istream(0), buffer_(server.accept()) {
          init(&buffer_);
      }


      /** \brief Destruct object and close underlying stream device */
      virtual ~isocketstream () {}


      /** \brief Return pointer to underlying socket device */
      socketbuf_type *rdbuf () const {
          return const_cast<socketbuf_type *>(&buffer_);
      }


      /** \brief Return pointer to underlying socket device */
      socketbuf_type *operator -> () const {
          return rdbuf();
      }


      /** \brief Check if connection is established */
      bool is_open () {
          return buffer_.is_open();
      }


      /** \brief Open connection to address \a a */
      void open (const address_type &a) {
          if (!buffer_.open(a))
              setstate(failbit);
      }


      /** \brief Close connection */
      void close () {
          if (!buffer_.close())
              setstate(failbit);
      }


  private:

      socketbuf_type buffer_;
};


/**
 * \brief Output stream class.
 *
 * Provides member functions to perform output operations with a socket stream
 * buffer.
 */
template <typename family>
class COOL_API osocketstream: public STD_ ostream
{
  public:

      /** \brief Family socket buffering device type */
      typedef socketbuf<family> socketbuf_type;

      /** \brief Family address type */
      typedef typename socketbuf_type::address_type address_type;

      /** \brief Family socket type */
      typedef typename socketbuf_type::socket_type socket_type;

      /** \brief Family listener type */
      typedef typename socketbuf_type::listener_type listener_type;


      /** \brief Create new socket stream */
      osocketstream (): STD_ ostream(0), buffer_() {
          init(&buffer_);
      }


      /**
       * \brief Construct output socket stream object that automatically
       * opens connection to \a.
       */
      osocketstream (const address_type &a): STD_ ostream(0), buffer_() {
          init(&buffer_);
          open(a);
      }


      /**
       * \brief Construct output socket stream object that uses
       * low-level socket \a s for underlying socket output operations
       */
      osocketstream (const socket_type &s): STD_ ostream(0), buffer_(s) {
          init(&buffer_);
      }


      /**
       * \brief Construct output socket stream object from low-level
       * socket accepted by \a server object.
       */
      osocketstream (listener_type &server)
          : STD_ ostream(0), buffer_(server.accept()) {
          init(&buffer_);
      }


      /** \brief Destruct object and close underlying stream device */
      virtual ~osocketstream () {}


      /** \brief Return pointer to underlying socket device */
      socketbuf_type *rdbuf () const {
          return const_cast<socketbuf_type *>(&buffer_);
      }


      /** \brief Return pointer to underlying socket device */
      socketbuf_type *operator -> () const {
          return rdbuf();
      }


      /** \brief Check if connection is established */
      bool is_open () {
          return buffer_.is_open();
      }


      /** \brief Open connection to address \a a */
      void open (const address_type &a) {
          if (!buffer_.open(a))
              setstate(failbit);
      }


      /** \brief Close connection */
      void close () {
          if (!buffer_.close())
              setstate(failbit);
      }


  private:

      socketbuf_type buffer_;
};


/**
 * \brief Input/output stream class.
 *
 * Provides functionality to perform I/O operations with a socket stream.
 */
template <typename family>
class COOL_API socketstream: public STD_ iostream
{
  public:

      /** \brief Family socket buffering device type */
      typedef socketbuf<family> socketbuf_type;

      /** \brief Family address type */
      typedef typename socketbuf_type::address_type address_type;

      /** \brief Family socket type */
      typedef typename socketbuf_type::socket_type socket_type;

      /** \brief Family listener type */
      typedef typename socketbuf_type::listener_type listener_type;


      /** \brief Create new socket stream */
      socketstream (): STD_ iostream(0), buffer_() {
          init(&buffer_);
      }


      /**
       * \brief Construct input/output socket stream object that automatically
       * opens connection to \a.
       */
      socketstream (const address_type &a): STD_ iostream(0), buffer_() {
          init(&buffer_);
          open(a);
      }


      /**
       * \brief Construct input/output socket stream object that uses
       * low-level socket \a s for underlying socket I/O operations
       */
      socketstream (const socket_type &s): STD_ iostream(0), buffer_(s) {
          init(&buffer_);
      }


      /**
       * \brief Construct input/output socket stream object from low-level
       * socket accepted by \a server object.
       */
      socketstream (listener_type &server)
          : STD_ iostream(0), buffer_(server.accept()) {
          init(&buffer_);
      }


      /** \brief Destruct object and close underlying stream device */
      virtual ~socketstream () {}


      /** \brief Return pointer to underlying socket device */
      socketbuf_type *rdbuf () const {
          return const_cast<socketbuf_type *>(&buffer_);
      }


      /** \brief Return pointer to underlying socket device */
      socketbuf_type *operator -> () const {
          return rdbuf();
      }


      /** \brief Check if connection is established */
      bool is_open () {
          return buffer_.is_open();
      }


      /** \brief Open connection to address \a a */
      void open (const address_type &a) {
          if (!buffer_.open(a))
              setstate(failbit);
      }


      /** \brief Close connection */
      void close () {
          if (!buffer_.close())
              setstate(failbit);
      }


  private:

      socketbuf_type buffer_;
};


/**
 * \brief Store '\\r\\n' sequence into stream \a os and flush content of
 * buffer into underlying socket device.
 */
inline STD_ ostream &
crlf (STD_ ostream &os)
{
  os << "\r\n" << STD_ flush;
  return os;
}


/** \} */
}   // namespace cool


#endif  /* cool_net_socketstream_hpp_included */
