/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_net_url_hpp_included
#define cool_net_url_hpp_included

/**
 * \file url.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::url
 *
 * In module \ref cool_net.
 */

#include <cool/cool.hpp>

#ifdef ostream_header
#  include ostream_header
#else
#  include iostream_header
#endif


namespace cool {
/** \addtogroup cool_net */
/** \{ */


/**
 * \brief URL parser and holder
 *
 * Parses (absolute and relative) URLs.
 *
 * See RFC1808: "Relative Uniform Resource Locators",
 * by R. Fielding, UC Irvine, June 1995
 */
class COOL_API url
{
  public:


      /**
       * \ Parse URL \a u and create new object using extracted values
       *
       * URL is parsed into 8 components:
       * (scheme)://(login)\(host):(port)/(path);(params)?(query)#(ref)
       *
       * \note % escapes aren't expanded.
       */
      url (const char *u = 0): url_(0), end(0) {
          operator=(u);
      }


      /** \brief Create copy of URL \a u */
      url (const url &u): url_(0), end(0) {
          operator=(u);
      }


      /** \brief Destroy URL object and release internal resources */
      ~url () {
          alloc(0);
      }


      /** \brief Set new URL using parsed values from \a u */
      url &operator = (const char *u);


      /** \brief Set new URL using previously parser URL \a u */
      url &operator = (const url &u);


      /**
       * \brief Return \e scheme part of URL.
       *
       * If this part wasn't specified, then \c NULL is returned
       */
      const char *scheme () const {
          return scheme_;
      }
 

      /**
       * \brief Return \e login part of URL.
       * \copydoc scheme()
       */
      const char *login () const {
          return login_;
      }
 

      /**
       * \brief Return \e host part of URL.
       * \copydoc scheme()
       */
      const char *host () const {
          return host_;
      }
 

      /**
       * \brief Return \e port part of URL.
       *
       * If \e port wasn't specified then -1 is returned.
       * \see default_port()
       */
      int port () const {
          return port_;
      }
 

      /**
       * \brief Return \e scheme was specified and it is some standard service
       * name, \e default_port is set to service's port value i.e. http = 80
       * etc.
       *
       * If \e scheme isn't standard service name, it's value is set to -1.
       */
      int default_port () const {
          return default_port_;
      }
 

      /**
       * \brief Return \e path part of URL.
       * \copydoc scheme()
       */
      const char *path () const {
          return path_;
      }
 

      /**
       * \brief Return \e params part of URL.
       * \copydoc scheme()
       */
      const char *params () const {
          return params_;
      }
 

      /**
       * \brief Return \e query part of URL.
       * \copydoc scheme()
       */
      const char *query () const {
          return query_;
      }
 

      /**
       * \brief Return \e ref part of URL.
       * \copydoc scheme()
       */
      const char *ref () const {
          return ref_;
      }


      /** \brief Resolve relative URL (\c this) to absolute \a base URL */
      url &resolve (const url &base);


      /**
       * \brief Return \c this URL as string.
       *
       * Returned string is guaranteed to be same as parsed URL string.
       */
      char *c_str (char *buf, int max);


  private:

      char *url_;
      char *end;

      const char *scheme_;
      const char *login_;
      const char *host_;
      const char *path_;
      const char *params_;
      const char *query_;
      const char *ref_;
      int port_;
      int default_port_;

      void alloc (int size);
      void join (bool free_path);
      friend STD_ ostream &operator << (STD_ ostream &os, url &u);
};


/** \brief Write URL \a u as string into stream \a os */
inline STD_ ostream &
operator << (STD_ ostream &os, url &u)
{
  // XXX 512 bytes should be enough for now.
  // If not, think something else out
  char buf[512];
  os << u.c_str(buf, sizeof(buf));

  return os;
}


/** \} */
}   // namespace cool


#endif  /* cool_net_url_hpp_included */
