/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_process_bits_process_hpp_included
#define cool_process_bits_process_hpp_included

#if defined(HAVE_UNISTD_H)
#  include <unistd.h>
#endif
#if defined(HAVE_SYS_TYPES_H)
#  include <sys/types.h>
#endif
#if defined(HAVE_SYS_WAIT_H)
#  include <sys/wait.h>
#endif

#include <cool/exception/os_error.hpp>


namespace cool {
namespace bits {
namespace process {


typedef pid_t pid_type;
typedef int status_type;


int spawn (const char *exe, const char **argv);
void wait_error (const char *func);


inline pid_type
getpid ()
{
  return ::getpid();
}


inline pid_type
getppid ()
{
  return ::getppid();
}


inline pid_type
fork ()
{
  pid_type r = ::fork();
  if (r == -1)
      raise<resource_error>(COOL_LOCUS);
  return r;
}


inline void
check_status (int s, status_type *status_value)
{
  if (WIFSIGNALED(s))
      raise<interrupted>(COOL_LOCUS);
  if (WIFSTOPPED(s))
      raise<stopped>(COOL_LOCUS);
  if (status_value)
      *status_value = WEXITSTATUS(s);
}


inline pid_type
wait_any (status_type *status)
{
  int s = 0, r;
  if ((r = ::wait(&s)) == -1)
      wait_error("wait");
  check_status(s, status);
  return r;
}


inline pid_type
wait_pid (pid_type pid, status_type *status, int option = WUNTRACED)
{
  int s = 0, r;
  if ((r = ::waitpid(pid, &s, option)) == -1)
      wait_error("waitpid");
  check_status(s, status);
  return r;
}


inline bool
finished (pid_type pid, status_type *status)
{
  return wait_pid(pid, status, (WNOHANG | WUNTRACED)) != 0;
}


inline void
detach ()
{
  if (fork() == 0) {
      ::setsid();
      if (fork() == 0)
          return;
  }

  ::exit(0);
}


inline void
setenv (const char *key, const char *value, bool overwrite)
{
#if defined(HAVE_SETENV)
  ::setenv(key, value, overwrite);
#else
  char buf[256];
  snprintf(buf, sizeof(buf), "%s=%s", key, value);
  if (!overwrite && ::getenv(key))
      return;
  ::putenv(buf);
#endif
}


inline const char *
getenv (const char *key)
{
  return ::getenv(key);
}


}   // namespace process
}   // namespace bits
}   // namespace cool


#endif  /* cool_process_bits_process_hpp_included */
