/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_process_process_hpp_included
#define cool_process_process_hpp_included

/**
 * \file process.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::process
 *
 * In module \ref cool_process.
 */

#include <cool/cool.hpp>
#include "bits/process.hpp"


namespace cool {
/** \addtogroup cool_process */
/** \{ */


/**
 * \brief Process handling
 *
 * Non-static methods are related to current process that modify or use
 * current process's id. Static methods can be used to modify or use other
 * processes.
 */
class COOL_API process
{
  public:

      /** \brief Process id type */
      typedef bits::process::pid_type id_type;


      /** \brief Process exiting status */
      typedef bits::process::status_type status_type;


      /**
       * \brief Create new process object and initialize it's id to current
       * process's id
       */
      process (): id_(bits::process::getpid()) { }


      /** \brief Return current process's id */
      id_type id () const {
          return id_;
      }


      /**
       * \brief Create child process.
       *
       * Child process differs from the parent only in it's id and parent id.
       * File locks and pending signals are not inherited.
       *
       * On success, id of child process is returned in parent's thread of
       * execution, and 0 in child's thread.
       *
       * If child process can't be forked, cool::resource_error is thrown.
       */
      id_type fork () {
          id_type r = bits::process::fork();
          if (!r)
              id_ = bits::process::getpid();
          return r;
      }


      /**
       * \brief Detach process from current parent i.e. become a daemon.
       *
       * Standard I/O streams aren't closed or redirected -- it is caller's
       * duty.
       */
      void detach () {
          if (parent() == 1)
              // already detached, our parent is grandfather of all processes
              // -- init
              return;

          bits::process::detach();
          id_ = bits::process::getpid();
      }


      /** \brief Return parent's id */
      static id_type parent () {
          return bits::process::getppid();
      }


      /**
       * \brief Spawn external application and wait for it's exit code.
       *
       * Returns exit code from application. By convention, first element of
       * \a argv must be same as \a exe. If not, assertion is thrown. As
       * exception, argv can be not specified at all, in which case this is
       * done internally.
       */
      static status_type spawn (const char *exe, const char **argv = 0) {
          return bits::process::spawn(exe, argv);
      }


      /**
       * \brief Suspend execution of current process until one of child has
       * exited.
       *
       * If a child has already exited by the time of call (a so-called
       * 'zombie'), the function returns immediately.
       */
      static id_type wait_any (status_type *status = 0) {
          return bits::process::wait_any(status);
      }


      /**
       * \brief Suspend execution of current process until a child as
       * specified by \a pid has exited.
       *
       * \copydoc wait_any()
       */
      static void wait (id_type pid, status_type *status = 0) {
          bits::process::wait_pid(pid, status);
      }


      /**
       * \brief Same as wait() except returns immediately.
       *
       * If this method returns true, children has already exited and false
       * otherwise.
       */
      static bool finished (id_type pid, status_type *status = 0) {
          return bits::process::finished(pid, status);
      }


      /** \brief Set system environment variable \a key to \a value. */
      static void setenv (const char *key, const char *value,
              bool overwrite = true) {
          bits::process::setenv(key, value, overwrite);
      }


      /** \brief Get system enviroment variable \a key value */
      static const char *getenv (const char *key) {
          return bits::process::getenv(key);
      }


  private:

      id_type id_;
};


/** \} */
}   // namespace cool


#endif  /* cool_process_process_hpp_included */
