/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_signal_signal_hpp_included
#define cool_signal_signal_hpp_included

/**
 * \file signal.hpp
 * \author Sven Suursoho
 *
 * \brief Signal handling.
 *
 * All signal handlers must be derived from signal::handler and must override
 * signal::handler::operator()(signal_type) that is called whenever signal is
 * sent to application.
 *
 * Whenever signal::handler is instantiated, it's constructor will insert
 * itself into internal list of handlers for specific signal. When signal
 * is arrived, all associated handlers for this signal are invoked.
 *
 * signal::handler destructor will remove itself from list of handlers.
 *
 * \note Application shouldn't mix using signal class with other ways to
 * handler signalling.
 *
 * In module \ref cool_signal.
 */

#include <cool/cool.hpp>
#include "bits/signal.hpp"


namespace cool {
/** \addtogroup cool_signal */
/** \{ */


class signal
{
  public:

      /** \brief Enumerates signal types */
      enum signal_type
      {
          hup = SIGHUP      /**< Hangup detected on controlling terminal */
          , intr = SIGINT   /**< Interrupt from keyboard */
          , quit = SIGQUIT  /**< Quit from keyboard */
          , ill = SIGILL    /**< Illegal instruction */
          , abrt = SIGABRT  /**< Abort signal from abort() */
          , fpe = SIGFPE    /**< Floating point exception */
          , segv = SIGSEGV  /**< Invalid memory reference */
          , pipe = SIGPIPE  /**< Broken pipe */
          , alrm = SIGALRM  /**< Timer signal from alarm() */
          , term = SIGTERM  /**< Termination signal */
          , usr1 = SIGUSR1  /**< User defined signal 1 */
          , usr2 = SIGUSR2  /**< User defined signal 2 */
          , chld = SIGCHLD  /**< Child stopped or terminated */
          , cont = SIGCONT  /**< Continue if stopped */
          , tstp = SIGTSTP  /**< Stop typed at tty */
          , ttin = SIGTTIN  /**< tty input for background process */
          , ttou = SIGTTOU  /**< tty output for background process */
      };


      /**
       * \brief All signal handlers must be derived from this abstract base
       * handler class.
       */
      class handler
      {
        public:

            /**
             * \brief Construct new handler and add itself into handlers list
             * for signal \a s.
             */
            handler (signal_type s): signo_(s) {
                set(signo_, this);
            }


            /** \brief Remove itself from signal handlers list */
            virtual ~handler () {
                unset(signo_, this);
            }


            /** \brief This method is invoked whenever signal \a s arrives */
            virtual void operator() (signal_type s) = 0;


        private:

            signal_type signo_;
      };


      /**
       * \brief Add signal handler \a h for signal \a s.
       *
       * \return true if adding succeeded and false if handler \a h already
       * exists for signal \a s.
       */
      static bool set (signal_type s, handler *h);


      /**
       * \brief Remove signal handler \a h from \a s handlers list.
       *
       * If last of our installed signal handlers is removed, then default
       * handler is restored.
       *
       * \return true if removal succeeded and false if handler \a h not
       * registered.
       */
      static bool unset (signal_type s, handler *h);


      /**
       * \brief Remove all signal handlers from \a s handlers list and
       * restore default signal handler.
       */
      static void unset (signal_type s);


      /**
       * \brief Mask signal \a s ie prevent application to get signal \a s.
       *
       * Signal is delivered after unmask(signal_type) is called.
       *
       * \note Since each thread maintains its own blocked signal mask, this
       * method modifies only the calling thread's blocked signal mask.
       * \note Not all signals can be blocked
       */
      static void mask (signal_type s);


      /**
       * \brief Unmask signal \a s ie application will see signal \a s again.
       *
       * \note All signals that have delivered to application but were masked,
       * will be delivered after calling this function.
       */
      static void unmask (signal_type s);


      /**
       * \brief Block signal \a b during handling signal \a a.
       *
       * \note Application will see \a b after \a a handler is finished.
       */
      static void mask (signal_type a, signal_type b);


      /**
       * \brief Unmask signal \a b when inside \a a handler.
       *
       * \note Signal \a b will be delivered after handler for \a a is finished.
       */
      static void unmask (signal_type a, signal_type b);


      /** \brief Return true if signal \a s is waiting to be delivered */
      static bool is_pending (signal_type s);


      /** \brief Return set of signals that are waiting to be delivered */
      static sigset_t pending ();


      /** \brief Send signal \a s to current process */
      static void kill (signal_type s);
};


/** \} */
}   // namespace cool


#endif  /* cool_signal_signal_hpp_included */
