/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_bits_lock_hpp_included
#define cool_thread_bits_lock_hpp_included


#include <cool/thread/exception.hpp>


namespace cool {

class condition;

namespace bits {
namespace thread {


template <typename mutex_impl>
class lock_ops: private noncopyable
{
  public:

      typedef typename mutex_impl::cv_state lock_state;


      static void lock (mutex_impl &m) {
          m.do_lock();
      }


      static bool try_lock (mutex_impl &m) {
          return m.do_try_lock();
      }


      static bool timed_lock (mutex_impl &m, const cool::time &t) {
          return m.do_timed_lock(t);
      }


      static void unlock (mutex_impl &m) {
          m.do_unlock();
      }


      static void lock (mutex_impl &m, lock_state &state) {
          m.do_lock(state);
      }


      static void unlock (mutex_impl &m, lock_state &state) {
          m.do_unlock(state);
      }


  private:

      lock_ops () { }
};


template <typename mutex_impl>
class scoped_lock: private noncopyable
{
  public:

      typedef mutex_impl mutex_type;


      explicit
      scoped_lock (mutex_impl &mutex, bool do_lock = true)
          : mutex_(mutex), locked_(false)
      {
          if (do_lock)
              lock();
      }


      ~scoped_lock () {
          if (locked_)
              unlock();
      }


      void lock () {
          if (locked_)
              raise<lock_error>(COOL_LOCUS);
          lock_ops<mutex_impl>::lock(mutex_);
          locked_ = true;
      }


      void unlock () {
          if (!locked_)
              raise<lock_error>(COOL_LOCUS);
          lock_ops<mutex_impl>::unlock(mutex_);
          locked_ = false;
      }


      bool locked () const {
          return locked_;
      }


      operator const void * () const {
          return locked_ ? this : 0;
      }


  private:

      friend class cool::condition;

      mutex_impl &mutex_;
      bool locked_;
};


template <typename try_mutex_impl>
class scoped_try_lock: private noncopyable
{
  public:

      typedef try_mutex_impl mutex_type;


      explicit
      scoped_try_lock (try_mutex_impl &mutex): mutex_(mutex), locked_(false) {
          try_lock();
      }


      scoped_try_lock (try_mutex_impl &mutex, bool do_lock)
          : mutex_(mutex), locked_(false)
      {
          if (do_lock)
              lock();
      }


      ~scoped_try_lock () {
          if (locked_)
              unlock();
      }


      void lock () {
          if (locked_)
              raise<lock_error>(COOL_LOCUS);
          lock_ops<try_mutex_impl>::lock(mutex_);
          locked_ = true;
      }


      bool try_lock () {
          if (locked_)
              raise<lock_error>(COOL_LOCUS);
          return (locked_ = lock_ops<try_mutex_impl>::try_lock(mutex_));
      }


      void unlock () {
          if (!locked_)
              raise<lock_error>(COOL_LOCUS);
          lock_ops<try_mutex_impl>::unlock(mutex_);
          locked_ = false;
      }


      bool locked () const {
          return locked_;
      }


      operator const void * () const {
          return locked_ ? this : 0;
      }


  private:

      friend class cool::condition;

      try_mutex_impl &mutex_;
      bool locked_;
};


template <typename timed_mutex_impl>
class scoped_timed_lock: private noncopyable
{
  public:

      typedef timed_mutex_impl mutex_type;


      scoped_timed_lock (timed_mutex_impl &mutex, const cool::time &t)
          : mutex_(mutex), locked_(false)
      {
          timed_lock(t);
      }


      scoped_timed_lock (timed_mutex_impl &mutex, bool do_lock)
          : mutex_(mutex), locked_(false)
      {
          if (do_lock)
              lock();
      }


      ~scoped_timed_lock () {
          if (locked_)
              unlock();
      }


      void lock () {
          if (locked_)
              raise<lock_error>(COOL_LOCUS);
          lock_ops<timed_mutex_impl>::lock(mutex_);
          locked_ = true;
      }


      bool try_lock () {
          if (locked_)
              raise<lock_error>(COOL_LOCUS);
          return (locked_ = lock_ops<timed_mutex_impl>::try_lock(mutex_));
      }


      bool timed_lock (const cool::time &t) {
          if (locked_)
              raise<lock_error>(COOL_LOCUS);
          return (locked_ = lock_ops<timed_mutex_impl>::timed_lock(mutex_, t));
      }


      void unlock () {
          if (!locked_)
              raise<lock_error>(COOL_LOCUS);
          lock_ops<timed_mutex_impl>::unlock(mutex_);
          locked_ = false;
      }


      bool locked () const {
          return locked_;
      }


      operator const void * () const {
          return locked_ ? this : 0;
      }


  private:

      friend class cool::condition;

      timed_mutex_impl &mutex_;
      bool locked_;
};


}   // namespace thread
}   // namespace bits
}   // namespace cool


#endif  /* cool_thread_bits_lock_hpp_included */
