/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_bits_mutex_hpp_included
#define cool_thread_bits_mutex_hpp_included

#include <cool/thread/exception.hpp>
#if defined(HAVE_PTHREAD_H)
#  include <pthread.h>
#endif


namespace cool {
namespace bits {
namespace thread {


typedef ::pthread_mutex_t mutex_type;
struct mutex_cv_state
{
  ::pthread_mutex_t *pmutex;
};


typedef ::pthread_mutex_t try_mutex_type;
struct try_mutex_cv_state
{
  ::pthread_mutex_t *pmutex;
};


#if defined(HAVE_PTHREAD_MUTEX_TIMEDLOCK)
// huh, what a relief, we don't have to cook our own timed_lock :)
typedef ::pthread_mutex_t timed_mutex_type;
#else
// no escape, we have to create timed_lock using condition variable :(
struct timed_mutex_type
{
  ::pthread_mutex_t mutex_;
  ::pthread_cond_t cond_;
  bool locked_;
};
#endif
struct timed_mutex_cv_state
{
  ::pthread_mutex_t *pmutex;
};


template <typename mutex_impl>
inline void
mutex_init (mutex_impl &mutex)
{
  if (::pthread_mutex_init(&mutex, 0) != 0)
      raise<resource_error>(COOL_LOCUS);
}


template <typename mutex_impl>
inline void
mutex_destroy (mutex_impl &mutex)
{
  int r = 0;
  r = ::pthread_mutex_destroy(&mutex);
  assert(r == 0);
}


template <typename mutex_impl>
inline void
mutex_lock (mutex_impl &mutex)
{
  int r = ::pthread_mutex_lock(&mutex);
  if (r == EDEADLK)
      raise<lock_error>(COOL_LOCUS);
  assert(r == 0);
}


template <typename mutex_impl>
inline bool
mutex_try_lock (mutex_impl &mutex)
{
  int r = ::pthread_mutex_trylock(&mutex);
  if (r == EDEADLK)
      raise<lock_error>(COOL_LOCUS);
  assert(r == 0 || r == EBUSY);
  return r == 0;
}


template <typename mutex_impl>
inline bool
mutex_timed_lock (mutex_impl &mutex, const cool::time &t)
{
#if defined(HAVE_PTHREAD_MUTEX_TIMEDLOCK)
// use timedlock provided by pthread
  struct timespec ts = { t.sec, t.nsec };
  int r = ::pthread_mutex_timedlock(&mutex, &ts);
  if (r == EDEADLK)
      raise<lock_error>(COOL_LOCUS);
  assert(r == 0 || r == ETIMEDOUT);
  return r == 0;
#else
  // actually implementation without support by pthread is below
  // !defined(HAVE_PTHREAD_MUTEX_TIMEDLOCK)
  raise<not_supported_error>(COOL_LOCUS);
  return true;
#endif
}


template <typename mutex_impl>
inline void
mutex_unlock (mutex_impl &mutex)
{
  int r = ::pthread_mutex_unlock(&mutex);
  if (r == EPERM)
      raise<lock_error>(COOL_LOCUS);
  assert(r == 0);
}


template <typename mutex_impl, typename mutex_cv_state_impl>
inline void
mutex_lock (mutex_impl &mutex, mutex_cv_state_impl &state)
{
}


template <typename mutex_impl, typename mutex_cv_state_impl>
inline void
mutex_unlock (mutex_impl &mutex, mutex_cv_state_impl &state)
{
  state.pmutex = &mutex;
}


#if !defined(HAVE_PTHREAD_MUTEX_TIMEDLOCK)
// create our own timedlock. Unfortunately it means that we have to overwrite
// all other locking functions as well :(

template <>
inline void
mutex_init (timed_mutex_type &mutex)
{
  mutex.locked_ = false;

  if (::pthread_mutex_init(&mutex.mutex_, 0) != 0)
      raise<resource_error>(COOL_LOCUS);

  if (::pthread_cond_init(&mutex.cond_, 0) != 0) {
      ::pthread_mutex_destroy(&mutex.mutex_);
      raise<resource_error>(COOL_LOCUS);
  }
}


template <>
inline void
mutex_destroy (timed_mutex_type &mutex)
{
  assert(mutex.locked_ == false);

  int r = ::pthread_mutex_destroy(&mutex.mutex_);
  assert(r == 0);

  r = ::pthread_cond_destroy(&mutex.cond_);
  assert(r == 0);
}


template <>
inline void
mutex_lock (timed_mutex_type &mutex)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  while (mutex.locked_) {
      r = ::pthread_cond_wait(&mutex.cond_, &mutex.mutex_);
      assert(r == 0);
  }

  assert(mutex.locked_ == false);
  mutex.locked_ = true;

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);
}


template <>
inline bool
mutex_try_lock (timed_mutex_type &mutex)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  bool ret = false;
  if (!mutex.locked_)
      ret = mutex.locked_ = true;

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);

  return ret;
}


template <>
inline bool
mutex_timed_lock (timed_mutex_type &mutex, const cool::time &t)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  struct timespec ts = { t.sec, t.nsec };
  while (mutex.locked_) {
      r = ::pthread_cond_timedwait(&mutex.cond_, &mutex.mutex_, &ts);
      if (r == ETIMEDOUT)
          break;

      assert(r == 0);
  }

  bool ret = false;
  if (!mutex.locked_)
      ret = mutex.locked_ = true;

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);

  return ret;
}


template <>
inline void
mutex_unlock (timed_mutex_type &mutex)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  assert(mutex.locked_ == true);
  mutex.locked_ = false;

  r = ::pthread_cond_signal(&mutex.cond_);
  assert(r == 0);

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);
}


template <>
inline void
mutex_lock (timed_mutex_type &mutex, timed_mutex_cv_state &state)
{
  int r;
  while (mutex.locked_) {
      r = ::pthread_cond_wait(&mutex.cond_, &mutex.mutex_);
      assert(r == 0);
  }

  assert(mutex.locked_ == false);
  mutex.locked_ = true;

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);
}


template <>
inline void
mutex_unlock (timed_mutex_type &mutex, timed_mutex_cv_state &state)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  assert(mutex.locked_ == true);
  mutex.locked_ = false;

  r = ::pthread_cond_signal(&mutex.cond_);
  assert(r == 0);

  state.pmutex = &mutex.mutex_;
}

#endif


}   // namespace thread
}   // namespace bits
}   // namespace cool


#endif  /* cool_thread_bits_mutex_hpp_included */
