/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_bits_recursive_mutex_hpp_included
#define cool_thread_bits_recursive_mutex_hpp_included

#include <cool/thread/exception.hpp>
#if defined(HAVE_PTHREAD_H)
#  include <pthread.h>
#endif

#if defined(HAVE_MUTEX_STANDARD)
#  define COOL_RECURSIVE_MUTEX PTHREAD_MUTEX_RECURSIVE
#elif defined(HAVE_MUTEX_NP)
#  define COOL_RECURSIVE_MUTEX PTHREAD_MUTEX_RECURSIVE_NP
#else
#  error No recursive mutexes
#endif


namespace cool {
namespace bits {
namespace thread {


struct recursive_mutex_type
{
  ::pthread_mutex_t mutex_;
  unsigned count_;
};


struct recursive_mutex_cv_state
{
  long count;
  ::pthread_mutex_t *pmutex;
};


struct recursive_try_mutex_type
{
  ::pthread_mutex_t mutex_;
  unsigned count_;
};


struct recursive_try_mutex_cv_state
{
  long count;
  ::pthread_mutex_t *pmutex;
};


struct recursive_timed_mutex_type
{
  ::pthread_mutex_t mutex_;
  unsigned count_;

#if !defined(HAVE_PTHREAD_MUTEX_TIMEDLOCK)
  ::pthread_cond_t unlocked_;
  ::pthread_t thread_;
  bool valid_;
#endif
};


struct recursive_timed_mutex_cv_state
{
  long count;
  ::pthread_mutex_t *pmutex;
};


template <typename recursive_mutex_impl>
inline void
recursive_mutex_init (recursive_mutex_impl &mutex)
{
  mutex.count_ = 0;

  ::pthread_mutexattr_t attr;
  int r = ::pthread_mutexattr_init(&attr);
  assert(r == 0);

  r = SS_MUTEXATTR_SETTYPE(&attr, COOL_RECURSIVE_MUTEX);
  assert(r == 0);

  r = ::pthread_mutex_init(&mutex.mutex_, &attr);
  {
      int r = ::pthread_mutexattr_destroy(&attr);
      assert(r == 0);
  }
  if (r != 0)
      raise<resource_error>(COOL_LOCUS);
}


template <typename recursive_mutex_impl>
inline void
recursive_mutex_destroy (recursive_mutex_impl &mutex)
{
  int r = ::pthread_mutex_destroy(&mutex.mutex_);
  assert(r == 0);
}


template <typename recursive_mutex_impl>
inline void
recursive_mutex_lock (recursive_mutex_impl &mutex)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  if (++mutex.count_ > 1) {
      r = ::pthread_mutex_unlock(&mutex.mutex_);
      assert(r == 0);
  }
}


template <typename recursive_mutex_impl>
inline bool
recursive_mutex_try_lock (recursive_mutex_impl &mutex)
{
  int r = ::pthread_mutex_trylock(&mutex.mutex_);
  assert(r == 0 || r == EBUSY);

  if (r == EBUSY)
      return false;

  if (++mutex.count_ > 1) {
      r = ::pthread_mutex_unlock(&mutex.mutex_);
      assert(r == 0);
  }

  return true;
}


template <typename recursive_mutex_impl>
inline bool
recursive_mutex_timed_lock (recursive_mutex_impl &mutex, const cool::time &t)
{
#if defined(HAVE_PTHREAD_MUTEX_TIMEDLOCK)

  // use timedlock provided by pthread
  struct timespec ts = { t.sec, t.nsec };
  int r = ::pthread_mutex_timedlock(&mutex.mutex_, &ts);
  assert(r == 0 || r == ETIMEDOUT);

  if (r == ETIMEDOUT)
      return false;

  if (++mutex.count_ > 1) {
      r = ::pthread_mutex_unlock(&mutex.mutex_);
      assert(r == 0);
  }

  return true;

#else

  // actually implementation without support by pthread is below
  // !defined(HAVE_PTHREAD_MUTEX_TIMEDLOCK)
  raise<not_supported_error>(COOL_LOCUS);
  return true;

#endif
}


template <typename recursive_mutex_impl>
inline void
recursive_mutex_unlock (recursive_mutex_impl &mutex)
{
  if (--mutex.count_ == 0) {
      int r = ::pthread_mutex_unlock(&mutex.mutex_);
      assert(r == 0);
  }
}


template <typename recursive_mutex_impl, typename recursive_mutex_cv_state_impl>
inline void
recursive_mutex_lock (recursive_mutex_impl &mutex,
        recursive_mutex_cv_state_impl &state)
{
  mutex.count_ = state.count;
}


template <typename recursive_mutex_impl, typename recursive_mutex_cv_state_impl>
inline void
recursive_mutex_unlock (recursive_mutex_impl &mutex,
        recursive_mutex_cv_state_impl &state)
{
  state.pmutex = &mutex.mutex_;
  state.count = mutex.count_;
  mutex.count_ = 0;
}


#if !defined(HAVE_PTHREAD_MUTEX_TIMEDLOCK)
// create our own timedlock. Unfortunately it means that we have to overwrite
// all other locking functions as well :(

template <>
inline void
recursive_mutex_init (recursive_timed_mutex_type &mutex)
{
  mutex.valid_ = false;
  mutex.count_ = 0;

  if (::pthread_mutex_init(&mutex.mutex_, 0) != 0)
      raise<resource_error>(COOL_LOCUS);

  if (::pthread_cond_init(&mutex.unlocked_, 0) != 0) {
      ::pthread_mutex_destroy(&mutex.mutex_);
      raise<resource_error>(COOL_LOCUS);
  }
}


template <>
inline void
recursive_mutex_destroy (recursive_timed_mutex_type &mutex)
{
  int r = ::pthread_mutex_destroy(&mutex.mutex_);
  assert(r == 0);

  r = ::pthread_cond_destroy(&mutex.unlocked_);
  assert(r == 0);
}


template <>
inline void
recursive_mutex_lock (recursive_timed_mutex_type &mutex)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  ::pthread_t thread = ::pthread_self();
  if (mutex.valid_ && ::pthread_equal(mutex.thread_, thread))
      ++mutex.count_;
  else {
      while (mutex.valid_) {
          r = ::pthread_cond_wait(&mutex.unlocked_, &mutex.mutex_);
          assert(r == 0);
      }

      mutex.thread_ = thread;
      mutex.valid_ = true;
      mutex.count_ = 1;
  }

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);
}


template <>
inline bool
recursive_mutex_try_lock (recursive_timed_mutex_type &mutex)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  bool ret = false;
  ::pthread_t thread = ::pthread_self();
  if (mutex.valid_ && ::pthread_equal(mutex.thread_, thread)) {
      ++mutex.count_;
      ret = true;
  }
  else if (!mutex.valid_) {
      mutex.thread_ = thread;
      mutex.valid_ = true;
      mutex.count_ = 1;
      ret = true;
  }

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);

  return ret;
}


template <>
inline bool
recursive_mutex_timed_lock (recursive_timed_mutex_type &mutex,
        const cool::time &t)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  bool ret = false;
  ::pthread_t thread = ::pthread_self();
  if (mutex.valid_ && ::pthread_equal(mutex.thread_, thread)) {
      ++mutex.count_;
      ret = true;
  }
  else {
      struct timespec ts = { t.sec, t.nsec };

      while (mutex.valid_) {
          r = ::pthread_cond_timedwait(&mutex.unlocked_, &mutex.mutex_, &ts);
          if (r == ETIMEDOUT)
              break;
          assert(r == 0);
      }

      if (!mutex.valid_) {
          mutex.thread_ = thread;
          mutex.valid_ = true;
          mutex.count_ = 1;
          ret = true;
      }
  }

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);

  return ret;
}


template <>
inline void
recursive_mutex_unlock (recursive_timed_mutex_type &mutex)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  ::pthread_t thread = ::pthread_self();
  if (mutex.valid_ && !::pthread_equal(mutex.thread_, thread)) {
      r = ::pthread_mutex_unlock(&mutex.mutex_);
      assert(r == 0);
      raise<lock_error>(COOL_LOCUS);
  }

  if (--mutex.count_ == 0) {
      assert(mutex.valid_ == true);
      mutex.valid_ = false;

      r = ::pthread_cond_signal(&mutex.unlocked_);
      assert(r == 0);
  }

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);
}


template <>
inline void
recursive_mutex_lock (recursive_timed_mutex_type &mutex,
        recursive_timed_mutex_cv_state &state)
{
  int r;

  while (mutex.valid_) {
      r = ::pthread_cond_wait(&mutex.unlocked_, &mutex.mutex_);
      assert(r == 0);
  }

  mutex.thread_ = ::pthread_self();
  mutex.valid_ = true;
  mutex.count_ = state.count;

  r = ::pthread_mutex_unlock(&mutex.mutex_);
  assert(r == 0);
}


template <>
inline void
recursive_mutex_unlock (recursive_timed_mutex_type &mutex,
        recursive_timed_mutex_cv_state &state)
{
  int r = ::pthread_mutex_lock(&mutex.mutex_);
  assert(r == 0);

  assert(mutex.valid_ == true);
  mutex.valid_ = false;

  r = ::pthread_cond_signal(&mutex.unlocked_);
  assert(r == 0);

  state.pmutex = &mutex.mutex_;
  state.count = mutex.count_;
  mutex.count_ = 0;
}

#endif


}   // namespace thread
}   // namespace bits
}   // namespace cool


#endif  /* cool_thread_bits_recursive_mutex_hpp_included */
