/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_blocking_queue_hpp_included
#define cool_thread_blocking_queue_hpp_included

/**
 * \file blocking_queue.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::blocking_queue
 *
 * In module \ref cool_thread.
 */

#include <cool/cool.hpp>
#include <cool/thread/queue.hpp>
#include <cool/thread/mutex.hpp>
#include <cool/thread/condition.hpp>
#include <cool/exception/exception.hpp>
#include deque_header


namespace cool {
/** \addtogroup cool_thread */
/** \{ */


/**
 * \brief Like a cool::locked_queue, blocking_queue provides serialized access
 * to items in queue.
 *
 * It differs by causing caller accessing pop() to block until a next item
 * becomes available.
 */
template <class T,
         typename mutex_type = mutex,
         typename container_type = STD_ deque<T> >
class COOL_API blocking_queue: public queue<T>, public mutex_type
{
  public:

      /** \brief Type of items that can be inserted into queue */
      typedef T value_type;


      /** \brief Create a new empty queue */
      blocking_queue () { }


      /** \brief Destroy a queue */
      virtual ~blocking_queue () { }


      /** \brief Add an item to queue */
      virtual void push (const value_type &item) {
          typename mutex_type::scoped_lock guard(*this);
          queue_.push_back(item);
          not_empty_.notify_one();
      }


      /**
       * \brief Retrieve and remove item from queue.
       *
       * The caller waits indefinitely until there is new item available to
       * return.
       */
      virtual value_type pop () {
          typename mutex_type::scoped_lock guard(*this);
          while (queue_.size() == 0)
              // wait item to become available
              not_empty_.wait(guard);

          value_type item = queue_.front();
          queue_.pop_front();
          return item;
      }


      /**
       * \brief Retrieve and remove item from queue.
       *
       * The caller waits up to \a timeout until there is new item available to
       * return. If after \a timeout still no item available,
       * cool::timeout_error is thrown.
       */
      virtual value_type pop (const time::time &timeout) {
          typename mutex_type::scoped_lock guard(*this);
          while (queue_.size() == 0) {
              // wait up to timeout item to become available
              if (!not_empty_.wait(guard, timeout))
                  raise<timeout_error>(COOL_LOCUS);
          }

          value_type item = queue_.front();
          queue_.pop_front();
          return item;
      }


      /** \brief Return number of item in queue */
      virtual size_t size () {
          typename mutex_type::scoped_lock guard(*this);
          return queue_.size();
      }


      /** \brief Return true if collection has no item */
      virtual bool empty () {
          return size() == 0;
      }


  private:

      condition not_empty_;
      container_type queue_;
};


/** \} */
}   // namespace cool


#endif  /* cool_thread_blocking_queue_hpp_included */
