/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_bounded_queue_hpp_included
#define cool_thread_bounded_queue_hpp_included

/**
 * \file bounded_queue.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::bounded_queue
 *
 * In module \ref cool_thread.
 */

#include <cool/cool.hpp>
#include <cool/thread/queue.hpp>
#include <cool/thread/mutex.hpp>
#include <cool/thread/condition.hpp>
#include <cool/exception/exception.hpp>
#include deque_header


namespace cool {
/** \addtogroup cool_thread */
/** \{ */


/**
 * \brief Like a cool::locked_queue, bounded_queue provides serialized access
 * to items in queue.
 *
 * Differences from other queue implementations:
 *  - there is maximum capacity
 *  - threads calling empty() will be blocked until queue becomes emtpy
 *  - threads calling pop() will be blocked until there is item in queue to
 *    return
 *  - threads calling push() will be blocked until there is room in queue to
 *    add new item
 */
template <class T,
         typename mutex_type = mutex,
         typename container_type = STD_ deque<T> >
class COOL_API bounded_queue: public queue<T>, public mutex_type
{
  public:

      /** \brief Type of items that can be inserted into queue */
      typedef T value_type;


      /** \brief Create a new empty queue */
      bounded_queue (int capacity): capacity_(capacity) { }


      /** \brief Destroy a queue */
      virtual ~bounded_queue () { }


      /** \brief Return number of items that container is capable to hold */
      size_t capacity () const {
          return capacity_;
      }


      /**
       * \brief Add an item to queue
       *
       * If queue already has maximum number of items, the calling thread will
       * be suspended until at least one item is removed from queue, so that
       * new item can be inserted.
       */
      virtual void push (const value_type &item) {
          typename mutex_type::scoped_lock guard(*this);

          // wait until there is room
          while (queue_.size() == capacity_)
              not_full_.wait(guard);

          // insert item and wake one of waiter
          queue_.push_back(item);
          not_empty_.notify_one();
      }


      /**
       * \brief Add an item to queue
       *
       * If queue already has maximum number of items, the calling thread will
       * be blocked up to \a timeout until at least one item is removed from
       * queue, so that new item can be inserted.
       *
       * Returns true, if insertion succeeded and false otherwise.
       */
      virtual bool push (const value_type &item, const time::time &timeout) {
          typename mutex_type::scoped_lock guard(*this);

          // wait until there is room
          while (queue_.size() == capacity_)
              if (!not_full_.wait(guard, timeout))
                  return false;

          // insert item and wake one of waiter
          queue_.push_back(item);
          not_empty_.notify_one();
          return true;
      }


      /**
       * \brief Retrieve and remove item from queue.
       *
       * The caller waits indefinitely until there is new item available to
       * return.
       */
      virtual value_type pop () {
          typename mutex_type::scoped_lock guard(*this);
          while (queue_.size() == 0)
              // wait item to become available
              not_empty_.wait(guard);

          value_type item = queue_.front();
          queue_.pop_front();

          // wake thread that tries to add and threads waiting for empty()
          not_full_.notify_one();
          if (queue_.size() == 0)
              is_empty_.notify_all();

          return item;
      }


      /**
       * \brief Retrieve and remove item from queue.
       *
       * The caller waits up to \a timeout until there is new item available to
       * return. If after \a timeout still no item available,
       * cool::timeout_error is thrown.
       */
      virtual value_type pop (const time::time &timeout) {
          typename mutex_type::scoped_lock guard(*this);
          while (queue_.size() == 0) {
              // wait up to timeout item to become available
              if (!not_empty_.wait(guard, timeout))
                  raise<timeout_error>(COOL_LOCUS);
          }

          value_type item = queue_.front();
          queue_.pop_front();

          // wake threads that tries to add and threads waiting for empty
          not_full_.notify_one();
          if (queue_.size() == 0)
              is_empty_.notify_all();

          return item;
      }


      /** \brief Return number of item in queue */
      virtual size_t size () {
          typename mutex_type::scoped_lock guard(*this);
          return queue_.size();
      }


      /** \brief Return true if collection has no item */
      virtual bool empty () {
          typename mutex_type::scoped_lock guard(*this);
          while (queue_.size() > 0)
              is_empty_.wait(guard);
          return true;
      }


  private:

      size_t capacity_;
      condition not_full_;
      condition not_empty_;
      condition is_empty_;
      container_type queue_;
};


/** \} */
}   // namespace cool


#endif  /* cool_thread_bounded_queue_hpp_included */
