/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_condition_hpp_included
#define cool_thread_condition_hpp_included

/**
 * \file condition.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::condition
 *
 * In module \ref cool_thread.
 */

#include <cool/cool.hpp>
#include <cool/time/time.hpp>
#include <cool/util/noncopyable.hpp>
#include "bits/condition.hpp"
#include "bits/lock.hpp"


namespace cool {
/** \addtogroup cool_thread */
/** \{ */


/**
 * \brief Condition variable synchronizing primitive.
 *
 * Condition variable is synchronization device that allows signal and
 * schedule threads in single process.
 *
 * The basic operations on condition variables are following:
 *  - signal the condition when some predicate is satisfied
 *  - wait for condition: suspend calling thread until another thread signals
 *    the condition.
 *
 * A condition variable must always be associated with a mutex to avoid the
 * race condition where a thread prepares to wait on a condition variable and
 * another thread signals the condition just before the first thread actually
 * waits on it.
 */
class COOL_API condition: private noncopyable
{
  public:

      /** \brief Initialize condition variable */
      condition (): cond_() { }


      /**
       * \brief Destroy condition variable, freeing resources it might hold.
       *
       * No threads must be waiting on the condition variable on entrance to
       * destructor. If there are waiting thread(s), following behaviour
       * depends on how COOL is configured: if debugging is enabled, then
       * application is aborted with error message. Otherwise error is
       * silently discarded, leaking low level condition variable handle.
       */
      ~condition () { }


      /**
       * \brief Atomically unlock \a lock and wait for condition variable to
       * be signalled.
       *
       * The thread execution is suspended and doesn't consume CPU time until
       * condition variable is signalled. The mutex must be locked by the
       * calling thread. Before returning to the calling thread, mutex is
       * locked again.
       */
      template <typename lock_impl>
      void wait (lock_impl &lock) {
          if (!lock)
              raise<lock_error>(COOL_LOCUS);

          do_wait(lock.mutex_);
      }


      /** \see wait(lock_impl&) */
      template <typename lock_impl, typename pred_func>
      void wait (lock_impl &lock, pred_func pred) {
          if (!lock)
              raise<lock_error>(COOL_LOCUS);

          while (!pred())
              do_wait(lock.mutex_);
      }


      /**
       * \brief Atomically unlock \a lock and wait up to \a t for condition
       * variable to be signalled.
       *
       * If \a t passes before condition variable is signalled, this method
       * returns \c false, also if some signal interrupts waiting.
       *
       * The thread execution is suspended and doesn't consume CPU time until
       * condition variable is signalled. The mutex must be locked by the
       * calling thread. Before returning to the calling thread, mutex is
       * locked again.
       */
      template <typename lock_impl>
      bool wait (lock_impl &lock, const time::time &t) {
          if (!lock)
              raise<lock_error>(COOL_LOCUS);

          return do_wait(lock.mutex_, t);
      }


      /** \see wait(lock_impl&,const time::time &) */
      template <typename lock_impl, typename pred_func>
      bool wait (lock_impl &lock, const time::time &t, pred_func pred) {
          if (!lock)
              raise<lock_error>(COOL_LOCUS);

          while (!pred())
              if (!do_wait(lock.mutex_, t))
                  return false;

          return true;
      }


      /**
       * \brief Signals a single thread (if there is waiting at least one).
       *
       * That signalled thread is scheduled to wake up and acquire the
       * associated mutex. Although it is not required, if predictable
       * scheduling is desired, that mutex must be locked while calling this
       * method.
       *
       * Nothing happens if there is no waiting threads.
       */
      void notify_one () {
          cond_.notify_one();
      }


      /**
       * \brief Signals all threads blocking on the condition variable.
       *
       * Each thread that was signalled is then scheduled to wake up and
       * acquire the associated mutex. This will happen in a serialized
       * manner. The order of wake up is not defined.
       *
       * Nothing happens if there is no waiting threads.
       */
      void notify_all () {
          cond_.notify_all();
      }


  private:

      template <typename mutex_impl>
      void do_wait (mutex_impl &mutex) {
          typedef bits::thread::lock_ops<mutex_impl> lock_ops;
          typename lock_ops::lock_state state;

          lock_ops::unlock(mutex, state);
          cond_.wait(state.pmutex);
          lock_ops::lock(mutex, state);
      }


      template <typename mutex_impl>
      bool do_wait (mutex_impl &mutex, const time::time &t) {
          typedef bits::thread::lock_ops<mutex_impl> lock_ops;
          typename lock_ops::lock_state state;

          lock_ops::unlock(mutex, state);
          bool ret = cond_.wait(state.pmutex, t);
          lock_ops::lock(mutex, state);

          return ret;
      }


      bits::thread::condition cond_;
};


/** \} */
}   // namespace cool


#endif  /* cool_thread_condition_hpp_included */
