/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_locked_queue_hpp_included
#define cool_thread_locked_queue_hpp_included

/**
 * \file locked_queue.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::locked_queue
 *
 * In module \ref cool_thread.
 */

#include <cool/cool.hpp>
#include <cool/thread/queue.hpp>
#include <cool/thread/mutex.hpp>
#include <cool/exception/exception.hpp>
#include deque_header


namespace cool {
/** \addtogroup cool_thread */
/** \{ */


/**
 * \brief Simple cool::queue implementation that provides serialized access to
 * items in it.
 */
template <class T,
         typename mutex_type = mutex,
         typename container_type = STD_ deque<T> >
class COOL_API locked_queue: public queue<T>, public mutex_type
{
  public:

      /** \brief Type of items that can be inserted into queue */
      typedef T value_type;


      /** \brief Create a new empty queue */
      locked_queue () { }


      /** \brief Destroy a queue */
      virtual ~locked_queue () { }


      /** \brief Add an item to queue */
      virtual void push (const value_type &item) {
          typename mutex_type::scoped_lock guard(*this);
          queue_.push_back(item);
      }


      /**
       * \brief Retrieve and remove item from queue.
       *
       * If queue is empty, cool::not_found_error is thrown
       */
      virtual value_type pop () {
          typename mutex_type::scoped_lock guard(*this);
          if (queue_.size() == 0)
              raise<not_found_error>(COOL_LOCUS);

          value_type item = queue_.front();
          queue_.pop_front();
          return item;
      }


      /** \brief Return number of item in queue */
      virtual size_t size () {
          typename mutex_type::scoped_lock guard(*this);
          return queue_.size();
      }


      /** \brief Return true if collection has no item */
      virtual bool empty () {
          return size() == 0;
      }


  private:

      container_type queue_;
};


/** \} */
}   // namespace cool


#endif  /* cool_thread_locked_queue_hpp_included */
