/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_mutex_hpp_included
#define cool_thread_mutex_hpp_included

/**
 * \file mutex.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::mutex, cool::try_mutex, cool::timed_mutex
 *
 * <b>MUT</b>ually <b>EX</b>clusive access devices. A mutex is used to provide
 * serialized access (one thread at a time) to some section of code. This is
 * accomplished by attempting to lock the mutex before entering that protected
 * section and by unlocking mutex when leaving that region. While mutex is
 * locked, no other threads can lock the same mutex.
 *
 * Mutex object itself doesn't provide lock/unlock methods (or any at all).
 * Instead, there are inner classes scoped_lock, scoped_try_lock and
 * scoped_timed_lock that use pattern known as 'scoped locking'. Constructors
 * of these classes lock mutex and unlock in destructor. In this way, it is
 * guaranteed, that there can't forget any mutex locked.
 *
 * Example usage:
 * \code
 * cool::mutex m;   // all mutexes is initially always unlocked
 *
 * {
 *   // m will be locked
 *   cool::mutex::scoped_lock guard(m);
 *
 *   guard.unlock();    // unlock m
 *   ...
 *   guard.lock();      // lock m again
 *   ...
 *
 *   // guard's destructor takes care that m is unlocked when leaving scope
 *   // no matter if user left it locked or not.
 * }
 * \endcode
 *
 * Lock object are not thread-safe because they are not meant to be shared
 * between threads. Only mutexes are shared, but lock object belong to some
 * local scope.
 *
 * In module \ref cool_thread.
 */

#include <cool/cool.hpp>
#include <cool/time/time.hpp>
#include <cool/util/noncopyable.hpp>
#include "bits/mutex.hpp"
#include "bits/lock.hpp"


namespace cool {
/** \addtogroup cool_thread */
/** \{ */


/**
 * \brief Simple mutex.
 * For more information, see \ref mutex.hpp
 */
class COOL_API mutex: private noncopyable
{
  public:

      typedef bits::thread::scoped_lock<mutex> scoped_lock;


      /** \brief Create simple mutex */
      mutex () {
          bits::thread::mutex_init(mutex_);
      }


      /** \brief Destroy mutex */
      ~mutex () {
          bits::thread::mutex_destroy(mutex_);
      }


  private:

      friend class bits::thread::lock_ops<mutex>;

      typedef bits::thread::mutex_cv_state cv_state;


      void do_lock () {
          bits::thread::mutex_lock(mutex_);
      }


      void do_unlock () {
          bits::thread::mutex_unlock(mutex_);
      }


      void do_lock (cv_state &state) {
          bits::thread::mutex_lock(mutex_, state);
      }


      void do_unlock (cv_state &state) {
          bits::thread::mutex_unlock(mutex_, state);
      }


      bits::thread::mutex_type mutex_;
};


/**
 * \brief Simple try mutex.
 * For more information, see \ref mutex.hpp
 */
class COOL_API try_mutex: private noncopyable
{
  public:

      typedef bits::thread::scoped_lock<try_mutex> scoped_lock;
      typedef bits::thread::scoped_try_lock<try_mutex> scoped_try_lock;


      /** \brief Create try mutex */
      try_mutex () {
          bits::thread::mutex_init(mutex_);
      }


      /** \brief Destroy mutex */
      ~try_mutex () {
          bits::thread::mutex_destroy(mutex_);
      }


  private:

      friend class bits::thread::lock_ops<try_mutex>;

      typedef bits::thread::try_mutex_cv_state cv_state;


      void do_lock () {
          bits::thread::mutex_lock(mutex_);
      }


      bool do_try_lock () {
          return bits::thread::mutex_try_lock(mutex_);
      }


      void do_unlock () {
          bits::thread::mutex_unlock(mutex_);
      }


      void do_lock (cv_state &state) {
          bits::thread::mutex_lock(mutex_, state);
      }


      void do_unlock (cv_state &state) {
          bits::thread::mutex_unlock(mutex_, state);
      }


      bits::thread::try_mutex_type mutex_;
};


/**
 * \brief Simple timed mutex.
 * For more information, see \ref mutex.hpp
 */
class COOL_API timed_mutex: private noncopyable
{
  public:

      typedef bits::thread::scoped_lock<timed_mutex> scoped_lock;
      typedef bits::thread::scoped_try_lock<timed_mutex> scoped_try_lock;
      typedef bits::thread::scoped_timed_lock<timed_mutex> scoped_timed_lock;


      /** \brief Create timed mutex */
      timed_mutex () {
          bits::thread::mutex_init(mutex_);
      }


      /** \brief Destroy mutex */
      ~timed_mutex () {
          bits::thread::mutex_destroy(mutex_);
      }


  private:

      friend class bits::thread::lock_ops<timed_mutex>;

      typedef bits::thread::timed_mutex_cv_state cv_state;


      void do_lock () {
          bits::thread::mutex_lock(mutex_);
      }


      bool do_try_lock () {
          return bits::thread::mutex_try_lock(mutex_);
      }


      bool do_timed_lock (const time::time &t) {
          return bits::thread::mutex_timed_lock(mutex_, t);
      }


      void do_unlock () {
          bits::thread::mutex_unlock(mutex_);
      }


      void do_lock (cv_state &state) {
          bits::thread::mutex_lock(mutex_, state);
      }


      void do_unlock (cv_state &state) {
          bits::thread::mutex_unlock(mutex_, state);
      }


      bits::thread::timed_mutex_type mutex_;
};


/** \} */
}   // namespace cool


#endif  /* cool_thread_mutex_hpp_included */
