/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_recursive_mutex_hpp_included
#define cool_thread_recursive_mutex_hpp_included

/**
 * \file recursive_mutex.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::recursive_mutex, cool::recursive_try_mutex,
 * cool::recursive_timed_mutex
 *
 * In module \ref cool_thread.
 */

#include <cool/cool.hpp>
#include <cool/time/time.hpp>
#include <cool/util/noncopyable.hpp>
#include "bits/recursive_mutex.hpp"
#include "bits/lock.hpp"


namespace cool {
/** \addtogroup cool_thread */
/** \{ */


/**
 * \brief Simple recursive_mutex.
 * For more information, see \ref recursive_mutex.hpp
 */
class COOL_API recursive_mutex: private noncopyable
{
  public:

      typedef bits::thread::scoped_lock<recursive_mutex> scoped_lock;


      /** \brief Create simple recursive_mutex */
      recursive_mutex () {
          bits::thread::recursive_mutex_init(mutex_);
      }


      /** \brief Destroy recursive_mutex */
      ~recursive_mutex () {
          bits::thread::recursive_mutex_destroy(mutex_);
      }


  private:

      friend class bits::thread::lock_ops<recursive_mutex>;

      typedef bits::thread::recursive_mutex_cv_state cv_state;


      void do_lock () {
          bits::thread::recursive_mutex_lock(mutex_);
      }


      void do_unlock () {
          bits::thread::recursive_mutex_unlock(mutex_);
      }


      void do_lock (cv_state &state) {
          bits::thread::recursive_mutex_lock(mutex_, state);
      }


      void do_unlock (cv_state &state) {
          bits::thread::recursive_mutex_unlock(mutex_, state);
      }


      bits::thread::recursive_mutex_type mutex_;
};


/**
 * \brief Simple try recursive_mutex.
 * For more information, see \ref recursive_mutex.hpp
 */
class COOL_API recursive_try_mutex: private noncopyable
{
  public:

      typedef bits::thread::scoped_lock<recursive_try_mutex> scoped_lock;
      typedef bits::thread::scoped_try_lock<recursive_try_mutex>
          scoped_try_lock;


      /** \brief Create try recursive_mutex */
      recursive_try_mutex () {
          bits::thread::recursive_mutex_init(mutex_);
      }


      /** \brief Destroy recursive_mutex */
      ~recursive_try_mutex () {
          bits::thread::recursive_mutex_destroy(mutex_);
      }


  private:

      friend class bits::thread::lock_ops<recursive_try_mutex>;

      typedef bits::thread::recursive_try_mutex_cv_state cv_state;


      void do_lock () {
          bits::thread::recursive_mutex_lock(mutex_);
      }


      bool do_try_lock () {
          return bits::thread::recursive_mutex_try_lock(mutex_);
      }


      void do_unlock () {
          bits::thread::recursive_mutex_unlock(mutex_);
      }


      void do_lock (cv_state &state) {
          bits::thread::recursive_mutex_lock(mutex_, state);
      }


      void do_unlock (cv_state &state) {
          bits::thread::recursive_mutex_unlock(mutex_, state);
      }


      bits::thread::recursive_try_mutex_type mutex_;
};


/**
 * \brief Simple timed recursive_mutex.
 * For more information, see \ref recursive_mutex.hpp
 */
class COOL_API recursive_timed_mutex: private noncopyable
{
  public:

      typedef bits::thread::scoped_lock<recursive_timed_mutex>
          scoped_lock;
      typedef bits::thread::scoped_try_lock<recursive_timed_mutex>
          scoped_try_lock;
      typedef bits::thread::scoped_timed_lock<recursive_timed_mutex>
          scoped_timed_lock;


      /** \brief Create timed recursive_mutex */
      recursive_timed_mutex () {
          bits::thread::recursive_mutex_init(mutex_);
      }


      /** \brief Destroy recursive_mutex */
      ~recursive_timed_mutex () {
          bits::thread::recursive_mutex_destroy(mutex_);
      }


  private:

      friend class bits::thread::lock_ops<recursive_timed_mutex>;

      typedef bits::thread::recursive_timed_mutex_cv_state cv_state;


      void do_lock () {
          bits::thread::recursive_mutex_lock(mutex_);
      }


      bool do_try_lock () {
          return bits::thread::recursive_mutex_try_lock(mutex_);
      }


      bool do_timed_lock (const time::time &t) {
          return bits::thread::recursive_mutex_timed_lock(mutex_, t);
      }


      void do_unlock () {
          bits::thread::recursive_mutex_unlock(mutex_);
      }


      void do_lock (cv_state &state) {
          bits::thread::recursive_mutex_lock(mutex_, state);
      }


      void do_unlock (cv_state &state) {
          bits::thread::recursive_mutex_unlock(mutex_, state);
      }


      bits::thread::recursive_timed_mutex_type mutex_;
};


/** \} */
}   // namespace cool


#endif  /* cool_thread_recursive_mutex_hpp_included */
