/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_thread_hpp_included
#define cool_thread_thread_hpp_included

/**
 * \file thread.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::thread
 *
 * In module \ref cool_thread.
 */

#include <cool/cool.hpp>
#include <cool/util/noncopyable.hpp>
#include <cool/time/time.hpp>
#include "bits/thread.hpp"


namespace cool {
/** \addtogroup cool_thread */
/** \{ */


/**
 * \brief Concurrent thread of execution.
 *
 * When application starts up, there is only a single thread -- main(). This
 * main thread can create other threads that run concurrently.
 *
 * Threads are created undetached i.e. joinable. join() can be used to suspend
 * calling function until thread function returns. After that thread doesn't
 * longer exists though thread object continues to exist until out of scope.
 * If thread's object lifetime is shorter than thread function's running time,
 * then thread is automatically detached.
 */
class COOL_API thread: private noncopyable
{
  public:

      /** \brief Represent thread of execution object */
      struct handler {

          /** \brief Destructor */
          virtual ~handler () {}

          /**
           * \brief This method is called automatically when thread is
           * created.
           *
           * \see thread::thread(handler*)
           */
          virtual void operator() () = 0;
      };


      /**
       * \brief Construct thread object representing current thread of
       * execution.
       *
       * Created thread object can't be join() -ed because it would create
       * deadlock i.e. thread waits itself to be finished.
       */
      thread (): joinable_(false), thread_() { }


      /** \brief Create new joinable thread and run handler::operator(). */
      thread (handler *h);


      /**
       * \brief Create joinable thread and run \a h specifying \a arg as
       * parameter.
       */
      thread (void (*h)(void *), void *arg = 0);


      /**
       * \brief Destroy thread object. Actual thread of execution mey continue
       * to execute after object is destroyed.
       *
       * If thread is joinable, then thread is detached, i.e. resources used
       * by system low level thread will be released automatically when thread
       * finishes. To ensure such thread of execution run to completion before
       * thread object is destroyed, call join().
       */
      ~thread () {
          if (joinable_)
              thread_.detach();
      }


      /**
       * \brief Suspend caller until thread completes it's execution.
       *
       * join() can be called only if thread is joinable. Call to join()
       * automatically frees system low level resources after finishing.
       */
      void join () {
          assert(joinable_ == true);
          thread_.join();
          joinable_ = false;
      }


      /** \brief Suspend calling thread until time \a t is passed. */
      static void sleep (const time::time &t) {
          bits::thread::thread::sleep(t);
      }


      /** \brief Thread voluntarily yields it's remaining CPU time-slice */
      static void yield () {
          bits::thread::thread::yield();
      }


      /** \brief Return \c true if \c *this is same thread as \a o */
      bool operator == (const thread &o) const {
          return thread_.equal(o.thread_);
      }


      /** \brief Return \c true if \c *this is different thread than \a o */
      bool operator != (const thread &o) const {
          return thread_.equal(o.thread_) == false;
      }


  private:

      bool joinable_;
      bits::thread::thread thread_;
};


/** \} */
}   // namespace cool


#endif  /* cool_thread_thread_hpp_included */
