/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_thread_list_hpp_included
#define cool_thread_thread_list_hpp_included

/**
 * \file thread_list.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::thread_list
 *
 * In module \ref cool_thread.
 */

#include <cool/cool.hpp>
#include <cool/thread/thread.hpp>
#include <cool/thread/mutex.hpp>
#include list_header
#include memory_header
#include algorithm_header


namespace cool {
/** \addtogroup cool_thread */
/** \{ */


/**
 * \brief Collection of threads.
 *
 * thread_list can be used to maintain list of threads as one i.e. application
 * can start several threads using create() and when finishing, wait for all
 * threads to finish using join_all().
 */
class COOL_API thread_list: private noncopyable
{
  public:

      /** \brief Create empty list of threads */
      thread_list () {}


      /** \brief Destroy all thread objects but doesn't join them */
      ~thread_list () {
          STD_ for_each(threads_.begin(), threads_.end(), delete_thread());
      }


      /** \brief Create and start new thread using thread handler \a h. */
      thread *create (thread::handler *h) {
          assert(h != 0);
          STD_ auto_ptr<thread> t(new thread(h));
          add(t.get());
          return t.release();
      }


      /**
       * \brief Create and start new thread using function \a h.
       *
       * \a arg is specified as parameter for function \a h when it is
       * starting.
       */
      thread *create (void (*h)(void *), void *arg = 0) {
          assert(h != 0);
          STD_ auto_ptr<thread> t(new thread(h, arg));
          add(t.get());
          return t.release();
      }


      /**
       * \brief Add already running thread \a t to list.
       *
       * Each thread can be added only once to list. When attempted to add
       * multiple times, assertion is thrown.
       */
      void add (thread *t) {
          mutex::scoped_lock guard(mutex_);
          assert(STD_ find(threads_.begin(), threads_.end(), t) ==
                  threads_.end());
          threads_.push_back(t);
      }


      /** \brief Remove thread \a t from list of threads */
      void remove (thread *t) {
          mutex::scoped_lock guard(mutex_);
          thr_list::iterator i =
              STD_ find(threads_.begin(), threads_.end(), t);
          assert(i != threads_.end());
          threads_.erase(i);
      }


      /** \brief Wait all threads in list to be stopped */
      void join_all () {
          mutex::scoped_lock guard(mutex_);
          STD_ for_each(threads_.begin(), threads_.end(), join_thread());
      }


  protected:

      typedef STD_ list<thread *> thr_list;

      struct delete_thread {
          void operator() (thr_list::value_type t) {
              delete t;
          }
      };


      struct join_thread {
          void operator() (thr_list::value_type t) {
              t->join();
          }
      };

      thr_list threads_;
      mutex mutex_;
};



/** \} */
}   // namespace cool


#endif  /* cool_thread_thread_list_hpp_included */
