/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_thread_tss_hpp_included
#define cool_thread_tss_hpp_included

/**
 * \file tss.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::thread_specific_ptr
 *
 * In module \ref cool_thread.
 */

#include <cool/cool.hpp>
#include <cool/util/noncopyable.hpp>
#include "bits/tss.hpp"


namespace cool {
/** \addtogroup cool_thread */
/** \{ */


/**
 * \brief Thread specific storage (TSS)
 *
 * It is data associated with individual threads and is often used to make
 * operations that rely on global data. For example, \c errno variable in
 * multi-threaded application: it is used as variable but it must be
 * individual for each thread as any thread can override it's value during
 * examination in other thread.
 *
 * Template thread_specific_ptr<T> stores a pointer to variable of type \a T
 * on a thread-by-thread basis and calls a specified cleanup handler on the
 * contained pointer when thread terminates. The cleanup handlers are called
 * in the reverse order of construction of the thread_specific_ptr-s, and for
 * initial thread are called by the destructor. Each thread initially stores
 * the NULL in each instance.
 */
template <typename T>
class COOL_API thread_specific_ptr: private noncopyable
{
  public:

      /**
       * \brief Allocate and initialize thread-specific data visible to all
       * threads.
       *
       * The cleanup function \a cf is registered with key and will be called
       * for a thread during exit if it has a non-NULL value associated with
       * key.
       */
      thread_specific_ptr (void (*cf)(T*) = thread_specific_ptr<T>::cleanup)
          : tss_((void (*)(void*))cf)
      { }


      /** \brief Delete the thread-specific data allocate by constructor. */
      ~thread_specific_ptr () {
          reset();
      }


      /**
       * \brief Return object stored in TSS for the current thread.
       *
       * Each thread initially returns \c NULL.
       */
      T *get () const {
          return static_cast<T *>(tss_.get());
      }


      /** \brief get() */
      T *operator-> () const {
          return get();
      }


      /** \brief get() */
      T &operator* () const {
          return *get();
      }


      /** \brief Return current value and set new to \c NULL */
      T *release () {
          T *t = get();
          if (t)
              tss_.set(0);
          return t;
      }


      /** \brief Set value to \a p for current thread */
      void reset (T *p = 0) {
          T *c = get();
          if (c == p)
              return;
          tss_.set(p);
          if (c)
              tss_.cleanup(c);
      }


  private:

      static void cleanup (T *p) {
          delete p;
      }

      bits::thread::tss tss_;
};


/** \} */
}   // namespace cool


#endif  /* cool_thread_tss_hpp_included */
