/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_time_time_hpp_included
#define cool_time_time_hpp_included

/**
 * \file time.hpp
 * \author Sven Suursoho
 *
 * \brief Class(es): cool::time
 *
 * In module \ref cool_time.
 */

#include <cool/cool.hpp>
#include <cool/util/convert.hpp>
#include "bits/time.hpp"

#if defined(ostream_header)
#  include ostream_header
#else
#  include iostream_header
#endif


namespace cool {
/**
 * \addtogroup cool_time
 * \{
 */


/**
 * \brief Absolute time since Epoch (00:00:00 UTC, 1970-01-01) measured n
 * seconds and nanoseconds.
 *
 * The time::nsec value has granularity much greater than 1 because of the
 * resolution of system clock. Granularity is more likely equal to 1000000ns
 * or even 1sec if C standard library ::time() is used as low level time
 * getting function. Relying on any granularity (particularly on 1) will make
 * code non-portable. */
struct COOL_API time
{
  /** \brief Number of seconds */
  typedef bits::time::sec_type sec_type;


  /** \brief Number of nanoseconds */
  typedef bits::time::sec_type nsec_type;


  enum {
      nsec_per_sec = bits::time::nsec_per_sec
      , nsec_per_msec = bits::time::nsec_per_msec
      , usec_per_sec = bits::time::usec_per_sec
      , nsec_per_usec = bits::time::nsec_per_usec
      , usec_per_msec = bits::time::usec_per_msec
      , msec_per_sec = bits::time::msec_per_sec
  };


  /**
   * \brief Construct object that has time of construction.
   *
   * Parameters \a rel_sec and \a rel_nsec can be used to set this time
   * relative to current moment by adding \a rel_sec to time::sec and \a
   * rel_nsec to time::nsec. */
  time (sec_type rel_sec = 0, nsec_type rel_nsec = 0) throw() {
      now(rel_sec, rel_nsec);
  }


  /** \brief Construct a time object that has value parsed from as_string */
  time (const char *as_string, char **eptr = 0) {
      parse_iso8601(as_string, eptr);
  }


  /** \brief Construct object that will have time of \a t */
  time (const time &t): sec(t.sec), nsec(t.nsec) { }


  /** \brief Convert cool::time to struct timespec duration */
  operator struct timespec () const {
      return bits::time::as_timespec(sec, nsec);
  }


  /** \brief Convert cool::time to struct timeval duration */
  operator struct timeval () const {
      return bits::time::as_timeval(sec, nsec);
  }


  /**
   * \brief Update object's time fields to current time.
   *
   * \copydoc time::time() */
  time &now (sec_type rel_sec = 0, nsec_type rel_nsec = 0) throw() {
      sec = get(&nsec) + rel_sec;
      nsec += rel_nsec;
      sec += nsec/nsec_per_sec;
      nsec %= nsec_per_sec;
      return *this;
  }


  /**
   * \brief Copy string representation of time to \a buf not exceeding \a
   * bufsz byte(s).
   *
   * \a buf length (and \a bufsz) must be at least 26 bytes. If not, assertion
   * is thrown.
   *
   * Time format is YYYY-MM-DDThh:mm:ssTZD (see
   * http://www.w3.org/TR/NOTE-datetime). If timezone offset is 00:00, then
   * TZD is 'Z' otherwise offset is printed. */
  char *as_iso8601 (char *buf, size_t bufsz = 26) const {
      assert(bufsz >= 26);
      return bits::time::as_iso8601(sec, buf, bufsz);
  }


  /**
   * \brief Extract time value from string in ISO8601 format (subset).
   * \see as_iso8601 */
  time &parse_iso8601 (const char *buf, char **eptr = 0) {
      nsec = 0;
      sec = bits::time::parse_iso8601(buf, eptr);
      return *this;
  }


  /**
   * \brief Return number of seconds and possibly nanoseconds (if address is
   * specified) since Epoch. */
  static sec_type get (nsec_type *nanosec = 0) throw() {
      return bits::time::get(nanosec);
  }


  /** \brief Timezone offset from UTC (in seconds) */
  static const long &tz;


  /**
   * \brief Daylight saving.
   *
   * 0 if there is no daylight savings time rules, nonzero if there is a time
   * during year when daylight savings time applies. Note: it doesn't show if
   * daylight savings is currently in effect */
  static const int &ds;


  /**
   * \brief Time object representing moment of beginning Epoch.
   *
   * When declaring cool::time object and don't want to initialize to current
   * moment (that might do syscall), then use this as initializer. */
  static const time &null;


  /** \name Absolute time since Epoch */
  /*\{*/
  sec_type sec;     /**< Number for seconds */
  nsec_type nsec;   /**< Nanoseconds to be added to time::sec */
  /*\}*/


private:
  // ugly trick to create 'null' variable that is initialized to Epoch
  // null = epoch()
  enum dummy { };
  time (dummy): sec(0), nsec(0) { }
  static time epoch () { return time(dummy()); }
};


/**
 * \brief Store string representation of \a t to stream \a os.
 * \see cool::time::as_iso8601() */
inline STD_ ostream &
operator << (STD_ ostream &os, const time &t)
{
  char buf[26];
  os << t.as_iso8601(buf, sizeof(buf));
  return os;
}


/**
 * \brief Compare two time structures.
 *
 * \return
 *  - < 0: \a t1 has time before \a t2
 *  - = 0: \a t1 has same time as \a t2
 *  - > 0: \a t1 has time after \a t2 */
inline int
time_cmp (const time &t1, const time &t2)
{
  int cmp = t1.sec - t2.sec;
  if (cmp == 0)
      return t1.nsec - t2.nsec;
  return cmp;
}


/** \brief Return \c true, if t1 == t2 and \c false otherwise */
inline bool
operator == (const time &t1, const time &t2)
{ return time_cmp(t1, t2) == 0; }


/** \brief Return \c true, if t1 != t2 and \c false otherwise */
inline bool
operator != (const time &t1, const time &t2)
{ return time_cmp(t1, t2) != 0; }


/** \brief Return \c true, if t1 < t2 and \c false otherwise */
inline bool
operator < (const time &t1, const time &t2)
{ return time_cmp(t1, t2) < 0; }


/** \brief Return \c true, if t1 <= t2 and \c false otherwise */
inline bool
operator <= (const time &t1, const time &t2)
{ return time_cmp(t1, t2) <= 0; }


/** \brief Return \c true, if t1 > t2 and \c false otherwise */
inline bool
operator > (const time &t1, const time &t2)
{ return time_cmp(t1, t2) > 0; }


/** \brief Return \c true, if t1 >= t2 and \c false otherwise */
inline bool
operator >= (const time &t1, const time &t2)
{ return time_cmp(t1, t2) >= 0; }


template<> inline cool::time
as<cool::time> (const char *as_string, char **eptr)
{
  cool::time t(cool::time::null);
  t.parse_iso8601(as_string, eptr);
  return t;
}


/** \} */
}   // namespace cool


#endif  /* cool_time_time_hpp_included */
