/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#ifndef cool_util_convert_hpp_included
#define cool_util_convert_hpp_included

/**
 * \file convert.hpp
 * \author Sven Suursoho
 *
 * \brief Function(s): as(const char *, char **)
 *
 * In module \ref cool_util.
 */

#include <cool/cool.hpp>
#include "bits/convert.hpp"


namespace cool {
/**
 * \addtogroup cool_util
 * \{
 */


/** \brief Dummy name used in template function to show error */
template <typename T> T
ERROR_unsupported_type_in_string_conversion ();


/**
 * \brief C-style string parser for built-in types.
 *
 * Allowed only for certain types, for which this function is explicitly
 * specialized. Library user can add new parsers by specializing function:
 * \code
 * template <typename T> T as (const char *as_string, char **eptr = 0);
 * \endcode
 * for other types.
 * If parser is not defined for some specific type, then generic version will
 * be instantiated which result in a link error for symbol
 * ERROR_unsupported_type_in_string_conversion().
 *
 * This function returns new object of type \a T initializing with value
 * parsed from \a as_string. On return, if \a eptr isn't \c NULL, it will
 * point to first character in \a as_string that can't be converted. If whole
 * string was correct, it will point to end of string.
 *
 * Throws format_error if no characters were converted.
 * Throws limit_error if parsed value is out of range for type T.
 */
template <typename T> inline T
as (const char *as_string, char **eptr = 0)
{
  return ERROR_unsupported_type_in_string_conversion<T>();
}


/**
 * \brief Convert string to boolean variable.
 *
 * Only first character is checked, returns \c true if it is in ('t', 'T', 'y',
 * 'Y', '1'..'9') and \c false otherwise.
 */
template<> inline bool
as (const char *as_string, char **eptr)
{
  return bits::convert::bool_(as_string, eptr);
}


template<> inline long
as<long> (const char *as_string, char **eptr)
{
  return bits::convert::long_(as_string, eptr);
}


template<> inline unsigned long
as<unsigned long> (const char *as_string, char **eptr)
{
  return bits::convert::unsigned_long_(as_string, eptr);
}


template<> inline char
as<char> (const char *as_string, char **eptr)
{
  signed long v = bits::convert::long_(as_string, eptr);
  if (bits::convert::char_min > v || v > bits::convert::char_max)
      raise<limit_error>(COOL_LOCUS);
  return static_cast<char>(v);
}


template<> inline unsigned char
as<unsigned char> (const char *as_string, char **eptr)
{
  unsigned long v = bits::convert::unsigned_long_(as_string, eptr);
  if (v > bits::convert::uchar_max)
      raise<limit_error>(COOL_LOCUS);
  return static_cast<unsigned char>(v);
}


template<> inline short
as<short> (const char *as_string, char **eptr)
{
  signed long v = bits::convert::long_(as_string, eptr);
  if (bits::convert::short_min > v || v > bits::convert::short_max)
      raise<limit_error>(COOL_LOCUS);
  return static_cast<short>(v);
}


template<> inline unsigned short
as<unsigned short> (const char *as_string, char **eptr)
{
  unsigned long v = bits::convert::unsigned_long_(as_string, eptr);
  if (v > bits::convert::ushort_max)
      raise<limit_error>(COOL_LOCUS);
  return static_cast<unsigned short>(v);
}


template<> inline int
as<int> (const char *as_string, char **eptr)
{
  signed long v = bits::convert::long_(as_string, eptr);
  if (bits::convert::int_min > v || v > bits::convert::int_max)
      raise<limit_error>(COOL_LOCUS);
  return static_cast<int>(v);
}


template<> inline unsigned int
as<unsigned int> (const char *as_string, char **eptr)
{
  unsigned long v = bits::convert::unsigned_long_(as_string, eptr);
  if (v > bits::convert::uint_max)
      raise<limit_error>(COOL_LOCUS);
  return static_cast<unsigned int>(v);
}


/** \} */
}   // namespace cool


#endif  /* cool_util_convert_hpp_included */
