/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#include <cool/net/client_socket.hpp>
#include <cool/net/exception.hpp>


namespace cool {
namespace bits {
namespace net {


bool
connect (socket_base::socket_type s, const socket::address &a)
{
  if (::connect(s, a.addr(), a.size()) == -1) {
      os_error::error_type e = os_error::last();
      switch (e) {
          case EISCONN:
              return true;

          case EALREADY:
          case EINPROGRESS:
          case EWOULDBLOCK:
#if defined(EAGAIN)
#  if EAGAIN != EWOULDBLOCK
          case EAGAIN:
#  endif
#endif
              return false;

          case ECONNREFUSED:
              raise<connection_refused>(COOL_LOCUS);

          case EADDRINUSE:
              raise<resource_error>(COOL_LOCUS);

          case EACCES:
          case EPERM:
              raise<permission_error>(COOL_LOCUS);

          default:
              throw_last_netio(e, "connect", COOL_LOCUS);
      }
  }

  return true;
}


bool
is_connected (socket_base::socket_type s)
{
  if (s == -1)
      return false;

  sockaddr a;
  socket_base::socklen_type len = sizeof(a);
  if (::getpeername(s, &a, &len) == -1) {
      os_error::error_type e = os_error::last();
      if (e != ENOTCONN)
          throw_last_netio(e, "getpeername", COOL_LOCUS);
      return false;
  }
  return true;
}


int
recv (socket_base::socket_type s, char *buf, int len,
        socket_base::io_flags_type f)
{
  while (true) {
      int r = ::recv(s, buf, len, f);
      if (r >= 0)
          return r;

      os_error::error_type e = os_error::last();
      if (e == EAGAIN || e == EWOULDBLOCK)
          return -1;

      if (e == EINTR)
          continue;

      throw_last_netio(e, "recv", COOL_LOCUS);
  }

  return 0;
}


int
recvfrom (socket_base::socket_type s, socket::address &a,
        char *buf, int len, socket_base::io_flags_type f)
{
  socket_base::socklen_type addr_len = a.size();
  while (true) {
      int r = ::recvfrom(s, buf, len, f, a.addr(), &addr_len);
      if (r >= 0)
          return r;

      os_error::error_type e = os_error::last();
      if (e == EAGAIN || e == EWOULDBLOCK)
          return -1;

      if (e == EINTR)
          continue;

      throw_last_netio(e, "recvfrom", COOL_LOCUS);
  }

  return 0;
}


int
send (socket_base::socket_type s, char *buf, int len,
        socket_base::io_flags_type f)
{
  int r = 0;

  while (true) {
      if ((r = ::send(s, buf, len, f)) >= 0)
          break;

      os_error::error_type e = os_error::last();
      if (e == EAGAIN || e == EWOULDBLOCK)
          break;
      else if (e == EINTR)
          continue;

      throw_last_netio(e, "send", COOL_LOCUS);
  }

  return r;
}

int
sendto (socket_base::socket_type s, const socket::address &a,
        char *buf, int len, socket_base::io_flags_type f)
{
  int r = 0;

  while (true) {
      if ((r = ::sendto(s, buf, len, f, a.addr(), a.size())) >= 0)
          break;

      os_error::error_type e = os_error::last();
      if (e == EAGAIN || e == EWOULDBLOCK)
          break;
      else if (e == EINTR)
          continue;

      throw_last_netio(e, "sendto", COOL_LOCUS);
  }

  return r;
}


}   // namespace net
}   // namespace bits
}   // namespace cool
