/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#include <cool/net/poller.hpp>


namespace cool {
namespace bits {
namespace net {


#if defined(SS_HAVE_EPOLL)


void
poller_type::insert (socket *s, socket::event_type ev)
{
  // make room for receiving event
  if (size_ == allocated_) {
      int nsz = allocated_ << 1;
      epoll_event *tmp = (epoll_event *)realloc(events_, nsz*sizeof(*events_));
      if (!tmp)
          raise<resource_error>(COOL_LOCUS);

      // set new array
      allocated_ = nsz;
      events_ = tmp;
  }

  // insert to internal poller
  mod_fd(s, ev, true);
  size_++;
}


void
poller_type::mod_fd (socket *s, socket::event_type ev, bool add)
{
  epoll_event pev;
  pev.data.ptr = s;
  pev.events = 0;
  if (ev & socket::in) pev.events |= EPOLLIN;
  if (ev & socket::out) pev.events |= EPOLLOUT;
  if (ev & socket::except) pev.events |= EPOLLPRI;

  int op = add ? EPOLL_CTL_ADD : EPOLL_CTL_MOD;
  if (::epoll_ctl(epollfd_, op, *s, &pev) == -1)
      throw_last("epoll_ctl", COOL_LOCUS);
}


int
poller_type::wait (const cool::time *timeout)
{
  pending_ = last_ = 0;
  if (!size_)
      return 0;

  timeval tv = { 0, 0 }, *tp = 0;
  if (timeout) {
      tv = *timeout;
      if (tv.tv_sec || tv.tv_usec)
          tp = &tv;
  }
  int t = tp
      ? tp->tv_sec*time::msec_per_sec + tp->tv_usec/time::usec_per_msec
      : -1;

  if ((last_ = ::epoll_wait(epollfd_, events_, size_, t)) == -1)
      throw_last("epoll_wait", COOL_LOCUS);

  return last_;
}


#elif defined(SS_HAVE_POLL)


void
poller_type::insert (socket *s, socket::event_type ev)
{
  assert(sockets_.find(*s) == sockets_.end());

  // find free slot
  int at;
  if (begin_ > 0)
      // at head
      at = --begin_;
  else if (end_ < size_)
      // at tail
      at = end_++;
  else {
      // no room, increase size
      int nsz = size_ << 1;
      pollfd *tmp;
      if ((tmp = (pollfd *)realloc(fds_, nsz*sizeof(*fds_))) == 0)
          raise<resource_error>(COOL_LOCUS);
      fds_ = tmp;
      size_ = nsz;

      // append
      at = end_++;
  }

  // add to array
  fds_[at].fd = *s;
  fds_[at].events = cool_ev_to_sys_ev(ev);

  // add to map
  socket_data_type se(s, at);
  STD_ pair<socket::socket_type, socket_data_type> v(*s, se);
  sockets_.insert(v);
}



void
poller_type::erase (socket *s)
{
  socket_map::iterator i = sockets_.find(*s);
  assert(i != sockets_.end());

  // erase from array
  if (i->second.ndx_ == begin_)
      // head
      begin_++;
  else if (i->second.ndx_ == end_ - 1)
      // tail
      end_--;
  else {
      // middle
      socket_map::iterator b = sockets_.find(fds_[begin_].fd);
      if (b != sockets_.end()) {
          // move first to removed slot
          fds_[i->second.ndx_].fd = fds_[begin_].fd;
          fds_[i->second.ndx_].events = fds_[begin_].events;
          b->second.ndx_ = i->second.ndx_;
          begin_++;
      }
  }

  // erase from map
  sockets_.erase(i);
}


int
poller_type::wait (const cool::time *timeout)
{
  pending_ = sockets_.end();
  if (sockets_.size() == 0)
      return 0;

  timeval tv = { 0, 0 }, *tp = 0;
  if (timeout) {
      tv = *timeout;
      if (tv.tv_sec || tv.tv_usec)
          tp = &tv;
  }
  int t = tp
      ? tp->tv_sec*time::msec_per_sec + tp->tv_usec/time::usec_per_msec
      : -1;

  if ((last_ = ::poll(fds_ + begin_, end_ - begin_, t)) == -1)
      throw_last("poll", COOL_LOCUS);
  if (last_ > 0)
      pending_ = sockets_.begin();

  return last_;
}


#elif defined(SS_HAVE_SELECT)


void
poller_type::insert (socket *s, socket::event_type ev)
{
  assert(sockets_.find(*s) == sockets_.end());

  if (size() == FD_SETSIZE)
      raise<limit_error>(COOL_LOCUS);

  socket::event_type old_ev = socket::no_events;
  socket::event_type new_ev = ev;
  set(socket::in, old_ev, new_ev, s, &rd);
  set(socket::out, old_ev, new_ev, s, &wr);
  set(socket::except, old_ev, new_ev, s, &ex);

  socket_data_type se(s, ev);
  STD_ pair<socket::socket_type, socket_data_type> v(*s, se);
  sockets_.insert(v);
}


void
poller_type::modify (socket *s, socket::event_type ev)
{
  socket_map::iterator i = sockets_.find(*s);
  assert(i != sockets_.end());

  socket::event_type old_ev = i->second.event_;
  socket::event_type new_ev = ev;
  set(socket::in, old_ev, new_ev, s, &rd);
  set(socket::out, old_ev, new_ev, s, &wr);
  set(socket::except, old_ev, new_ev, s, &ex);

  i->second.event_ = ev;
}


void
poller_type::erase (socket *s)
{
  socket_map::iterator i = sockets_.find(*s);
  assert(i != sockets_.end());

  socket::event_type old_ev = i->second.event_;
  socket::event_type new_ev = socket::no_events;
  set(socket::in, old_ev, new_ev, s, &rd);
  set(socket::out, old_ev, new_ev, s, &wr);
  set(socket::except, old_ev, new_ev, s, &ex);

  sockets_.erase(i);
}


int
poller_type::wait (const cool::time *timeout)
{
  pending_ = sockets_.end();
  if (sockets_.size() == 0)
      return 0;

  struct timeval tv, *tvp = 0;
  if (timeout) {
      tv = *timeout;
      tvp = &tv;
  }

  memcpy(&rrd, &rd, sizeof(rrd));
  memcpy(&rwr, &wr, sizeof(rwr));
  memcpy(&rex, &ex, sizeof(rex));

  if ((last_ = ::select(max_handle() + 1, &rrd, &rwr, &rex, tvp)) == -1)
      throw_last("select", COOL_LOCUS);
  if (last_ > 0)
      pending_ = sockets_.begin();

  return last_;
}


#endif


}  // namespace net
}  // namespace bits
}  // namespace cool
