/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#include <cool/net/socket.hpp>
#include <cool/net/exception.hpp>


// no point to hassle with epoll_* functions here as we are checking only 1
// socket here and no good scalability is needed
// poll() is preferred over select() though
#if defined(SS_HAVE_POLL)
#  if defined(HAVE_SYS_POLL_H)
#    include <sys/poll.h>
#  endif
#elif defined(SS_HAVE_SELECT)
#  if defined(HAVE_SYS_SELECT_H)
#    include <sys/select.h>
#  else
#    if defined(HAVE_SYS_TIME_H)
#      include <sys/time.h>
#    endif
#    if defined(HAVE_UNISTD_H)
#      include <unistd.h>
#    endif
#  endif
#endif


namespace cool {
namespace bits {
namespace net {


void
throw_last (const char *func_name, const char *locus)
{
  os_error::error_type e = os_error::last();
  switch (e) {
      case ENFILE:
      case EMFILE:
      case ENOBUFS:
      case ENOMEM:
          raise<resource_error>(locus);

      case EAFNOSUPPORT:
      case EPROTONOSUPPORT:
      case EOPNOTSUPP:
          raise<not_supported_error>(locus);

      case EACCES:
          raise<permission_error>(locus);

      case EINTR:
          raise<interrupted>(locus);

      case EBADF:
          raise<invalid_value_error>(locus);

      case EADDRINUSE:
          raise<in_use_error>(locus);

      default:
          raise(cool::os_error(e, func_name), locus);
  }
}


void
throw_last_netio (os_error::error_type e,
        const char *func_name, const char *locus)
{
  switch (e) {
      case EHOSTUNREACH:
      case ENETDOWN:
      case ENETUNREACH:
      case ETIMEDOUT:
          raise<network_error>(locus);

      case EPIPE:
      case ECONNRESET:
      case ENOTCONN:
          raise<connection_error>(locus);

      default:
          throw_last(func_name, locus);
  }
}


void
bind (socket_base::socket_type s,
        const sockaddr *a, socket_base::socklen_type l)
{
  if (::bind(s, a, l) == -1) {
      os_error::error_type e = os_error::last();
      switch (e) {
          case EINVAL:
              // XXX according to bind(2), this might change in future
              raise<already_bound>(COOL_LOCUS);

          case EACCES:
              raise<permission_error>(COOL_LOCUS);

          default:
              throw_last_netio(e, "bind", COOL_LOCUS);
      }
  }
}


socket_base::event_type
wait (socket_base::socket_type s, socket_base::event_type ev,
        const cool::time *timeout)
{
  assert(s > -1);
  assert(ev != socket_base::no_events);

  timeval tv = { 0, 0 }, *tvp = 0;
  if (timeout) {
      tv = *timeout;
      if (tv.tv_sec || tv.tv_usec)
          tvp = &tv;
  }

#if defined(SS_HAVE_POLL)

  pollfd pfd;
  pfd.fd = s;
  pfd.events = 0;
  if (ev & socket_base::in) pfd.events |= POLLIN;
  if (ev & socket_base::out) pfd.events |= POLLOUT;
  if (ev & socket_base::except) pfd.events |= POLLPRI;

  int t = tvp
      ? tvp->tv_sec*time::msec_per_sec + tvp->tv_usec/time::usec_per_msec
      : -1;

  ev = socket_base::no_events;
  os_error::error_type res = ::poll(&pfd, 1, t);
  if (res == -1)
      throw_last("poll", COOL_LOCUS);
  if (res == 0)
      return ev;

  if (pfd.revents & POLLIN) ev |= socket_base::in;
  if (pfd.revents & POLLOUT) ev |= socket_base::out;
  if (pfd.revents & POLLPRI) ev |= socket_base::except;

#elif defined(SS_HAVE_SELECT)

  fd_set r, *rp = 0;
  if (ev & socket_base::in) {
      rp = &r;
      FD_ZERO(rp);
      FD_SET(s, rp);
  }

  fd_set w, *wp = 0;
  if (ev & socket_base::out) {
      wp = &w;
      FD_ZERO(wp);
      FD_SET(s, wp);
  }

  fd_set e, *ep = 0;
  if (ev & socket_base::except) {
      ep = &e;
      FD_ZERO(ep);
      FD_SET(s, ep);
  }

  ev = socket_base::no_events;
  os_error::error_type res = ::select(s + 1, rp, wp, ep, tvp);
  if (res == -1)
      throw_last("select", COOL_LOCUS);
  if (res == 0) {
      return ev;
  }

  if (rp && FD_ISSET(s, rp)) ev |= socket_base::in;
  if (wp && FD_ISSET(s, wp)) ev |= socket_base::out;
  if (ep && FD_ISSET(s, ep)) ev |= socket_base::except;

#else

#  error No socket multiplexor function

#endif

  return ev;
}


}  // namespace net
}  // namespace bits
}  // namespace cool
