/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#include <cool/net/url.hpp>
#include <cool/net/exception.hpp>
#include <cool/net/inet_socket.hpp>

#include <stdio.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#ifdef HAVE_CTYPE_H
#  include <ctype.h>
#endif


namespace cool {


void
url::alloc (int size)
{
  if (!size) {
      // release
      if (url_) {
          free(url_);
          url_ = end = 0;
      }
      return;
  }

  if (end - url_ < size) {
      char *tmp = (char *)(url_ ? realloc(url_, size) : malloc(size));
      if (!tmp)
          raise<resource_error>(COOL_LOCUS);

      url_ = tmp;
      end = url_ + size;
  }
}


url &
url::operator = (const url &u)
{
  if (&u != this) {
      int size = u.end - u.url_;
      alloc(size);
      memcpy(url_, u.url_, size);

      scheme_ = u.scheme_ ? url_ + (u.scheme_ - u.url_) : 0;
      login_ = u.login_ ? url_ + (u.login_ - u.url_) : 0;
      host_ = u.host_ ? url_ + (u.host_ - u.url_) : 0;
      path_ = u.path_ ? url_ + (u.path_ - u.url_) : 0;
      params_ = u.params_ ? url_ + (u.params_ - u.url_) : 0;
      query_ = u.query_ ? url_ + (u.query_ - u.url_) : 0;
      ref_ = u.ref_ ? url_ + (u.ref_ - u.url_) : 0;

      port_ = u.port_;
      default_port_ = u.default_port_;
  }

  return *this;
}


// RFC 2396: 3 URI Syntactic Components
url &
url::operator = (const char *u_)
{
  scheme_ = login_ = host_ = path_ = params_ = query_ = ref_ = 0;
  port_ = default_port_ = -1;

  if (!u_ || !*u_)
      return *this;

  // begin of URL
  register char *ub = (char *)u_;
  while (*ub && isspace(*ub))
      ub++;

  // end of URL
  register char *ue = ub + strlen(ub);
  while (ue > ub && isspace(*ue))
      ue--;

  if (ue == ub)
      return *this;

  alloc((ue - ub) + 8*sizeof('\0'));
  
  // extract scheme
  path_ = url_; 
  register char *p = url_;
  register char *up = ub;
  if (isalpha(*up)) {
      while (up < ue &&
              (isalnum(*up) || *up == '+' || *up == '-' || *up == '.'))
          *p++ = *up++;
      if (up < ue && *up == ':') {
          up++;
          *p++ = '\0';
          scheme_ = url_;
          path_ = 0;
      }
      *p = '\0';
  }
  
  // extract authority (if available)
  if (up < ue && up[0] == '/' && up[1] == '/') {
      up += 2;
      host_ = p;
      while (up < ue && *up != '/' && *up != '?' && *up != '#') {
          if ((*p++ = *up++) == '@') {
              // authority contains login -- actual host starts now
              login_ = host_;
              p[-1] = '\0';
              host_ = ++p;
          }
      }
      *p++ = '\0';
      path_ = 0;
  }
  
  // extract path
  if (up < ue) {
      if (path_ == 0)
          path_ = p; 
      while (up < ue && *up != ';' && *up != '?' && *up != '#')
          *p++ = *up++;
      *p++ = 0;
  }
  if (path_ && !*path_)
      path_ = 0;
  
  // extract params
  if (up < ue && *up == ';') {
      up++;
      params_ = p;
      while (up < ue && *up != '?' && *up != '#')
          *p++ = *up++;
      *p++ = '\0';
  }
  
  // extract query
  if (up < ue && *up == '?') {
      up++;
      query_ = p; 
      while (up < ue && *up != '#')
          *p++ = *up++;
      *p++ = '\0';
  }

  // extract ref
  if (up < ue && *up == '#') {
      up++;
      ref_ = p;
      while (up < ue)
          *p++ = *up++;
      *p++ = '\0';
  }

  // calculate port numbers
  if ((p = (char *)host_) != 0) {
      // specified port
      while (*p && *p != ':')
          p++;
      if (*p) {
          // found port separator
          *p = '\0';
          if ((port_ = atoi(p + 1)) == 0) {
              // but after that isn't numeric value -- restore port separator
              *p = ':';
              port_ = -1;
          }
      }
  }
  if (scheme_) {
      // default port
      try {
          default_port_ = inet::address::service(scheme_);
      }
      catch (service_not_found &e) {
          default_port_ = -1;
      }
  }

  return *this;
}


// RFC 1808: 4 Resolving Relative URLs
url &
url::resolve (const url &base)
{
  if (this == &base)
      return *this;

  // 1
  if (!base.url_)
      return *this;

  // 2.a
  if (!url_)
      return operator=(base);

  // 2.b, 2.c
  if (scheme_)
      return *this;
  scheme_ = base.scheme_;
  default_port_ = base.default_port_;

  // use "do {} while (false);" construct to avoid goto
  bool allocated_path = false;
  do {
      // 3
      if (host_ || login_)
          break;
      host_ = base.host_;
      login_ = base.login_;
      port_ = base.port_;

      // 4
      if (path_ && *path_ == '/')
          break;

      // 5
      if (!path_) {
          path_ = base.path_;

          // 5.a
          if (params_)
              break;
          params_ = base.params_;

          // 5.b
          if (query_)
              break;
          query_ = base.query_;

          if (ref_)
              break;
          ref_ = base.ref_;

          break;
      }

      // 6
      // calculate approx result path length before resolving
      register char *p = (char *)base.path_ + strlen(base.path_);
      while (p > base.path_ && p[-1] != '/')
          p--;
      int base_path_len = p - base.path_;
      int path_len = strlen(path_);

      p = (char *)malloc(base_path_len + path_len + 1);
      if (!p) {
          alloc(0);
          scheme_ = login_ = host_ = path_ = params_ = query_ = ref_ = 0;
          raise<resource_error>(COOL_LOCUS);
      }
      allocated_path = true;

      memcpy(p, base.path_, base_path_len);
      memcpy(p + base_path_len, path_, path_len + 1);

      path_ = p;
      register char *dst = (char *)path_;
      char *begin = p;
      while (*p) {
          // 6.a
          if (p[0] == '.' && p[1] == '/') {
              p += 2;
              continue;
          }

          // 6.c
          if (p[0] == '.' && p[1] == '.' && p[2] == '/') {
              if (dst > begin + 1) {
                  p += 3;
                  dst--;
                  while (dst > begin && dst[-1] != '/')
                      dst--;
              }
              else {
                  *dst++ = *p++;
                  *dst++ = *p++;
                  *dst++ = *p++;
                  begin = dst;
              }
              continue;
          }

          *dst++ = *p++;
      }

      if (dst > begin && dst[-1] == '.') {
          if (dst > begin + 1 && dst[-2] == '/')
              // 6.b
              dst--;
          else if (dst > begin + 2 && dst[-2] == '.' && dst[-3] == '/') {
              // 6.d
              dst -= 3;
              while (dst > begin && dst[-1] != '/')
                  dst--;
          }
      }
      *dst = '\0';
  } while (false);

  join(allocated_path);
  return *this;
}


void
url::join (bool free_path)
{
  int total = 0;
  int lscheme = scheme_ ? strlen(scheme_) + 1 : 0;
  total += lscheme;
  int llogin = login_ ? strlen(login_) + 1 : 0;
  total += llogin;
  int lhost = host_ ? strlen(host_) + 1 : 0;
  total += lhost;
  int lpath = path_ ? strlen(path_) + 1 : 0;
  total += lpath;
  int lparams = params_ ? strlen(params_) + 1 : 0;
  total += lparams;
  int lquery = query_ ? strlen(query_) + 1 : 0;
  total += lquery;
  int lref = ref_ ? strlen(ref_) + 1 : 0;
  total += lref;

  end = 0;
  if (total) {
      if ((end = (char *)malloc(total)) == 0) {
          alloc(0);
          if (free_path)
              free((void *)path_);
          scheme_ = login_ = host_ = path_ = params_ = query_ = ref_ = 0;
          raise<resource_error>(COOL_LOCUS);
      }
      end += total;

      if (ref_) {
          end -= lref;
          memcpy(end, ref_, lref);
          ref_ = end;
      }
      if (query_) {
          end -= lquery;
          memcpy(end, query_, lquery);
          query_ = end;
      }
      if (params_) {
          end -= lparams;
          memcpy(end, params_, lparams);
          params_ = end;
      }
      if (path_) {
          end -= lpath;
          memcpy(end, path_, lpath);
          if (free_path)
              free((void *)path_);
          path_ = end;
      }
      if (host_) {
          end -= lhost;
          memcpy(end, host_, lhost);
          host_ = end;
      }
      if (login_) {
          end -= llogin;
          memcpy(end, login_, llogin);
          login_ = end;
      }
      if (scheme_) {
          end -= lscheme;
          memcpy(end, scheme_, lscheme);
          scheme_ = end;
      }
  }

  free(url_);
  url_ = end;
  end = url_ + total;
}


char *
url::c_str (char *buf, int max)
{
  int l = 0;
  *buf = '\0';

  if (max - l > 1 && scheme_)
      l += snprintf(buf + l, max - l, "%s:", scheme_);
  if (max - l > 1 && host_ || login_)
      l += snprintf(buf + l, max - l, "//");
  if (max - l > 1 && login_)
      l += snprintf(buf + l, max - l, "%s@", login_);
  if (max - l > 1 && host_)
      l += snprintf(buf + l, max - l, "%s", host_);
  if (max - l > 1 && port_ != -1)
      l += snprintf(buf + l, max - l, ":%d", port_);
  if (max - l > 1 && path_)
      l += snprintf(buf + l, max - l, "%s", path_);
  if (max - l > 1 && params_)
      l += snprintf(buf + l, max - l, ";%s", params_);
  if (max - l > 1 && query_)
      l += snprintf(buf + l, max - l, "?%s", query_);
  if (max - l > 1 && ref_)
      l += snprintf(buf + l, max - l, "#%s", ref_);

  return buf;
}


}  // namespace cool
