/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#include <cool/signal/signal.hpp>
#include <cool/exception/os_error.hpp>
#include list_header
#include map_header
#include algorithm_header


namespace cool {


typedef STD_ list<signal::handler *> handler_list;
typedef STD_ map<signal::signal_type, handler_list> handler_map;
static handler_map handlers;


struct signal_procsig
{
  signal::signal_type s_;

  signal_procsig (signal::signal_type s): s_(s) {}

  void operator() (signal::handler *&ph) {
      (*ph)(s_);
  }
};


typedef void (*handler_type)(int);


static void
signal_handler (signal::signal_type s)
{
  signal::kill(s);
}


#define CHECK_RESULT(func,params) \
    do { \
        if (func params == -1) \
            raise(os_error(#func), COOL_LOCUS); \
    } while (0)

      
bool
signal::set (signal_type s, handler *h)
{ 
  assert(h != 0);

  handler_list &v = handlers[s];
  if (v.empty()) {
      struct sigaction sa;
      CHECK_RESULT(sigaction, (s, 0, &sa));

      if (sa.sa_handler != handler_type(signal_handler)) {
          // first time -- set our handler
          sa.sa_handler = handler_type(signal_handler);
          CHECK_RESULT(sigemptyset, (&sa.sa_mask));

          sa.sa_flags = 0;
          CHECK_RESULT(sigaction, (s, &sa, 0));
      }

      v.push_back(h);
      return true;
  }

  if (find(v.begin(), v.end(), h) == v.end()) {
      // Doesn't exist -- add
      v.push_back(h);
      return true;
  }

  // Already exists
  return false;
}


bool
signal::unset (signal_type s, handler *h)
{
  assert(h != 0);

  handler_list &v = handlers[s];
  handler_list::iterator i = find(v.begin(), v.end(), h);
  if (i != v.end()) {
      v.erase(i);
      if (v.empty())
          // Last our handler -- restore default handler
          unset(s);
      return true;
  }

  return false;
}


void
signal::unset (signal_type s)
{
  handler_list &v = handlers[s];
  v.erase(v.begin(), v.end());

  struct sigaction sa;
  CHECK_RESULT(sigaction, (s, 0, &sa));

  if (sa.sa_handler == handler_type(signal_handler)) {
      // Yes, this is our handler
      sa.sa_handler = SIG_DFL;
      CHECK_RESULT(sigemptyset, (&sa.sa_mask));

      sa.sa_flags = 0;
      CHECK_RESULT(sigaction, (s, &sa, 0));
  }
}


void
signal::mask (signal_type s)
{
  sigset_t set;

  CHECK_RESULT(sigemptyset, (&set));
  CHECK_RESULT(sigaddset, (&set, s));
  CHECK_RESULT(sigprocmask, (SIG_BLOCK, &set, 0));
}


void
signal::unmask (signal_type s)
{
  sigset_t set;

  CHECK_RESULT(sigemptyset, (&set));
  CHECK_RESULT(sigaddset, (&set, s));
  CHECK_RESULT(sigprocmask, (SIG_UNBLOCK, &set, 0));
}


void
signal::mask (signal_type a, signal_type b)
{
  struct sigaction sa;
  CHECK_RESULT(sigaction, (a, 0, &sa));

  if (sa.sa_handler != handler_type(signal_handler)) {
      sa.sa_handler = handler_type(signal_handler);
      CHECK_RESULT(sigemptyset, (&sa.sa_mask));
      sa.sa_flags = 0;
  }
  CHECK_RESULT(sigaddset, (&sa.sa_mask, b));
  CHECK_RESULT(sigaction, (a, &sa, 0));
}


void
signal::unmask (signal_type a, signal_type b)
{
  struct sigaction sa;
  CHECK_RESULT(sigaction, (a, 0, &sa));

  if (sa.sa_handler != handler_type(signal_handler)) {
      sa.sa_handler = handler_type(signal_handler);
      CHECK_RESULT(sigemptyset, (&sa.sa_mask));
      sa.sa_flags = 0;
  }
  else
      CHECK_RESULT(sigdelset, (&sa.sa_mask, b));

  CHECK_RESULT(sigaction, (a, &sa, 0));
}


bool
signal::is_pending (signal_type s)
{
  sigset_t set = pending();

  switch (sigismember(&set, s)) {
      case 0: return false;
      case 1: return true;
  }

  raise(os_error("sigismember"), COOL_LOCUS);

  // never reached
  return false;
}


sigset_t
signal::pending ()
{
  sigset_t set;
  CHECK_RESULT(sigemptyset, (&set));
  CHECK_RESULT(sigpending, (&set));

  return set;
}


void
signal::kill (signal_type s)
{
  handler_list &v = handlers[s];
  for_each(v.begin(), v.end(), signal_procsig(s));
}


}   // namespace cool
