/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#include <cool/time/time.hpp>
#include <cool/exception/exception.hpp>


namespace cool {
namespace bits {
namespace time {


struct tz_data
{
  tz_data () {
      ::tzset();
  }
};

static tz_data tz;


char *
as_iso8601 (sec_type sec, char *buf, size_t bufsz)
{
  struct tm tms;
#if defined(COOL_HAVE_LOCALTIME_R)
  localtime_r((time_t *)&sec, &tms);
#else
  memcpy(&tms, localtime((time_t *)&sec), sizeof(tms));
#endif

  size_t len = strftime(buf, bufsz, "%Y-%m-%dT%H:%M:%S", &tms);
  if (len) {
      if (cool::time::tz != 0)
          snprintf(buf + len, bufsz - len, "%+03d:%02d",
                  (int)-cool::time::tz/(60*60),
                  (int)cool::time::tz%(60*60));
      else if (len < bufsz - 2) {
          buf[len] = 'Z';
          buf[len + 1] = '\0';
      }
  }
  else
      // buffer too small
      *buf = '\0';

  return buf; 
}


sec_type
parse_iso8601 (const char *buf, char **eptr)
{
  // init time structure
  struct tm tms;
  tms.tm_isdst = -1;
  tms.tm_mday = 1; tms.tm_mon = tms.tm_year = 0;
  tms.tm_sec = tms.tm_min = tms.tm_hour = 0;
      
  // parse values
  bool done = false;
  char *eptr_;

  // year
  tms.tm_year = cool::as<int>(buf, &eptr_);
  if (!(done = (*eptr_ == '\0')) && *eptr_ != '-') {
      if (eptr)
          *eptr = eptr_;
      cool::raise<format_error>(COOL_LOCUS);
  }
  tms.tm_year -= 1900;

  // month
  if (!done) {
      tms.tm_mon = cool::as<int>(eptr_ + 1, &eptr_);
      if (!(done = (*eptr_ == '\0')) && *eptr_ != '-') {
          if (eptr)
              *eptr = eptr_;
          cool::raise<format_error>(COOL_LOCUS);
      }
      tms.tm_mon--;
  }

  // day
  if (!done) {
      tms.tm_mday = cool::as<int>(eptr_ + 1, &eptr_);
      if (!(done = (*eptr_ == '\0')) && *eptr_ != 'T') {
          if (eptr)
              *eptr = eptr_;
          cool::raise<format_error>(COOL_LOCUS);
      }
  }

  // hour
  if (!done) {
      tms.tm_hour = cool::as<int>(eptr_ + 1, &eptr_);
      if (!(done = (*eptr_ == '\0')) && *eptr_ != ':') {
          if (eptr)
              *eptr = eptr_;
          cool::raise<format_error>(COOL_LOCUS);
      }
  }

  // minute
  if (!done) {
      tms.tm_min = cool::as<int>(eptr_ + 1, &eptr_);
      done = (*eptr_ == '\0') ||
          (*eptr_ == 'Z') ||
          (*eptr_ == '+') ||
          (*eptr_ == '-');
      if (!done && *eptr_ != ':') {
          if (eptr)
              *eptr = eptr_;
          cool::raise<format_error>(COOL_LOCUS);
      }
  }

  // second
  if (!done)
      tms.tm_sec = cool::as<int>(eptr_ + 1, &eptr_);

  // create result time
  sec_type t = ::mktime(&tms);
  if (*eptr_ == 'Z')
      // UTC -> local
      t -= cool::time::tz;

  return t;
}


}   // namespace time
}   // namespace bits


const cool::time &cool::time::null = epoch();
const long &cool::time::tz = ::timezone;
const int &cool::time::ds = ::daylight;


}   // namespace cool
