/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#undef NDEBUG
#include <assert.h>

#if defined(INCLUDE_MODULE)
#  include <cool/thread.hpp>
#  include <cool/time.hpp>
#else
#  include <cool/thread/condition.hpp>
#  include <cool/thread/thread.hpp>
#  include <cool/thread/thread_list.hpp>
#  include <cool/time/time.hpp>
#endif
#include iostream_header


struct cond_test_args
{
  cond_test_args (): notified_(0), awoken_(0) { }

  cool::mutex m_;
  cool::condition c_;
  int notified_;
  int awoken_;
};


void
cond_test_thread (void *args)
{
  cond_test_args *data = static_cast<cond_test_args *>(args);
  cool::mutex::scoped_lock g(data->m_);
  assert((!g) == false);

  while (!(data->notified_ > 0))
      data->c_.wait(g);

  assert((!g) == false);
  data->awoken_++;
}


struct cond_pred
{
  int &var_;
  int &val_;

  cond_pred (int &var, int val): var_(var), val_(val) {}

  bool operator() () {
      return var_ == val_;
  }
};


void
cond_test_wait (void *args)
{
  cond_test_args *data = static_cast<cond_test_args *>(args);
  cool::mutex::scoped_lock g(data->m_);
  assert((!g) == false);

  // test wait
  while (data->notified_ != 1)
      data->c_.wait(g);
  assert((!g) == false);
  assert(data->notified_ == 1);
  data->awoken_++;
  data->c_.notify_one();

  // test predicate wait
  data->c_.wait(g, cond_pred(data->notified_, 2));
  assert((!g) == false);
  assert(data->notified_ == 2);
  data->awoken_++;
  data->c_.notify_one();

  // test timed wait
  while (data->notified_ != 3)
      data->c_.wait(g, cool::time::time(3));
  assert((!g) == false);
  assert(data->notified_ == 3);
  data->awoken_++;
  data->c_.notify_one();

  // test timed predicate wait
  cond_pred pred(data->notified_, 4);
  assert(data->c_.wait(g, cool::time::time(3), pred) == true);
  assert((!g) == false);
  assert(pred());
  assert(data->notified_ == 4);
  data->awoken_++;
  data->c_.notify_one();
}


void
test_notify_one ()
{
  cond_test_args data;
  cool::thread t(cond_test_thread, &data);

  {
      cool::mutex::scoped_lock g(data.m_);
      assert((!g) == false);
      data.notified_++;
      data.c_.notify_one();
  }

  t.join();
  assert(data.awoken_ == 1);
}


void
test_notify_all ()
{
  const int num_threads = 5;
  cool::thread_list threads;
  cond_test_args data;

  int i;
  for (i = 0;  i < num_threads;  ++i)
      threads.create(cond_test_thread, &data);

  {
      cool::mutex::scoped_lock g(data.m_);
      assert((!g) == false);
      data.notified_++;
      data.c_.notify_all();
  }

  threads.join_all();
  assert(data.awoken_ == num_threads);
}


void
test_waits ()
{
  cond_test_args data;
  cool::thread t(cond_test_wait, &data);

  {
      cool::mutex::scoped_lock g(data.m_);
      assert((!g) == false);

      int i;
      for (i = 1;  i < 5;  ++i) {
          cool::thread::sleep(cool::time::time(0, 5000));
          data.notified_++;
          data.c_.notify_one();
          while (data.awoken_ != i)
              data.c_.wait(g);
          assert((!g) == false);
          assert(data.awoken_ == i);
      }
  }

  t.join();
  assert(data.awoken_ == 4);
}


int
main (int argc, char *argv[])
{
  try {
      test_notify_one();
      test_notify_all();
      test_waits();
  }
  catch (const cool::exception &e) {
      STD_ cout << e << STD_ endl;
      return 1;
  }
  catch (STD_ exception &e) {
      STD_ cout << e.what() << STD_ endl;
      return 1;
  }

  return 0;
}
