/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#undef NDEBUG
#include <assert.h>

#if defined(INCLUDE_MODULE)
#  include <cool/net.hpp>
#  include <cool/time.hpp>
#else
#  include <cool/net/poller.hpp>
#  include <cool/net/inet_socket.hpp>
#  include <cool/time/time.hpp>
#endif
#include iostream_header


void
check_socket_status (cool::socket *s, cool::poller::iterator &i,
        bool r, bool w, bool e)
{
  assert(i == s);
  assert(s->readable() == r);
  assert(i->readable() == r);
  assert(s->writable() == w);
  assert(i->writable() == w);
  assert(s->has_exception() == e);
  assert(i->has_exception() == e);
}


int
main (int argc, char *argv[])
{
  char data[] = "hello\n";
  try {
      cool::poller poller;
      cool::poller::iterator i;
      cool::time::nsec_type timeout = 50000000;
      cool::inet::address a(cool::inet_address::loopback, 8192);

      // check that empty poller do nothing
      assert(poller.size() == 0);
      assert(poller.begin(cool::time::time()) == poller.end());

      // register server socket and check that it is not readable yet
      cool::inet::listener server(a);
      poller.insert(&server);
      i = poller.begin(cool::time::time(0, timeout));
      assert(!i == true);
      assert(i == poller.end());
      assert(i++ == poller.end());
      assert(++i == poller.end());

      // create connection to server socket
      cool::inet::socket client;
      assert(client.connect(a) == true);

      // server must be now readable
      i = poller.begin();
      assert(!i == false);
      assert(i != poller.end());
      check_socket_status(&server, i, true, false, false);
      assert(++i == poller.end());
      assert(i++ == poller.end());
      assert(!i == true);

      // accept connection and register both ends of communication channel
      cool::inet::socket channel = server.accept();
      poller.insert(&client, cool::socket::out);
      assert(poller.size() == 2);
      poller.insert(&channel);
      assert(poller.size() == 3);

      // client must be writable (and send data)
      i = poller.begin();
      assert(!i == false);
      assert(i != poller.end());
      assert(*i == client);
      check_socket_status(&client, i, false, true, false);
      assert(client.send(data, sizeof(data) - 1) == sizeof(data) - 1);
      assert(++i == poller.end());
      assert(i++ == poller.end());
      assert(!i == true);

      // client & channel must be changed their status
      char buf[sizeof(data)];
      for (i = poller.begin();  i != poller.end();  ++i) {
          if (*i == client)
              check_socket_status(&client, i, false, true, false);
          else if (*i == channel) {
              check_socket_status(&channel, i, true, false, false);
              int length = channel.recv(buf, sizeof(buf));
              assert(length == sizeof(data) - 1);
              assert(strncmp(data, buf, length) == 0);
          }
          else
              assert(!"Unknown socket returned!");
      }
      assert(++i == poller.end());
      assert(i++ == poller.end());
      assert(!i == true);

      // remove and close client
      poller.erase(&client);
      assert(poller.size() == 2);
      i = poller.begin(cool::time::time(0, timeout));
      assert(i == poller.end());
      client.close();

      // channel must now be readable but have no data
      i = poller.begin();
      assert(!i == false);
      assert(i != poller.end());
      assert(*i == channel);
      check_socket_status(&channel, i, true, false, false);
      assert(channel.recv(buf, sizeof(buf)) == 0);
      poller.erase(i);
      assert(++i == poller.end());
      assert(i++ == poller.end());
      assert(!i == true);
      assert(poller.size() == 1);
      channel.close();

      // unregister and close server socket also
      poller.erase(&server);
      assert(poller.size() == 0);
      server.close();

      // check again that empty poller do nothing
      assert(poller.size() == 0);
      assert(poller.begin(cool::time::time()) == poller.end());
  }
  catch (const cool::os_error &e) {
      STD_ cout << e << STD_ endl;
      return 1;
  }
  catch (const cool::exception &e) {
      STD_ cout << e << STD_ endl;
      return 1;
  }
  catch (STD_ exception &e) {
      STD_ cout << e.what() << STD_ endl;
      return 1;
  }

  return 0;
}
