/* Copyright (c) 2003-2005, Sven Suursoho
 * See LICENSE for details */

#undef NDEBUG
#include <assert.h>

#if defined(INCLUDE_MODULE)
#  include <cool/signal.hpp>
#  include <cool/net.hpp>
#else
#  include <cool/signal/signal.hpp>
#  include <cool/net/socketbuf.hpp>
#  include <cool/net/inet_socket.hpp>
#  include <cool/net/local_socket.hpp>
#endif
#include iostream_header


template <typename family>
void
test_socketbuf_of (const typename family::address &a)
{
  char *write_buf[] = {
      "11111111",
      "222222",
      "3333",
      "444444",
      "55555555"
  };
  int write_count = sizeof(write_buf)/sizeof(write_buf[1]);

  typedef typename cool::socketbuf<family> socketbuf;
  typedef typename socketbuf::listener_type listener;

  // I/O timeout (msec)
  const int timeout = 5;

  // create listening server
  listener server(a);

  // create client and connect to listening server
  socketbuf client;
  assert(client.recv_timeout(timeout).recv_timeout() == timeout);
  assert(client.send_timeout(timeout).send_timeout() == timeout);
  assert(client.throw_timeout(true).throw_timeout() == true);

  // use external buffer
  char buffer[1024];
  client.pubsetbuf(buffer, sizeof(buffer));

  assert(client.is_open() == false);
  assert(client.open(a) != 0);
  assert(client.is_open() == true);

  // accept i.e. create communication channel
  assert(server.wait(cool::socket::in).readable() == true);
  socketbuf channel(server.accept());
  assert(channel.recv_timeout(timeout).recv_timeout() == timeout);
  assert(channel.send_timeout(timeout).send_timeout() == timeout);
  assert(channel.throw_timeout(true).throw_timeout() == true);

  // do send/receive test(s) multiple times and both ways
  socketbuf *c[2];
  c[0] = &client;
  c[1] = &channel;
  for (int i = 0;  i < write_count;  ++i) {
      STD_ streamsize write_size = strlen(write_buf[i]);

      // switch sender / receiver sides
      int s = i%2, r = !s;

      // test sending data
      assert(c[s]->sputc(write_buf[i][0]) == write_buf[i][0]);
      assert(c[s]->sputn(write_buf[i] + 1, write_size - 1) == write_size - 1);
      assert(c[s]->pubsync() == 0);

      // test receiving data
      char read_buf[1024];
      assert(c[r]->sgetc() == write_buf[i][0]);
      assert((read_buf[0] = c[r]->sbumpc()) == write_buf[i][0]);
      assert(c[r]->sgetn(read_buf + 1, write_size - 1) == write_size - 1);
      assert(strncmp(read_buf, write_buf[i], write_size) == 0);

      try {
          // check timeout during read
          assert(c[r]->sgetc() == EOF);
          assert(c[r]->sgetn(read_buf, sizeof(read_buf)) == 0);
      }
      catch (cool::timeout_error &e) {}
  }

  // check connection error
  client.close();
  try {
      channel.sgetc();
      assert(!"connection_error should be thrown!");
  }
  catch (cool::connection_error &e) {}
  channel.close();
  server.close();
}


int
main (int argc, char *argv[])
{
  cool::signal::mask(cool::signal::pipe);
  try {
      // test local (i.e. unix) socket
      const char *local_socket_name = "./my_socket";
      remove(local_socket_name);
      test_socketbuf_of<cool::local>(local_socket_name);

      // test inet socket
      test_socketbuf_of<cool::inet>("localhost:8192");
  }
  catch (cool::exception &e) {
      STD_ cout << e << STD_ endl;
      return 1;
  }
  catch (STD_ exception &e) {
      STD_ cout << e.what() << STD_ endl;
      return 1;
  }

  return 0;
}
