#ifndef common_protocol_hpp_included
#define common_protocol_hpp_included

/**
 * \file common/protocol.hpp
 * \brief Communication protocol between proxy-db and connection pooler.
 */


#include <cool/net/inet_socket.hpp>
#include <cool/net/local_socket.hpp>
#include <string>
#include <limits>
#include <iostream>
#include <iomanip>


namespace common
{


typedef cool::local net_space;


struct net_string
{
  uint32_t length;  // in network format
  uint32_t data[1]; // actual length in length
};


/** \brief Communication protocol between PL/Proxy and connection pooler */
class protocol
{
  public:


      /** \brief Rendevouz point. common */
      static net_space::address rendevouz;


      /**
       * \brief Is listener waiting for incoming data? Otherwise wait channel
       * to become writable
       */
      net_space::socket::event_type wait_events;


      /** \brief I/O buffer */
      enum { default_bufsz = 64*1024 };
      size_t size_max;
      uint8_t *data; //[size_max + 1 /* for safety */];
      uint32_t bptr, eptr;


      /** \brief Communication channel */
      net_space::socket &channel;


      /** \brief Create new protocol object belonging to \a owner */
      protocol (net_space::socket &owner, size_t bufsize = default_bufsz)
          : size_max(bufsize), bptr(0), eptr(0), channel(owner)
      {
          data = new uint8_t[size_max + 1 /*safety*/];
          wait_readable();
      }


      ~protocol () {
          delete [] data;
      }


      /** \brief Return whether channel socket is connected */
      bool is_connected () const {
          return channel.is_connected();
      }


      /** \brief Wait for new requests */
      void wait_readable () {
          eptr = 0;
          wait_events = channel.in;
      }


      /** \brief Request handled. Wait until whole response is sent */
      void wait_writable () {
          bptr = 0;
          wait_events = channel.out;
      }


      /**
       * \brief Return current writing pointer in data buffer
       *
       * more() shows how much more data have to be written
       */
      uint8_t *begin () {
          return data + bptr;
      }


      /**
       * \brief Return pointer to end of data area in buffer i.e. free area
       *
       * free() shows how much more data can be stored
       */
      uint8_t *end () {
          return data + eptr;
      }


      /** \brief Return size of free area in buffer */
      uint32_t free () const {
          return size_max - eptr;
      }


      /** \brief Return size of writable area in buffer */
      uint32_t more () const {
          return eptr - bptr;
      }


      /**
       * \brief Advance read pointer by \a r bytes and return true if we have
       * now full request.
       *
       * server
       */
      bool recvd (uint32_t r);


      /**
       * \brief Advance write pointer by \a w bytes and return true if whole
       * response is sent
       */
      bool sent (uint32_t w);


      /** \brief Return true if buffer is full but no request -- overflow */
      bool full () const {
          return eptr == size_max;
      }


      /** \brief Daemon status */
      enum pooler_status_type {
          offline     /**< Offline or not initialized yet */
          , online      /**< Online, working */
          , locked      /**< Reconfiguring */
      };


      /** \brief Handshake packet */
      // server
      void send_hello (pooler_status_type status, const std::string &name);
      // client
      void recv_hello (pooler_status_type &status, std::string &name);


      enum command_type {
          cmd_invalid = 0
          , cmd_get_info
          , cmd_set_status
          , cmd_exec_query
          , cmd_last
      };


      /** \brief Return last request's type. server */
      command_type last_command () const {
          assert(eptr > 0);
          return (command_type)data[0];
      }


      enum response_type {
          rsp_ok                // success (optional data follow)
          , rsp_invalid         // invalid DB request
          , rsp_enotimpl        // feature not implemented yet
          , rsp_enores          // no resources to handle request
          , rsp_enoconn         // not connected
          , rsp_enocluster      // cluster not found
          , rsp_enopartition    // partition not found
      };


      static const char *as_string (response_type r);


      /** \brief Set response for last command. server */
      void response (response_type r) {
          data[0] = (uint8_t)r;
          eptr = sizeof(uint8_t);
          wait_writable();
      }

      /** \brief Get pool statistics. client */
      std::string recv_statistics ();


      /** \brief Send pool statistics. server */
      void send_statistics (const char *info, uint32_t length);


      /** \brief Set new status */
      response_type set_status (pooler_status_type &status);


      /** \brief Get new status sent by client */
      pooler_status_type new_status () const {
          assert(last_command() == cmd_set_status);
          return (pooler_status_type)data[1];
      }


      struct query_type
      {
          std::string &query;
          std::string &cluster;
          std::string &conf_map;
          uint16_t &hash_nr;

          query_type (std::string &query
                  , std::string &cluster
                  , std::string &conf_map, uint16_t &hash_nr)
              : query(query)
              , cluster(cluster)
              , conf_map(conf_map)
              , hash_nr(hash_nr)
          { }
      };


      /** \brief Send query to server. client */
      void send_query (const query_type &q);


      /** \brief Receive query from proxydb. server */
      void recv_query (query_type &q);


      /** \brief Send query result header. server */
      void send_result (response_type r,  int32_t rows, int32_t cols);


      /** \brief Receive query result header. client */
      response_type recv_result (int32_t &rows, int32_t &cols);


  // protocol base I/O
  // TODO: optimize client side (accumulate sending data in buffer and send as
  // one packet)
  public:

      protocol &
      do_send (const char *s, uint32_t len);

      protocol &
      send (const char *s, uint32_t len);

      template <typename T>
      protocol &
      send (const T &value);

      protocol &
      send_null();

      template <typename T>
      T recv ();

      void flush () {}
};


template <>
inline protocol &
protocol::send<uint32_t> (const uint32_t &value)
{
  uint32_t v = htonl(value);
  return do_send((char *)&v, sizeof(v));
}


template <>
inline protocol &
protocol::send<int32_t> (const int32_t &value)
{
  int32_t v = htonl(value);
  return do_send((char *)&v, sizeof(v));
}


template <>
inline protocol &
protocol::send<uint16_t> (const uint16_t &value)
{
  uint16_t v = htons(value);
  return do_send((char *)&v, sizeof(v));
}


template <>
inline protocol &
protocol::send<uint8_t> (const uint8_t &value)
{
  uint8_t v = value;
  return do_send((char *)&v, sizeof(v));
}


template <>
inline protocol &
protocol::send<std::string> (const std::string &value)
{
  return send(value.c_str(), value.length());
}


inline protocol &
protocol::send_null()
{
  return send(std::numeric_limits<uint32_t>::max());
}


inline protocol &
protocol::send (const char *value, uint32_t len)
{
  send(len);
  return do_send(value, len);
}


inline protocol &
protocol::do_send (const char *s, uint32_t len)
{
  if (len == 0)
      // nothing to send
      return *this;
  int l = channel.send((char *)s, len);
  if (l <= 0)
      // FIXME: other end is blocking
      cool::raise<cool::connection_error>(COOL_LOCUS);

  return *this;
}


template <>
inline uint32_t
protocol::recv<uint32_t> ()
{
  uint32_t v;
  int len = channel.recv((char *)&v, sizeof(v));
  if (!len)
      cool::raise<cool::connection_error>(COOL_LOCUS);
  return ntohl(v);
}

template <>
inline int32_t
protocol::recv<int32_t> ()
{
  int32_t v;
  int len = channel.recv((char *)&v, sizeof(v));
  if (!len)
      cool::raise<cool::connection_error>(COOL_LOCUS);
  return ntohl(v);
}


template <>
inline uint16_t
protocol::recv<uint16_t> ()
{
  uint16_t v;
  int len = channel.recv((char *)&v, sizeof(v));
  if (!len)
      cool::raise<cool::connection_error>(COOL_LOCUS);
  return ntohs(v);
}


template <>
inline uint8_t
protocol::recv<uint8_t> ()
{
  uint8_t v;
  int len = channel.recv((char *)&v, sizeof(v));
  if (!len)
      cool::raise<cool::connection_error>(COOL_LOCUS);
  return v;
}


template <>
inline std::string
protocol::recv<std::string> ()
{
  uint32_t len = recv<uint32_t>();
  if (len == std::numeric_limits<uint32_t>::max())
      // ugly trick to let caller know that null received
      throw uint32_t(0);

  std::string s;
  s.resize(len);

  // read until we have all data
  uint32_t read = 0;
  while (read < len) {
      uint32_t r = channel.recv((char *)s.data() + read, len - read);
      if (!r)
          cool::raise<cool::connection_error>(COOL_LOCUS);
      read += r;
  }

  return s;
}


}   // namespace common


#endif  // common_protocol_hpp_included
