#include <common/protocol.hpp>
#include <netinet/in.h>


namespace common
{


void
protocol::send_hello (pooler_status_type status, const std::string &name)
{
  // send status
  send<uint8_t>(status);
  if (status == offline)
      // send pooler name
      send(name);
  flush();
}


bool
protocol::recvd (uint32_t r)
{
  eptr += r;

  bool bad_cmd = false;
  uint32_t expect_size = 0;

  // calculate number of bytes we expect for specific request
  switch (last_command()) {
      case cmd_set_status:
          expect_size = sizeof(uint8_t[2]);
          break;

      case cmd_get_info:
          expect_size = sizeof(uint8_t);
          break;

      case cmd_exec_query: {
          // cmd_exec_query
          expect_size += sizeof(uint8_t);
          if (eptr < expect_size)
              return false;

          // cluster name
          expect_size += sizeof(uint32_t) +
              ntohl(*(uint32_t *)(data + expect_size));
          if (eptr < expect_size)
              return false;

          // conf_map
          expect_size += sizeof(uint32_t) +
              ntohl(*(uint32_t *)(data + expect_size));
          if (eptr < expect_size)
              return false;

          // hash_nr
          expect_size += sizeof(uint16_t);
          if (eptr < expect_size)
              return false;

          // query
          expect_size += sizeof(uint32_t) +
              ntohl(*(uint32_t *)(data + expect_size));
          if (eptr < expect_size)
              return false;

          break;
      }

      default:
          bad_cmd = true;
          break;
  }

  if (bad_cmd || expect_size < eptr) {
      // bad command or overflood
      data[0] = (uint8_t)cmd_invalid;
      return true;
  }
  else if (expect_size == eptr)
      // we have full request
      // mark that we don't want to read until request is fully handled
      wait_events = channel.no_events;

  // return false if we have no enough data yet
  return expect_size == eptr;
}


bool
protocol::sent (uint32_t w)
{
  bptr += w;
  if (bptr == eptr) {
      // all sent -- request handled
      wait_readable();
      return true;
  }

  return false;
}


void
protocol::recv_query (query_type &q)
{
  assert(wait_events == channel.no_events);

  // skip cmd_exec_query
  bptr = sizeof(uint8_t);

  // cluster_name
  uint32_t len = ntohl(*(uint32_t *)(data + bptr));
  bptr += sizeof(len);
  q.cluster.assign((const char *)data + bptr, len);
  bptr += len;

  // conf_map
  len = ntohl(*(uint32_t *)(data + bptr));
  bptr += sizeof(len);
  q.conf_map.assign((const char *)data + bptr, len);
  bptr += len;

  // hash_nr
  q.hash_nr = ntohs(*(uint16_t *)(data + bptr));
  bptr += sizeof(q.hash_nr);

  // query
  len = ntohl(*(uint32_t *)(data + bptr));
  bptr += sizeof(len);
  q.query.assign((const char *)data + bptr, len);
}


void
protocol::send_result (response_type r,  int32_t rows, int32_t cols)
{
  send<uint8_t>(r);
  send(rows);
  send(cols);
}


void
protocol::send_statistics (const char *info, uint32_t length)
{
  send<uint8_t>(rsp_ok);
  send(info, length);
  wait_readable();
}


}   // namespace common
