#include <db/func.hpp>
#include <db/error.hpp>
#include <db/pool.hpp>
#include <db/serial.hpp>
#include <boost/lexical_cast.hpp>
#include <string.h>
#include <map>

// PostgreSQL headers
extern "C" {
#include <db/hide_cpp_keywords.hpp>
#include <funcapi.h>
#include <catalog/pg_proc.h>
#include <catalog/pg_type.h>
#include <catalog/pg_namespace.h>
#include <utils/array.h>
#include <utils/builtins.h>
#include <utils/memutils.h>
#include <utils/lsyscache.h>
#include <db/show_cpp_keywords.hpp>
}


namespace db
{


// function cache: Oid <--> function
typedef std::map<Oid, function *> function_map;
static function_map func_cache;


void
function::init ()
{
  func_cache.clear();
}


void
function::reset ()
{
  profiler::function f("function::reset");

  // cluster
  cluster_name.clear();

  // partition by
  if (hash_plan) {
      int rc = SPI_freeplan(hash_plan);
      if (rc < 0)
          elog(FATAL, LOCUS"%d:SPI %s (freeplan)\n"
               , error::spi
               , SPI_result_code_string(rc));
  }
  hash_plan = 0;
  hash_func.clear();
  hash_fields.clear();

  // run on all
  run_on_all = single_transaction = false;

  // serial
  serial_next.clear();
  serial_parameter = 0;
  if_null = false;

  reset_current_call();
}


void
function::dump (PG_FUNCTION_ARGS)
{
  if (!plproxy_debug)
      return;

  // cluster
  dprint(LOCUS"%s:\n  cluster %s;\n",
          name.c_str(), cluster_name.c_str());

  if (run_on_all) {
      // run on all
      dprint("  run on all");
      if (single_transaction)
          dprint("\n    as single transaction");
      dprint(";\n");
  }
  else {
      // partition by
      hash_parameter_list::const_iterator i = hash_fields.begin();
      dprint("  partition by %s($%d", hash_func.c_str(), *i);
      while (++i != hash_fields.end())
          dprint(", $%d", *i);
      dprint(");\n");
  }

  // serial
  if (!serial_next.empty())
      dprint("  serial %s\n    replace $%d%s;\n"
              , serial_next.c_str()
              , serial_parameter
              , (if_null ? " if null" : "")
            );

  // show call
  dprint("\n"LOCUS"select * from %s(", name.c_str());
  parameter_list::const_iterator pi;
  int pno;
  for (pno = 0, pi = parameters.begin(); pi != parameters.end(); ++pi, ++pno) {
      if (pi != parameters.begin())
          dprint(", ");
      if (fcinfo->argnull[pno]) {
          dprint("<null>");
          continue;
      }
      char *as_string = parameters[pno].as_string(fcinfo->arg[pno]);
      dprint("<%s>", as_string);
      pfree(as_string);
  }
  dprint(")\n");
}


function *
function::get (PG_FUNCTION_ARGS)
{
  profiler::function f("function::get");

  // fetch target procedure tuple
  Oid fnoid = fcinfo->flinfo->fn_oid;
  HeapTuple fntup = SearchSysCache(PROCOID, ObjectIdGetDatum(fnoid), 0, 0, 0);
  if (!HeapTupleIsValid(fntup))
      elog(ERROR, LOCUS"%d:SYS_CACHE %u", error::sys_cache, fnoid);

  // See if there's already a cache entry for the current FmgrInfo. If not,
  // try to find one in the hash table
  function *fn = (function *)fcinfo->flinfo->fn_extra;
  function_map::iterator fi = func_cache.end();

  // if not in FmgrInfo, search in cache
  if (!fn && (fi = func_cache.find(fnoid)) != func_cache.end())
      // got from cache
      fn = fi->second;

  // if got from FmgrInfo or cache, check if still valid
  function *old_fn = 0;
  if (fn && !(fn->transaction_ == HeapTupleHeaderGetXmin(fntup->t_data) &&
              fn->command_ == HeapTupleHeaderGetCmin(fntup->t_data))) {
      // invalidated

      // Do not free 'fn' as we can reuse data structure for new function --
      // less allocations
      old_fn = fn;
      fn = 0;

      if (fi != func_cache.end()) {
          // if in cache, remove
          func_cache.erase(fi);
          fi = func_cache.end();
      }
  }

  if (!fn) {
      // not found or out-of-date
      // tough, we have to load function and compile
      if (old_fn)
          // reuse old storage
          fn = old_fn;

      create(fntup, fn, fcinfo);

      // insert into cache for future references
      std::pair<function_map::iterator, bool> ins_result =
          func_cache.insert(std::make_pair(fnoid, fn));

      // weird, it should succeed but didn't -- drop dead
      Assert(ins_result.second == true);
  }

  ReleaseSysCache(fntup);

  // save procedure pointer in FmgrInfo to avoid search on subsequent calls
  // NB! because fn_extra is in use, we must handle itself SRF (set returning
  // functions).
  fcinfo->flinfo->fn_extra = (void *)fn;

  return fn;
}


void
function::create (const HeapTuple &fntup, function *&fn, PG_FUNCTION_ARGS)
{
  profiler::function f("function::create");

  // make sure we have function structure, whether reuse old or create new
  if (!fn) {
      if ((fn = new function()) == 0)
          throw std::bad_alloc();
  }
  fn->reset();

  PG_TRY(); {
      // fetch function source
      bool isnull;
      Datum srcdat =
          SysCacheGetAttr(PROCOID, fntup, Anum_pg_proc_prosrc, &isnull);
      if (isnull)
          // no source
          elog(ERROR, LOCUS"%d:SRC_NULL", error::src_null);

      // setup procedure information
      Form_pg_proc prostruct = (Form_pg_proc)GETSTRUCT(fntup);
      fn->transaction_ = HeapTupleHeaderGetXmin(fntup->t_data);
      fn->command_ = HeapTupleHeaderGetCmin(fntup->t_data);
      fn->name = NameStr(prostruct->proname);

      // get schema HeapTuple
      HeapTuple nsTup = SearchSysCache(NAMESPACEOID
          , ObjectIdGetDatum(prostruct->pronamespace), 0, 0, 0);
      if (!HeapTupleIsValid(nsTup))
          elog(ERROR, LOCUS"%d:SYS_CACHE %u", error::sys_cache
               , prostruct->pronamespace);
      Form_pg_namespace nsStruct = (Form_pg_namespace) GETSTRUCT(nsTup);
      fn->name = std::string(NameStr(nsStruct->nspname)) + '.' + fn->name;
      ReleaseSysCache(nsTup);

      // get result info
      Oid relid = fn->result.init(prostruct->prorettype, false);
      fn->is_tuple = relid != InvalidOid || prostruct->prorettype == RECORDOID;
      fn->is_set = prostruct->proretset;

      // get parameters info
      fn->parameters.clear();
      fn->parameters.resize(prostruct->pronargs);
      for (int i = 0;  i < prostruct->pronargs;  ++i)
          fn->parameters[i].init(prostruct->proargtypes[i], true);

      // get body as C string and compile
      char *source = DatumGetCString(DirectFunctionCall1(textout, srcdat));
      try {
          fn->compile(source);
          pfree(source);
      }
      catch (const syntax_error &e) {
          pfree(source);
          ereport(ERROR,
                  (errcode(ERRCODE_SYNTAX_ERROR),
                   errmsg(LOCUS"%d:SYNTAX: \"%s\"", error::syntax, e.what())));
      }

      if (!fn->hash_func.empty())
          // prepare hash func calling plan if specified
          fn->create_hash_plan();
  }
  PG_CATCH(); {
      // error creating function, free storage
      delete fn;
      fn = 0;
      PG_RE_THROW();
  }
  PG_END_TRY();
}


void
function::create_hash_plan ()
{
  profiler::function f("function::create_hash_plan");

  Assert(hash_func.empty() == false);

  // array of parameter type Oids
  Oid *argtypes = (Oid *)palloc(sizeof(Oid)*hash_fields.size());
  int argtypes_count = 0;

  // create plan and set first parameter as cluster_name
  std::string plan =
      "select conf_map, hash_nr"
      " from " + hash_func + "('" + cluster_name + '\'';

  // parameters
  function::hash_parameter_list::const_iterator hf;
  for (hf = hash_fields.begin();  hf != hash_fields.end();  ++hf) {
      argtypes[argtypes_count++] = parameters[*hf - 1].oid;
      plan += ",$" + boost::lexical_cast<std::string>(argtypes_count);
  }
  plan += ')';

  // prepare and save for future references
  if ((hash_plan = SPI_prepare(plan.c_str(), argtypes_count, argtypes)) == 0)
      elog(FATAL, LOCUS"%d:SPI %s (prepare)", error::spi
           , SPI_result_code_string(SPI_result));
  if ((hash_plan = SPI_saveplan(hash_plan)) == 0)
      elog(FATAL, LOCUS"%d:SPI %s (saveplan)", error::spi
           , SPI_result_code_string(SPI_result));

  pfree(argtypes);
}


function::hash_key_type
function::get_hash (PG_FUNCTION_ARGS)
{
  profiler::function f("function::get_hash");

  uint32_t attempts = 0;

retry:
  attempts++;
  if (hash_plan == 0)
      create_hash_plan();

  // prepare SPI_execute_plan parameters
  Datum *args = (Datum *)palloc(sizeof(Datum)*hash_fields.size());
  char *nulls = (char *)palloc(sizeof(char)*hash_fields.size());
  int ndx = 0;

  hash_parameter_list::const_iterator hf;
  for (hf = hash_fields.begin();  hf != hash_fields.end();  ++hf, ++ndx) {
      int pno = *hf - 1;

      // check if argument is null
      if ((nulls[ndx] = fcinfo->argnull[pno] ? 'n' : ' ') == 'n')
          // yes, do not set argument
          continue;

      // set argument
      args[ndx] = fcinfo->arg[pno];
  }

  PG_TRY(); {
      // execute plan
      int rc = SPI_execute_plan(hash_plan, args, nulls, true, 1);
      if (rc < 0)
          elog(FATAL, LOCUS"%d:SPI %s (execute)", error::spi
               , SPI_result_code_string(SPI_result));
      Assert(SPI_processed == 1);
  }
  PG_CATCH(); {
      if (attempts == 2)
          // already tried to recreate get_hash SQL plan, didn't succeed
          PG_RE_THROW();

      // there is no way than leaking get_hash plan, because we have invalid
      // handle :(
      SPI_finish();
      SPI_connect();
      hash_plan = 0;
      goto retry;
  }
  PG_END_TRY();

  // fetch hash function calling result
  bool isnull = false;
  hash_key_type hash_value;

  // conf_map
  Datum d = SPI_getbinval(
          SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 1, &isnull);
  if (isnull)
      elog(ERROR, LOCUS"%d:NULL_HASH", error::null_hash);
  text *confmap = DatumGetTextP(d);
  hash_value.first = std::string(VARDATA(confmap), VARSIZE(confmap) - VARHDRSZ);

  // hash_nr
  hash_value.second = DatumGetUInt16(SPI_getbinval(
              SPI_tuptable->vals[0], SPI_tuptable->tupdesc, 2, &isnull));
  if (isnull)
      elog(ERROR, LOCUS"%d:NULL_HASH", error::null_hash);

  SPI_freetuptable(SPI_tuptable);
  pfree(args);
  pfree(nulls);

  return hash_value;
}


Datum
function::build_return_datum (const char *as_string, PG_FUNCTION_ARGS)
{
  profiler::function f("function::build_return_datum");

  // void
  if (result.oid == VOIDOID) {
      // void
      fcinfo->isnull = false;
      return (Datum)0;
  }

  // normal scalar or tuple
  Datum rv = result.as_datum(as_string);
  if (!result.by_value) {
      // not by-value, copy the value into upper executor memory context
      Size len = datumGetSize(rv, false, result.type_length);
      void *tmp = (void *)SPI_palloc(len);
      memcpy(tmp, DatumGetPointer(rv), len);
      rv = PointerGetDatum(tmp);
  }

  return rv;
}


Datum
function::exec_on (hash_key_type partno, PG_FUNCTION_ARGS)
{
  profiler::function f("function::exec_on");

  if (!is_set || set_row_no == -1) {
      // query is not sent yet, do it

      // create query string
      std::string q = "select r from " + name + '(';
      unsigned int i;
      parameter_list::iterator p;
      for (i = 0, p = parameters.begin();  p != parameters.end();  ++p, ++i) {
          if (i > 0)
              q += ',';

          if (serial_parameter - 1 != i) {
              // ordinary parameter
              if (PG_ARGISNULL(i))
                  q += "null";
              else {
                  char *s = p->as_string(fcinfo->arg[i]);
                  q += s;
                  pfree(s);
              }
          }
          else {
              // serial 'serial_next' replace $i replace if null
              int64_t next = (!if_null || PG_ARGISNULL(i))
                  ? serial::next(serial_next)
                  : PG_GETARG_INT64(i);
              q += boost::lexical_cast<std::string>(next);
          }
      }
      q += ") r;";

      // send
      pool->send_query(q, cluster_name, partno.first, partno.second);

      // wait for initial result
      int32_t cols;
      protocol::response_type r = pool->recv_result(set_row_count, cols);
      switch (r) {
          case protocol::rsp_ok:
              // result is ok
              break;

          default:
              // error or smth
              // TODO: error reporting
              elog(ERROR, LOCUS"%d:EPOOL %s\n"
                   , r + error::pool_error, protocol::as_string(r));
      }
  }

  if (!is_set) {
      // one-shot call
      try {
          std::string result = pool->recv_result_field();
          Datum rv = build_return_datum(result.c_str(), fcinfo);
          return rv;
      }
      catch (const uint32_t &) {
          // if field is null, recv_result_field throws uint32_t(0);
          fcinfo->isnull = true;
          return (Datum)0;
      }
  }

  // set returning function
  ReturnSetInfo *rsi = (ReturnSetInfo *)fcinfo->resultinfo;
  if (++set_row_no == 0) {
      // first row -- do checks and setup
      if (!rsi || !IsA(rsi, ReturnSetInfo) ||
              (rsi->allowedModes & SFRM_ValuePerCall) == 0)
          ereport(ERROR,
                  (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                   errmsg(LOCUS"%d:SET_CONTEXT", error::set_context)));
      rsi->returnMode = SFRM_ValuePerCall;
  }

  if (set_row_no == set_row_count) {
      // no more rows in result set
      reset_current_call();
      rsi->isDone = ExprEndResult;
      fcinfo->isnull = true;
      return 0;
  }

  // signal that there is more rows coming
  rsi->isDone = ExprMultipleResult;

  // build next row
  try {
      std::string result = pool->recv_result_field();
      Datum rv = build_return_datum(result.c_str(), fcinfo);
      return rv;
  }
  catch (const uint32_t &) {
      // if field is null, recv_result_field throws uint32_t(0)
      fcinfo->isnull = true;
      return (Datum)0;
  }
}


Datum
function::execute (PG_FUNCTION_ARGS)
{
  profiler::function f("function::execute");

  // debug output
  //dump(fcinfo);

  if (!run_on_all) {
      // run on one specific partition
      hash_key_type hash_value;

      if (!is_set || set_row_no == -1)
          // hash function not called yet
          hash_value = get_hash(fcinfo);

      return exec_on(hash_value, fcinfo);
  }

  // run on all partitions
  elog(ERROR, LOCUS"%d:NOT_IMPL run_on_all()", error::not_impl);

  return (Datum)0;
}


}   // namespace db
