/**
 * \file db/func.hpp
 * \brief PL/Proxy language function stuff
 */

#ifndef db_func_hpp_included
#define db_func_hpp_included


#include <db/plproxy.hpp>
#include <db/typeinfo.hpp>
#include <vector>
#include <string>


extern "C" {
#include <db/hide_cpp_keywords.hpp>
#include <executor/spi.h>
#include <utils/syscache.h>
#include <db/show_cpp_keywords.hpp>
}


namespace db
{


/** \brief Class representing PL/Proxy function */
class function
{
  public:

      /** \brief Initialize db::function subsystem */
      static void init ();


      /** \brief Create or fetch from cache PL/Proxy function */
      static function *get (PG_FUNCTION_ARGS);


      /** \brief Dump function's content */
      void dump (PG_FUNCTION_ARGS);


      /** \brief Run this function */
      Datum execute (PG_FUNCTION_ARGS);


      /** \brief Reset procedure's variables */
      void reset ();


      /** \brief PL/Proxy procedure name */
      std::string name;


      /** \brief Procedure result type info */
      typeinfo result;
      bool is_tuple;
      bool is_set;


      /** \brief Procedure arguments (array of types). */
      typedef std::vector<typeinfo> parameter_list;
      parameter_list parameters;

      /** \brief Cluster name */
      std::string cluster_name;

      /** \brief Saved hashfunc calling plan (if not run on all) */
      void *hash_plan;

      /** \brief Hashfunc name */
      std::string hash_func;

      /**
       * \brief Indexes of parameters (in parameters[]) used to
       * calculate hash (1 as base)
       */
      typedef std::vector<int> hash_parameter_list;
      hash_parameter_list hash_fields;

      bool run_on_all;
      bool single_transaction;

      std::string serial_next;
      unsigned int serial_parameter;
      bool if_null;


      // Following attributes are used during actual function call

      // SRF: number of rows already returned
      int32_t set_row_count;   // number of rows in set
      int32_t set_row_no;      // number of rows already returned


  private:


      /**
       * \brief Create new function.
       *
       * Use static method get() instead of direct instantation because that
       * does everything (cache, parse, compile etc).
       */
      function () {
          // must be initialized before call to reset
          hash_plan = 0;
      }


      /**
       * \brief Compile PL/Proxy function from source \a src.
       *
       * Because language structure is so simple, we just remember some stuff
       * extracted from source.
       *
       * \a src is destroyed as it is parsed using strtok(). But it is ok,
       * because we won't use it any more.
       */
      void compile (const char *src);


      /** \brief Prepare hash function calling plan. Called from create() */
      void create_hash_plan ();


      /** \brief Return hash calculated by fields specified in partition by */
      typedef std::pair<std::string, uint16_t> hash_key_type;
      hash_key_type get_hash (PG_FUNCTION_ARGS);


      /** \brief Build return datum */
      Datum build_return_datum (const char *as_string, PG_FUNCTION_ARGS);


      /** \brief Run query on partition \a partno */
      Datum exec_on (hash_key_type partno, PG_FUNCTION_ARGS);


      /** \brief Reset current function call variables (for SRF) */
      void reset_current_call () {
          set_row_count = set_row_no = -1;
      }


      /**
       * \brief Create new function. If \a fn is not 0, then reuse existing
       * data structure.
       */
      static void create (const HeapTuple &fntup, function *&fn,
              PG_FUNCTION_ARGS);

      TransactionId transaction_;
      CommandId command_;
      std::string query;
};


}   // namespace db


#endif  // db_func_hpp_included
