// Because parser has great potential to grow bigger and bigger it is moved to
// separate source file

#include <db/error.hpp>
#include <db/func.hpp>

#include <boost/spirit/core.hpp>
#include <boost/spirit/actor/assign_actor.hpp>
#include <boost/spirit/utility/confix.hpp>


namespace db
{



/* PL/Proxy language description:
 
cluster /CLUSTER_NAME/;

-- if neither specified 'run on all' is assumed
[
  { partition by /hash_func/($/PARAMETER_NUMBER/ [, $/PARAMETER_NUMBER/]); }
  |
  { run on all [ as single transaction ]; }
]

-- if 'if null' not specified, parameter PARAMETER_NUMBER is unconditionally
-- replaced otherwise only if specified parameter is null
[ serial /SEQUENCE_NAME/ replace $/PARAMETER_NUMBER/ [ if null ] ; ]

*/


using namespace boost::spirit;


struct plproxy_parser: public grammar<plproxy_parser>
{
  function &f;

  plproxy_parser (function &f_): f(f_) { }

  // grammar definition
  template <typename ScannerT>
  struct definition
  {
      definition (const plproxy_parser &self) {
          unsigned int param_count = self.f.parameters.size();

          first = (

              grammar =
                  as_lower_d
                  [
                      cluster
                      >> !( run_all | partition )
                      >> !( serial )
                  ]

              , var_name =
                  lexeme_d
                  [
                      +( alnum_p | '_' | '.' )
                  ]

              , cluster =
                  "cluster"
                  >> var_name[assign(self.f.cluster_name)]
                  >> ';'

              , partition =
                  str_p("partition")
                  >> "by"
                  >> var_name[assign(self.f.hash_func)]
                  >> '('
                  >> (
                        '$'
                        >> limit_d(1u, param_count)
                            [uint_p][push_back_a(self.f.hash_fields)]
                        >> *(
                              ch_p(',')
                              >> '$'
                              >> limit_d(1u, param_count)
                                  [uint_p][push_back_a(self.f.hash_fields)]
                            )
                     )
                  >> ')'
                  >> ';'

              , run_all =
                  str_p("run")[assign_a(self.f.run_on_all, true)]
                  >> "on"
                  >> "all"
                  >> !(
                        str_p("as")[assign_a(self.f.single_transaction, true)]
                        >> "single"
                        >> "transaction"
                      )
                  >> ';'

              , serial =
                  "serial"
                  >> var_name[assign(self.f.serial_next)]
                  >> "replace"
                  >> '$'
                  >> min_limit_d(1u)[uint_p][assign(self.f.serial_parameter)]
                  >> !(
                        str_p("if")[assign_a(self.f.if_null, true)]
                        >> "null"
                      )
                  >> ';'
          );
      }


      rule<ScannerT> first;
      subrule<0> grammar;
      subrule<1> var_name;
      subrule<2> cluster;
      subrule<3> partition;
      subrule<4> run_all;
      subrule<5> serial;


      const rule<ScannerT> &start () const {
          return first;
      }
  };
};


void
function::compile (const char *src)
{
  profiler::function f("function::compile");

  // parse and set new function information
  plproxy_parser parser(*this);
  parse_info<const char *> result = parse(src, parser
          , space_p | comment_p("/*", "*/") | comment_p("--"));
  if (!result.full) {
      // thrown an exception with source line where parser stopped
      const char *begin = result.stop;
      const char *end = begin;
      while (*end && *end != '\n')
          end++;
      throw syntax_error(std::string(begin, end));
  }

  // if 'partition by' nor 'run on all' not specified
  if (hash_func.empty())
      run_on_all = true;
  if (!serial_next.empty())
      // little speedup for later -- because we don't use serial_next variable
      // name anymore, we can create SQL statement, assigning it to
      // serial_next and save ourselves to do it each time
      serial_next = "select nextval('" + serial_next + "')";
}


}   // namespace db
