#include <db/plproxy.hpp>
#include <db/error.hpp>
#include <db/func.hpp>
#include <db/pool.hpp>
#include <stdexcept>


extern "C"
{


// PostgreSQL headers
#include <db/hide_cpp_keywords.hpp>
#include <commands/trigger.h>
#include <utils/memutils.h>
#include <db/show_cpp_keywords.hpp>


bool plproxy_debug =
#if !defined(NDEBUG)
    true;
#else
    false;
#endif


PG_FUNCTION_INFO_V1(plproxy_call_handler);


void
perm_fmgr_info (Oid functionId, FmgrInfo *finfo)
{
  fmgr_info_cxt(functionId, finfo, TopMemoryContext);
}


// One time initialization stuff
static void
plproxy_init (void)
{
  // Do initialization only once
  static bool plproxy_initialized = false;
  if (plproxy_initialized)
      return;
  plproxy_initialized = true;

  // initialize subsystems

  // create pooler client
  init();

  // function cache
  db::function::init();
}


// Language handler
Datum
plproxy_call_handler (PG_FUNCTION_ARGS)
{
  db::profiler::function f("plproxy_call_handler");

  if (CALLED_AS_TRIGGER(fcinfo))
      // This language can't be called as trigger
      ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                  errmsg(LOCUS"%d:NO_TRIGGER trigger", db::error::not_impl)));

  try {
      // Do one time initialization
      plproxy_init();

      int rc;
      if ((rc = SPI_connect()) != SPI_OK_CONNECT)
          elog(FATAL, LOCUS"%d:SPI %s (connect)\n", db::error::spi
               , SPI_result_code_string(rc));

      // real action
      Datum rv;
      PG_TRY(); {
          db::function *fn = db::function::get(fcinfo);
          rv = fn->execute(fcinfo);
      }
      PG_CATCH(); {
          // reset current profiler call chain
          db::profiler::cancel_call_chain();
          PG_RE_THROW();
      }
      PG_END_TRY();

      if ((rc = SPI_finish()) != SPI_OK_FINISH)
          elog(FATAL, LOCUS"%d:SPI %s (finish)", db::error::spi
               , SPI_result_code_string(rc));

      return rv;
  }
  catch (const std::bad_alloc &e) {
      ereport(ERROR, (errcode(ERRCODE_OUT_OF_MEMORY),
                      errmsg(LOCUS"%d:NO_MEM", db::error::no_mem)));
  }
  catch (const cool::exception &e) {
      ereport(FATAL,
              (errcode(ERRCODE_INTERNAL_ERROR),
               errmsg(LOCUS"%d:GENERIC %s (%s)", db::error::generic
                      , e.what(), e.locus)));
  }
  catch (const std::exception &e) {
      ereport(FATAL,
              (errcode(ERRCODE_INTERNAL_ERROR),
               errmsg(LOCUS"%d:GENERIC %s", db::error::generic, e.what())));
  }

  // should never get here
  return (Datum)0;
}


void
init ()
{
  pool = new db::pooler_client();
}


void
done ()
{
  delete pool;
}


#include <stdio.h>
#include <stdarg.h>


void
dprint (const char *fmt, ...)
{
#if !defined(NDEBUG)
  if (!plproxy_debug)
      return;

  va_list ap;
  va_start(ap, fmt);
  vprintf(fmt, ap);
  va_end(ap);
  fflush(stdout);
#endif
}


}   // extern "C"
