#include <db/pool.hpp>
#include <db/error.hpp>
#include <db/typeinfo.hpp>
#include <sstream>


using namespace common;


extern "C" {

#include <db/hide_cpp_keywords.hpp>
#include <executor/spi.h>
#include <db/show_cpp_keywords.hpp>

db::pooler_client *pool = 0;
}


namespace db
{

// connnection to pooler
net_space::socket pooler_client::daemon;
protocol pooler_client::proto(pooler_client::daemon);


bool
pooler_client::connect ()
{
  dprint(LOCUS"reconnect (%s)\n", proto.rendevouz.c_str());

  try {
      if (daemon.handle() != -1)
          // already were connected, create new socket
          daemon.open(net_space::family);

      daemon.connect(proto.rendevouz);

      // success
      return true;
  }
  catch (const cool::connection_refused &e) {
      // do nothing, let's hope that can connect later
      elog(FATAL, LOCUS"%d:POOL_CONNECT (%s)"
           , error::pool_connect, e.locus);
  }
  catch (const cool::os_error &e) {
      elog(FATAL, LOCUS"%d:POOL_CONNECT %s (%s)"
           , error::pool_connect, e.detail(), e.locus);
  }
  catch (const cool::exception &e) {
      elog(FATAL, LOCUS"%d:POOL_CONNECT %s (%s)"
           , error::pool_connect, e.what(), e.locus);
  }
  catch (const std::exception &e) {
      elog(FATAL, LOCUS"%d:POOL_CONNECT %s", error::pool_connect, e.what());
  }

  return false;
}


void
pooler_client::disconnect ()
{
  daemon.close();
}


void
pooler_client::send_query (std::string &query
        , std::string &cluster
        , std::string &conf_map, uint16_t &hash_nr)
{
  profiler::function f("pooler_client::send_query");

#if !defined(NDEBUG)
  dprint("*** Q *** %s\n", query.c_str());
#endif

  for (int i = 0;  i < 2;  ++i) {
      try {
          protocol::query_type q(query, cluster, conf_map, hash_nr);
          proto.send_query(q);
          proto.flush();
          return;
      }
      catch (const cool::connection_error &e) {
          if (i == 0 && connect())
              // try again
              continue;

          // rethrow
          throw;
      }
  }
}


}   // namespace db



extern "C" {


PG_FUNCTION_INFO_V1(set_pool_status);

Datum
set_pool_status (PG_FUNCTION_ARGS)
{
  if (PG_ARGISNULL(0))
      elog(ERROR, LOCUS"invalid status");

  // get parameter
  text *st = PG_GETARG_TEXT_P(0);
  uint32_t stl = VARSIZE(st) - VARHDRSZ;

  // copy for strcasecmp
  char status_text[16];
  if (sizeof(status_text) - 1 < stl)
      elog(ERROR, LOCUS"invalid status");
  memcpy(status_text, VARDATA(st), stl);
  status_text[stl] = '\0';

  protocol::pooler_status_type status;
  if (strcasecmp(status_text, "LOCKED") == 0)
      status = protocol::locked;
  else if (strcasecmp(status_text, "ONLINE") == 0)
      status = protocol::online;
  else if (strcasecmp(status_text, "OFFLINE") == 0)
      status = protocol::offline;
  else
      elog(ERROR, LOCUS"invalid status");

  const char *rtxt = 0;
  for (int i = 0;  i < 2;  ++i) {
      try {
          protocol::response_type r = pool->proto.set_status(status);
          rtxt = protocol::as_string(r);
          break;
      }
      catch (const cool::connection_error &e) {
          if (i == 0 && pool->connect())
              // reconnect and try again
              continue;

          rtxt = "POOL_ENOCONN";
          break;
      }
  }

  if (!rtxt)
      elog(ERROR, LOCUS"unknown response");

  size_t len = strlen(rtxt);
  text *result = (text *)palloc(len + VARHDRSZ);
  VARATT_SIZEP(result) = len + VARHDRSZ;
  memcpy(VARDATA(result), rtxt, len);

  PG_RETURN_TEXT_P(result);
}


PG_FUNCTION_INFO_V1(get_pool_status);

Datum
get_pool_status (PG_FUNCTION_ARGS)
{
  const char *rtxt = "not connected";
  if (pool->proto.is_connected())
      rtxt = "connected";

  size_t len = strlen(rtxt);
  text *result = (text *)palloc(len + VARHDRSZ);
  VARATT_SIZEP(result) = len + VARHDRSZ;
  memcpy(VARDATA(result), rtxt, len);

  PG_RETURN_TEXT_P(result);
}


PG_FUNCTION_INFO_V1(profiler);

Datum
profiler (PG_FUNCTION_ARGS)
{
  if (PG_ARGISNULL(0))
      elog(ERROR, LOCUS"%d:NOT_IMPL", db::error::not_impl);

  text *ot = PG_GETARG_TEXT_P(0);
  uint32_t otl = VARSIZE(ot) - VARHDRSZ;

  char op[16];
  if (sizeof(op) - 1 < otl)
      elog(ERROR, LOCUS"%d:NOT_IMPL", db::error::not_impl);
  memcpy(op, VARDATA(ot), otl);
  op[otl] = '\0';

  const char *result = "";
  uint32_t len = 0;

  std::ostringstream os;
  if (strcasecmp(op, "ON") == 0) {
      // switch profiler data gathering on
      db::profiler::active(true);
      result = "ok";
      len = 2;
  }
  else if (strcasecmp(op, "OFF") == 0) {
      // switch profiler data gathering off
      db::profiler::active(false);
      result = "ok";
      len = 2;
  }
  else if (strcasecmp(op, "RESET") == 0) {
      // reset profiler data
      db::profiler::reset();
      result = "ok";
      len = 2;
  }
  else if (strcasecmp(op, "GET") == 0) {
      // get profiler data
      db::profiler::dump(os);
      result = os.str().c_str();
      len = os.str().length();
  }
  else
      elog(ERROR, LOCUS"%d:NOT_IMPL", db::error::not_impl);

  // convert to text and return
  text *r = (text *)palloc(len + VARHDRSZ);
  VARATT_SIZEP(r) = len + VARHDRSZ;
  memcpy(VARDATA(r), result, len);
  PG_RETURN_TEXT_P(r);
}


}
