#include <db/typeinfo.hpp>
#include <db/error.hpp>


extern "C" {
#include <db/hide_cpp_keywords.hpp>
#include <catalog/pg_type.h>
#include <utils/builtins.h>
#include <utils/syscache.h>
#include <utils/lsyscache.h>
#include <db/show_cpp_keywords.hpp>
#include <libpq-fe.h>
}   // extern "C"


namespace db
{


Oid
typeinfo::init (Oid typid, bool in)
{
  profiler::function f("typeinfo::init");

  oid = typid;
  if (oid == ANYELEMENTOID || oid == ANYARRAYOID)
      // we do not support polymorphic procedures
      elog(ERROR, LOCUS"%d:ANY", error::any);

  // lookup from system's cache
  HeapTuple tup = SearchSysCache(TYPEOID, ObjectIdGetDatum(oid), 0, 0, 0);
  if (!HeapTupleIsValid(tup))
      elog(ERROR, LOCUS"%d:SYS_CACHE %u", error::sys_cache, oid);

  // get tuple data
  Form_pg_type stru = (Form_pg_type)GETSTRUCT(tup);

  // find variable's type
  switch (stru->typtype) {
      case 'b': // base type
      case 'd': // domain
          kind = scalar;
          break;

      case 'c':
          kind = row;
          break;

      case 'p':
          if (oid != VOIDOID)
              ereport(ERROR,
                      (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
                       errmsg(LOCUS"%d:TYPE \"%s\""
                              , error::type, format_type_be(oid))));
          break;

      default:
          elog(ERROR, LOCUS"%d:TYPE \"%c\": unrecoginized type"
               , error::type, stru->typtype);
          break;
  }

  perm_fmgr_info((in ? stru->typoutput : stru->typinput), &io);
  io_param = getTypeIOParam(tup);
  by_value = stru->typbyval;
  type_length = stru->typlen;
  need_quotes = (type_length < 1 && oid != NUMERICOID)
      // types that have fixed length but must be quoted
      || (oid == BOOLOID)
      || (oid == NAMEOID)
      || (oid == CHAROID)
      || (oid == MACADDROID)
      || (oid == DATEOID)
      || (oid == TIMEOID)
      || (oid == TIMETZOID)
      || (oid == TIMESTAMPOID)
      || (oid == TIMESTAMPTZOID)
      || (oid == INTERVALOID)
      ;

  Oid relid = stru->typrelid;
  ReleaseSysCache(tup);

  return relid;
}


// TODO: optimize, too many strlen, memcpy
char *
typeinfo::as_string (Datum arg)
{
  profiler::function f("typeinfo::as_string");

  char *s = DatumGetCString(FunctionCall3(&io, arg,
              ObjectIdGetDatum(io_param), Int32GetDatum(-1)));

  /*if (oid == BYTEAOID) {
      // specific case of need_quotes: do binary string escape
      size_t len = 0;
      unsigned char *es =
          PQescapeBytea((const unsigned char *)s, strlen(s), &len);

      // free old string, allocate enough for new and copy escaped version
      s = (char *)repalloc(s, len + 2);
      memcpy(s + 1, es, len);
      s[0] = s[len] = '\'';
      s[len + 1] = '\0';
      PQfreemem(es);
  }
  else*/ if (need_quotes) {
      // do string escape
      size_t len = strlen(s);
      char *es = (char *)palloc(2*len + 1 + 2);
      len = PQescapeString(es + 1, s, len);

      // free unescaped version and set new, escaped version
      pfree(s);
      s = (char *)repalloc(es, len + 1 + 2);
      s[0] = s[len + 1] = '\'';
      s[len + 2] = '\0';
  }

  return s;
}


}   // namespace db
