#include <pgclient/connection.hpp>
#include <cool/net/exception.hpp>


namespace pgclient {


void
connection::start (const char *conn)
{
  // start connecting
  state_ = connecting;
  start_time_ = time(0);
  last_result_ = 0;
  if (conn) {
      // create
      assert(connection_ == 0);
      if ((connection_ = PQconnectStart(conn)) == 0)
          cool::raise<cool::resource_error>(COOL_LOCUS);
  }
  else {
      // reset
      assert(connection_ != 0);
      PQresetStart(connection_);
  }

  // quick check if failed early
  if (!connection_ || PQstatus(connection_) == CONNECTION_BAD) {
      state_ = failed;
      pgclient::error e(*this);
      cool::raise(e, COOL_LOCUS);
  }

  // prepare for wait_connected
  PQsetnonblocking(connection_, 1);
  socket::operator=(PQsocket(connection_));
  wait_ = PGRES_POLLING_WRITING;
}


connection &
connection::wait_internal (const cool::time *timeout)
{
  assert(connection_ != 0);
  assert(state_ != failed);

  // wait for ...
  event_type ev = no_events;
  if (state_ == connecting) {
      // ... connecting
      if (wait_ == PGRES_POLLING_READING)
          ev = in;
      else if (wait_ == PGRES_POLLING_WRITING)
          ev = out;
  }
  else if (state_ == ok || state_ == recv_result)
      // ... receiving result
      ev = in;
  else if (state_ == send_query)
      // ... sending query
      ev = out;

  // and now actually wait
  assert(ev != no_events);
  timeout ? socket::wait(*timeout, ev) : socket::wait(ev);

  return *this;
}


bool
connection::ready ()
{
  assert(connection_ != 0);
  assert(state_ != failed);

  switch (state_) {
      case connecting: {
          switch ((wait_ = poll())) {
              case PGRES_POLLING_OK:
                  initialized_ = true;
                  state_ = ok;
                  if (deferred_result_) {
                      // handle postponed query
                      result *r = deferred_result_;
                      deferred_result_ = 0;
                      exec(r);
                  }
                  return initialized_;

              case PGRES_POLLING_FAILED: {
                  pgclient::error e(*this);
                  state_ = failed;
                  cool::raise(e, COOL_LOCUS);
              }

              default:
                  return false;
          }
      }

      case send_query:
          flush();
          return false;

      case ok: {
          // actually there should be no data -- maybe connection is broken?
          if (PQconsumeInput(connection_) == 0) {
              if (PQstatus(connection_) == CONNECTION_BAD) {
                  // yes
                  state_ = failed;
                  cool::raise<cool::connection_error>(COOL_LOCUS);
              }

              assert("ehh? unhandled situation");
          }
          return true;
      }

      case recv_result:
          return consume();

      default:
          break;
  }

  assert(!"ehh? we shouldn't be here");
  return false;
}


connection &
connection::exec (result *r)
{
  assert(state_ == ok);
  assert(deferred_result_ == 0);
  assert(last_result_ == 0);
  assert(r != 0);
  assert(r->sql() != 0);

  // send query
  // TODO: really use separated parameters
  if (!PQsendQueryParams(connection_, r->sql(), 0, 0, 0, 0, 0, 0)) {
      // failure
      state_ = failed;
      if (PQstatus(connection_) == CONNECTION_BAD)
          // connection broken
          cool::raise<cool::connection_error>(COOL_LOCUS);

      // unknown
      cool::raise(pgclient::error(*this), COOL_LOCUS);
  }

  // hook result to self
  state_ = send_query;
  (last_result_ = r)->release(COOL_LOCUS);
  last_result_->db_ = this;
  last_result_->result_ = 0;

  return *this;
}


bool
connection::flush ()
{
  switch (PQflush(connection_)) {
      case -1: {
          // error, unhook from result and throw
          state_ = ok;
          last_result_->db_ = 0;
          last_result_ = 0;
          cool::raise(pgclient::error(*this), COOL_LOCUS);
      }

      case 0:
          // all sent
          state_ = recv_result;
          return true;
  }

  // still blocks
  return false;
}


bool
connection::consume ()
{
  if (PQconsumeInput(connection_) == 1) {
      // return true if next PQgetResult wouldn't block
      if (PQisBusy(connection_) == 1)
          // still not enough data
          return false;

      // PQgetResult wouldn't block any more, get result
      assert(last_result_ != 0);
      assert(last_result_->db_ == this);
      assert(last_result_->result_ == 0);
      last_result_->result_ = PQgetResult(connection_);

      // we allow only 1 query per exec() -- get next result and it must be 0
      PGresult *res = PQgetResult(connection_);
      if (res != 0 &&
              (PQresultStatus(res) == PGRES_FATAL_ERROR ||
               PQresultStatus(res) == PGRES_BAD_RESPONSE))
          // there should be no result, probably connection error
          cool::raise<cool::connection_error>(COOL_LOCUS);
      res = 0; // keep compiler happy when assert() not used

      // free connection for other tasks
      // we assume that caller has remembered pointer in last_result_
      result *r = last_result_;
      last_result_ = 0;
      state_ = ok;

      // check result code
      switch (PQresultStatus(r->result_)) {
          case PGRES_EMPTY_QUERY:
              assert(!"PGRES_EMPTY_QUERY? bug?");
              break;

          case PGRES_TUPLES_OK:
          case PGRES_COMMAND_OK:
              return true;

          case PGRES_BAD_RESPONSE:
              assert(!"PGRES_BAD_RESPONSE");

          case PGRES_FATAL_ERROR:
              cool::raise(
                      pgclient::error(PQresultErrorMessage(r->result_), true),
                      COOL_LOCUS);

          default:
              assert(!"don't know how to handle PQresultStatus()");
      }
  }

  cool::raise(pgclient::error(*this), COOL_LOCUS);
  return true;  // to keep compiler quiet
}


}   // namespace pgclient
