#ifndef pgclient_connection_hpp_included
#define pgclient_connection_hpp_included


/**
 * \file pgclient/connection.hpp
 * \brief PostgreSQL connection class
 */

#include <pgclient/pg.hpp>
#include <pgclient/error.hpp>
#include <pgclient/result.hpp>
#include <cool/net/socket.hpp>


namespace pgclient
{


/** \brief PostgreSQL database client */
class connection: public cool::socket
{
  public:

      /**
       * \brief Create connection object and start connecting
       *
       * Call wait() to wait until connection is established and ready() to
       * check if it succeeded.
       *
       * Optional \a immediate_query specifies query to be executed
       * immediately after connection is established.
       */
      connection (const char *conn, result *immediate_query = 0)
          : connection_(0)
          , initialized_(false)
          , deferred_result_(immediate_query)
      {
          assert(conn != 0);
          start(conn);
      }


      /** \brief Disconnect and destroy connection object */
      ~connection () throw() {
          stop();
      }


      time_t start_time () const {
          return start_time_;
      }


      /**
       * \brief Reset connection
       *
       * Use wait() after reset() to wait until connection is
       * re-established and ready() to check if it succeeded.
       */
      connection &reset () {
          start(0);
          return *this;
      }


      /**
       * \brief Send query r.sql() to database backend. Result is stored to
       * \a r when it arrives.
       *
       * Call wait() to wait for result and ready() to check if result has
       * been arrived.
       */
      connection &exec (result *r);


      /**
       * \brief Wait up to \a timeout until last operation is finished.
       *
       * Last operation can be creating object (constructor), resetting
       * connection (reset()), executing SQL statement (exec()). Call ready()
       * to check if really is finished.
       */
      connection &wait (const cool::time &timeout) {
          return wait_internal(&timeout);
      }


      /**
       * \brief Wait indefinitely until last operation is finished.
       *
       * \copydoc wait(const cool::time&)
       */
      connection &wait () {
          return wait_internal(0);
      }


      /** \brief Return true if last operation finished and false otherwise */
      bool ready ();


  private:

      friend class error;
      friend class result;
      friend class pool;


      /** \brief Start connecting */
      void start (const char *conn = 0);


      /** \brief Drop connection */
      void stop () {
          if (connection_) {
              PQfinish(connection_);
              connection_ = 0;
              //BUG socket::operator=(-1);
          }
      }


      /** \brief Wait until socket is ready */
      PostgresPollingStatusType poll () {
          assert(state_ == connecting);
          return initialized_
              ? PQresetPoll(connection_)
              : PQconnectPoll(connection_);
      }


      /** \brief Wait up to \a timeout until last operation is finished */
      connection &wait_internal (const cool::time *timeout);


      /** \brief Flush pending query to backend */
      bool flush ();


      /** \brief Receive data from backend. Return true if ready */
      bool consume ();


      /** \brief State type of connection */
      enum state_type {
          connecting        /**< Connecting to database server */
          , ok              /**< Connection is ready for exec() */
          , send_query      /**< Sending query */
          , recv_result     /**< Fetching result */
          , failed          /**< Connection establishing has been failed */
      };

      // TEST
      std::string state_string (state_type s) {
          switch (s) {
              case connecting:
                  return "connecting";
              case ok:
                  return "ok";
              case send_query:
                  return "send_query";
              case recv_result:
                  return "recv_result";
              case failed:
                  return "failed";
          }

          return "unknown";
      }


      /** \brief Low-level connection data structure */
      PGconn *connection_;

      /** \brief True if connection has been established */
      bool initialized_;

      /** \brief Current state of connection */
      state_type state_;

      /** \brief Next wait() type */
      PostgresPollingStatusType wait_;

      /** \brief Address of structure where expected result is stored */
      result *last_result_;

      /**
       * \brief In case of broken connection, query is postponed until
       * connection is restored. This is temporary storage for query
       */
      result *deferred_result_;

      /** \brief Time of last connection establishment */
      time_t start_time_;
};


}   // namespace pgclient


#endif  // pgclient_connection_hpp_included
