#include <pgclient/pool.hpp>
#include <cool/net/exception.hpp>
#include <sstream>


namespace pgclient
{


pool::pool (const char *conn
        , size_t min_count, size_t max_count
        , int reconnect_interval
        , int reset_interval)
    : pair_()
    , idle_(0)
    , min_count_(min_count)
    , max_count_(max_count)
    , reconnect_interval_(reconnect_interval)
    , reset_interval_(reset_interval)
{
  assert(conn != 0);
  assert(max_count_ > 0);
  assert(min_count_ <= max_count_);

  poller_.insert(&pair_.first);
  snprintf(conn_info_, sizeof(conn_info_), "%s", conn);
  create(min_count_);

  // set system manager intervals
  time_t now = time(0);
  next_reconnect_ = now + reconnect_interval_;
  next_reset_time_ = now + 60;

#if defined(PGCLIENT_STATS)
  total_stats_.startup = last_stats_.startup = time(0);
  total_stats_.activity_duration_usec = last_stats_.activity_duration_usec = 0;
  total_stats_.activity_duration_sec = last_stats_.activity_duration_sec = 0;
  total_stats_.query_count = last_stats_.query_count = 0;
  total_stats_.failure_count = last_stats_.failure_count = 0;
#endif
}


pool &
pool::create (size_t count)
{
  while (count--) {
      try {
          // create new connection and register in poller
          create_and_exec(0);
      }
      catch (...) {
          // created not as much as wanted but that's life
          // do whatever we can
          break;
      }
  }

  return *this;
}


pool &
pool::create_and_exec (result *r)
{
  // create new connection with optional postponed query and register in poller
  connection *c = new connection(conn_info_, r);
  connections_.push_back(c);
  poller_.insert(c, connection::out);
  return *this;
}


pool &
pool::exec (result *r)
{
#if defined(PGCLIENT_STATS)
  last_stats_.query_count++;
#endif

  if (idle_) {
      // find idle connection
      connection_list::iterator i;
      for (i = connections_.begin();  i != connections_.end();  ++i) {
          if ((*i)->state_ == connection::ok) {
              idle_--;
              try {
                  // found
                  (*i)->exec(r);
                  poller_.modify(*i, connection::out);
                  return *this;
              }
              catch (...) {
                  // broken, restart and find next idle
                  try {
                      poller_.erase(*i);
                  }
                  catch (...) {
                      // ignore this error
                  }
                  try  {
                      (*i)->reset();
                      poller_.insert(*i, connection::out);
                  }
                  catch (...) {
                      (*i)->state_ = connection::failed;
                  }
              }
          }
      }
  }

  error current_error("");
  if (connections_.size() < max_count_) {
      // create new connection with postponed query that will be executed
      // immediately after connection is established
      try {
          create_and_exec(r);
          return *this;
      }
      catch (const cool::exception &e) {
          std::ostringstream os;
          os << e;
          current_error.what(os.str().c_str());
      }
  }
  // else: all existing connections are resetting (some global failure) or
  // there is already max connections.

  // no idle connections
#if defined(PGCLIENT_STATS)
  last_stats_.failure_count++;
#endif

  r->on_error(result::no_connection, current_error);

  return *this;
}


//
// sorry, hairy code warning
//

pool &
pool::wait_internal (const cool::time *timeout)
{
  time_t now = time(0);
  maintain_min_connections(now);

  // wait for events
  cool::poller::iterator i = timeout ?
      poller_.begin(*timeout) : poller_.begin();

#if defined(PGCLIENT_STATS)
  // connections handling start time
  timeval start;
  gettimeofday(&start, 0);
  bool handled_anything = false;
#endif

  // process each active connection
  for ( ; i != poller_.end();  ++i) {
      cool::socket as_socket = *i;
      result *r;

      if (i == &pair_.first) {
          size_t l = pair_.first.recv((char *)&r, sizeof(r));
          assert(l == sizeof(r));
          l = 0;  // to keep compiler happy :)
          exec(r);
          continue;
      }

      // get connection and it's current state
      connection *c = static_cast<connection *>(i.operator->());
      connection::state_type old = c->state_;

      // remember result for possible later releasing
      r = static_cast<result *>(c->last_result_);

      // act on connection
      try {
          c->ready();
          if (old == c->state_ && c->state_ != connection::connecting)
              // state hasn't changed, do nothing
              continue;
      }
      catch (const cool::connection_error &e) {
          // connection broken
#if defined(PGCLIENT_STATS)
          last_stats_.failure_count++;
#endif
          if (connections_.size() > min_count_) {
              // too many of them, anyway
              // let next big switch do drop it
              c->state_ = connection::failed;
          }
          else {
              // restart
              try {
                  poller_.erase(c);
              }
              catch (...) {
                  // ignore this error, socket probably gone bad
              }
              try {
                  c->reset();
                  poller_.insert(c, connection::out);
                  if (old == connection::ok)
                      idle_--;
                  continue;
              }
              catch (...) {
                  // failed to reset, let next big switch do handle it
                  c->state_ = connection::failed;
              }
          }
      }
      catch (const error &e) {
          if (r || c->deferred_result_) {
              if (c->deferred_result_) {
                  r = dynamic_cast<result *>(c->deferred_result_);
                  c->deferred_result_ = 0;
              }

              // let caller handle error and mark connection as failed
              r->on_error(result::db_error, e);
              c->state_ = connection::failed;
#if defined(PGCLIENT_STATS)
              last_stats_.failure_count++;
#endif
          }

          // let rest of error to be handled by next big switch
      }
      catch (const STD_ exception &e) {
          // another error, drop connection
          c->state_ = connection::failed;
#if defined(PGCLIENT_STATS)
          last_stats_.failure_count++;
#endif
      }

#if defined(PGCLIENT_STATS)
      handled_anything = true;
#endif

      // no last_stats_.failure_count modifying in last half

      // update stuff
state_changed:
      try {
          switch (c->state_) {
              case connection::connecting:
                  switch (c->wait_) {
                      case PGRES_POLLING_READING:
                          poller_.modify(c, connection::in);
                          break;

                      case PGRES_POLLING_WRITING:
                          poller_.modify(c, connection::out);
                          break;

                      default:
                          assert(!"connection::connecting: invalid state");
                  }
                  break;

              case connection::send_query:
                  poller_.modify(c, connection::out);
                  break;

              case connection::recv_result:
                  poller_.modify(c, connection::in);
                  break;

              case connection::ok:
                  if (old == connection::recv_result) {
                      // result is ready
                      r->on_ready();

                      // r on_ready must call release itself because of possible
                      // multithreading issues
                      // r->release(COOL_LOCUS);
                  }
                  idle_++;
                  poller_.modify(c, connection::in);
                  break;

              case connection::failed:
                  // connection has been failed, drop
                  if (old == connection::ok)
                      idle_--;
                  try {
                      poller_.erase(&as_socket);
                  }
                  catch (...) {
                      // ignore this error, socket probably already gone bad
                  }
                  connections_.remove(c);
                  delete c;
                  break;

              default:
                  assert(!"connection::state_: invalid");
          }
      }
      catch (...) {
          // unknown unhandled error, drop this connection
          c->state_ = connection::failed;
          goto state_changed;
      }
  }

#if defined(PGCLIENT_STATS)
  // connections handling end time
  if (handled_anything) {
      timeval stop;
      gettimeofday(&stop, 0);
      time_t s = stop.tv_sec - start.tv_sec;
      suseconds_t us = stop.tv_usec - start.tv_usec;
      if (us < 0) {
          --s;
          us += 1000000;
      }

      last_stats_.activity_duration_sec += s;
      if ((last_stats_.activity_duration_usec += us) >= usec_per_sec) {
          // normalize interval
          last_stats_.activity_duration_sec +=
              last_stats_.activity_duration_usec/usec_per_sec;
          last_stats_.activity_duration_usec %= usec_per_sec;
      }
  }
#endif

  if (reset_interval_ && next_reset_time_ <= now && connections_.size() > 0) {
      reset_old_connections(now);
      next_reset_time_ = now + 60;
  }

  return *this;
}


void
pool::maintain_min_connections (time_t now)
{
  size_t active_connections = connections_.size();
  if (active_connections < min_count_) {
      // some connections are broken, try to repair but not more than once per
      // reconnect_interval_ seconds
      if (next_reconnect_ <= now) {
          // now it's good time
          create(min_count_ - active_connections);

          // set next check time
          next_reconnect_ = now + reconnect_interval_;
      }
  }
}


void
pool::reset_old_connections (time_t now)
{
  int n = 0;
  connection_list::iterator i = connections_.begin();
  while (i != connections_.end()) {
      int random_const = (n++)%(reset_interval_/60)*60;
      if ((*i)->state_ != connection::ok
              || (*i)->start_time() + reset_interval_ + random_const > now) {
          // no need to reset yet
          ++i;
          continue;
      }

      // remove connection from poller
      --idle_;
      try {
          try {
              poller_.erase(*i);
          }
          catch (const cool::os_error &e) {
              // ignore this error, socket probably already gone bad
          }
      }
      catch (const cool::invalid_value_error &e) {}

      // reset and re-insert into poller
      // on failure, forget this connection
      try {
          (*i)->reset();
          poller_.insert(*i, connection::out);
          ++i;
      }
      catch (...) {
          connection *c = *i;
          i = connections_.erase(i);
          delete c;
      }
  }
}


}   // namespace pgclient
