#ifndef pgclient_pool_hpp_included
#define pgclient_pool_hpp_included


/** \brief Define to use optional code for collecting statistics */
#define PGCLIENT_STATS


#include <pgclient/pg.hpp>
#include <pgclient/error.hpp>
#include <pgclient/result.hpp>
#include <pgclient/connection.hpp>
#include <cool/net/poller.hpp>
#include <cool/net/local_socket.hpp>
#include <cool/net/socket_pair.hpp>
#include <list>


namespace pgclient
{


/** \brief PostgreSQL database client pooler */
class pool
{
  public:


      /** \brief Asynchronous query result */
      class result: public pgclient::result
      {
        public:

            /** \brief Construct empty result without query */
            result (const char *q): pgclient::result(q) { }


            /** \brief Destroy result */
            virtual ~result () {}


            /** \brief Set new SQL statement */
            virtual result &sql (const char *q) {
                pgclient::result::sql(q);
                return *this;
            }


            /** \brief Get current SQL statement */
            virtual const char *sql () const throw() {
                return pgclient::result::sql();
            }


            /** \brief Parameter for on_error() */
            enum error_type {
                no_connection
                , db_error
            };


            /**
             * \brief Called automatically whenever query failed
             *
             * Design by contract: this method must call
             * pglient::result::release(). Unfortunately framework itself
             * can't do it because of multithreading issues.
             */
            virtual void on_error (error_type code, const error &e) throw() = 0;


            /**
             * \brief Called automatically whenever result is ready
             *
             * Design by contract: this method must call
             * pglient::result::release(). Unfortunately framework itself
             * can't do it because of multithreading issues.
             */
            virtual void on_ready () throw() = 0;
      };


      /** \brief Create pool of connections to \a conn */
      pool (const char *conn
              , size_t min_count, size_t max_count
              , int reconnect_interval
              , int reset_interval);


      /** \brief Drop and destroy pool of connections */
      ~pool () throw() {
          connection_list::iterator i = connections_.begin();
          for ( ;  i != connections_.end();  ++i)
              delete *i;
      }


      /** \brief Return connection string */
      const char *options () const throw () {
          return conn_info_;
      }


      /** \brief Return number of active connections in pool */
      size_t size () const {
          return connections_.size();
      }


      /** \brief Return number of idle connections in pool */
      size_t idle () const {
          return idle_;
      }


      /**
       * \brief Add \a count new connections to pool using connection string
       * specified in constructor
       */
      pool &create (size_t count = 1);


      /**
       * \brief Add one new connection to pool using connection string
       * specified in constructor and with postponed query that will be
       * executed immediately after connection is created.
       *
       * \see exec
       */
      pool &create_and_exec (result *r);


      /**
       * \brief Register query \a r.sql() for asynchronous execution.
       * Whenever result is ready \a r.on_ready() is called
       */
      pool &exec (result *r);


      /**
       * \brief Register query \a r.sql() for asynchronous execution from
       * another thread.
       *
       * Whenever result is ready r.on_ready() is called and on error
       * r.on_error(). Callback are called in pool's thread context.
       */
      pool &exec_thread (result *r) {
          size_t l = pair_.second.send((char *)&r, sizeof(r));
          assert(l == sizeof(r));
          l = 0;  // to keep compiler happy :)
          return *this;
      }


      /**
       * \brief Wait up to \a timeout for activity in pooled connections and
       * act accordingly.
       */
      pool &wait (const cool::time &timeout) {
          return wait_internal(&timeout);
      }


      /**
       * \brief Wait indefinitely for activity in pooled connections and act
       * accrodingly
       */
      pool &wait () {
          return wait_internal(0);
      }


#if defined(PGCLIENT_STATS)
      /** \brief Structure containing connection pool statistics */
      struct statistics_type
      {
          /** \brief Time when pool started up */
          time_t startup;

          /** \brief Number of requests sent to database through pool */
          long query_count;

          /** \brief Number of failed requests */
          long failure_count;

          /** \brief Total time spent servicing this database (in usec) */
          long activity_duration_usec;
          long long activity_duration_sec;
      };


      static const int usec_per_sec = 1000000;


      /**
       * \brief Return statistics data for whole runtime
       * \note Must be called before querying last_stats() because it uses data
       * collected by last_stats() but last_stats() resets it's data.
       */
      const statistics_type &total_stats () {
          // update
          total_stats_.activity_duration_sec +=
              last_stats_.activity_duration_sec;
          total_stats_.activity_duration_usec +=
              last_stats_.activity_duration_usec;
          if (total_stats_.activity_duration_usec >= usec_per_sec) {
              // normalize time
              total_stats_.activity_duration_sec +=
                  total_stats_.activity_duration_usec/usec_per_sec;
              total_stats_.activity_duration_usec %= usec_per_sec;
          }
          total_stats_.query_count += last_stats_.query_count;
          total_stats_.failure_count += last_stats_.failure_count;
          return total_stats_;
      }


      /**
       * \brief Return statistics data for period from last call
       * \see total_stats()
       */
      statistics_type last_stats () {
          // get data
          statistics_type tmp = last_stats_;

          // reset
          last_stats_.startup = time(0);
          last_stats_.activity_duration_usec = 0;
          last_stats_.activity_duration_sec = 0;
          last_stats_.query_count = 0;
          last_stats_.failure_count = 0;

          return tmp;
      }
#endif


  private:


      /** \brief Pool's real work-horse */
      pool &wait_internal (const cool::time *timeout);

      /**
       * \brief Called during each wait_internal() to check that pooler has
       * at least min_count_ connections
       */
      void maintain_min_connections (time_t now);

      /**
       * \brief Called periodically by wait_internal() to reset too old
       * connections
       */
      void reset_old_connections (time_t now);

      /** \brief Size of connection string buffer */
      enum { conn_info_size = 128 };

      /** \brief Connection string */
      char conn_info_[conn_info_size + 1];

      /** \brief Container for pooled connections */
      typedef STD_ list<connection *> connection_list;

      /** \brief List of pooled connections */
      connection_list connections_;

      /** \brief Pool poller */
      cool::poller poller_;

      /** \brief Socket pair for sending another query to pool's thread */
      cool::socket_pair<cool::local> pair_;

      /** \brief Number of idle connections in pool */
      size_t idle_;

      /** \brief Minimum/maximum number of connections */
      size_t min_count_;
      size_t max_count_;

      /**
       * \brief Interval (seconds) for attempting to repair broken
       * connections
       */
      int reconnect_interval_;

      /** \brief Time for next connection repairing */
      time_t next_reconnect_;

      /** \brief Time for next old connections reset */
      time_t next_reset_time_;

      /** \brief Interval (seconds) to reset partition connections */
      int reset_interval_;

#if defined(PGCLIENT_STATS)
      /** \brief Statistics for pool */
      statistics_type total_stats_, last_stats_;
#endif
};


}   // namespace pgclient


#endif  // pgclient_pool_hpp_included
