#ifndef pgclient_result_hpp_included
#define pgclient_result_hpp_included


/**
 * \file pgclient/result.hpp
 * \brief Database query result and related classes
 */

#include <pgclient/pg.hpp>
#include <pgclient/error.hpp>


namespace pgclient
{


/**
 * \brief Database query result class
 *
 * \note Instance of this class shouldn't go out of scope until result is
 * received from connection. Otherwise assertion is thrown in destructor.
 */
class result
{
  public:

      class tuple;
      class field;

      /** \brief Row number type */
      typedef int size_type;


      /** \brief Class representing row in result set */
      class tuple
      {
        public:

            /** \brief Field number type */
            typedef int size_type;


            /** \brief Field type */
            typedef field reference;


            /** \brief Create object referencing \a row in resultset \a r */
            tuple (const result &r, result::size_type row)
                : owner_(r), row_(row) {}


            /** \brief Destructor */
            ~tuple () {}


            /** \brief Return number of fields in tuple */
            size_type size () const throw() {
                return owner_.fields();
            }


            /** \brief Return index of this tuple in resultset */
            result::size_type row () const throw() {
                return row_;
            }


            /** \brief Return field number corresponding to name \a f */
            size_type field_number (const char f[]) const {
                return owner_.field_number(f);
            }


            /** \brief Return field name for column \a f */
            const char *field_name (size_type f) const {
                return owner_.field_name(f);
            }


            /** \brief Get field by column number \a f */
            reference operator[] (size_type f) const {
                return field(*this, f);
            }


            /** \brief Get field by column name \a f */
            reference operator[] (const char f[]) const {
                return field(*this, field_number(f));
            }


            /**
             * \brief Get field by column number \a f.
             *
             * \throw cool::limit_error if \a f is out of range
             */
            reference at (size_type f) const {
                if (f >= size())
                    cool::raise<cool::limit_error>(COOL_LOCUS);
                return operator[](f);
            }


            /**
             * \brief Get field by column name \a f.
             *
             * \throw cool::not_found_error if field is not found.
             */
            reference at (const char f[]) const {
                return operator[](field_number(f));
            }


        private:

            tuple ();

            friend class field;
            const result &owner_;
            result::size_type row_;
      };

      /** \brief Class representing one field in tuple */
      class field
      {
        public:

            /** \brief Field size type */
            typedef int size_type;


            /** \brief Create field referencing field in tuple */
            field (const tuple &t, tuple::size_type f): tuple_(t), field_(f) {}


            /** \brief Destructor */
            ~field () {}


            /** \brief Return value as string */
            const char *c_str () const {
                return tuple_.owner_.field_value(tuple_.row_, field_);
            }


            /** \brief Return field name */
            const char *name () const {
                return tuple_.owner_.field_name(field_);
            }


            /** \brief Return true if field is null */
            bool is_null () const {
                return tuple_.owner_.is_null(tuple_.row_, field_);
            }


            /** \brief Return size of field */
            size_type size () const {
                return tuple_.owner_.field_size(tuple_.row_, field_);
            }


        private:

            const tuple &tuple_;
            tuple::size_type field_;
      };


      /** \brief Create empty result set */
      result (const char *q = 0): db_(0), result_(0), sql_(q) {
          release(COOL_LOCUS);
      }


      /** \brief Create copy of result. Ownership of result goes to this */
      result (const result &r) {
          operator=(r);
      }


      /** \brief Destroy result set, releasing internal resources */
      virtual ~result () {
          release(COOL_LOCUS);
      }


      /** \brief Create copy of result. Ownership of result goes to this */
      result &operator = (const result &r);


      /** \brief Return pointer to previously sent query statement */
      virtual const char *sql () const throw() {
          return sql_;
      }


      /** \brief Set new query */
      virtual result &sql (const char *q) {
          sql_ = q;
          return *this;
      }


      /** \brief Return number of rows in current result set */
      size_type size () const throw() {
          return result_ ? PQntuples(result_) : 0;
      }


      /** \brief Return true if result set is currently empty */
      bool empty () const throw() {
          return !result_ || (PQntuples(result_) == 0);
      }


      /** \brief Return number of fields in resultset */
      tuple::size_type fields () const throw() {
          return result_ ? PQnfields(result_) : 0;
      }


      /** \brief Get field name from resultset at \a field */
      const char *field_name (tuple::size_type f) const {
          assert(result_ != 0);
          const char *n = PQfname(result_, f);
          if (!n)
              cool::raise<cool::limit_error>(COOL_LOCUS);
          return n;
      }


      /** \brief Get field number corresponding to field name \a f */
      tuple::size_type field_number (const char f[]) const {
          assert(result_ != 0);
          int n = PQfnumber(result_, f);
          if (n == -1)
              cool::raise<cool::not_found_error>(COOL_LOCUS);
          return (tuple::size_type)n;
      }


      /** \brief Get value from resultset at \a row : \a field */
      const char *field_value (size_type r, tuple::size_type f) const {
          assert(result_ != 0);
          return PQgetvalue(result_, r, f);
      }


      /** \brief Get field size from resultset at \a field */
      field::size_type field_size (size_type r, tuple::size_type f) const {
          assert(result_ != 0);
          return PQgetlength(result_, r, f);
      }

      /** \brief Return true if \a field in \a row is null */
      bool is_null (size_type r, tuple::size_type f) const {
          assert(result_ != 0);
          return PQgetisnull(result_, r, f);
      }


      /** \brief Return tuple at \a row */
      const tuple operator[] (size_type row) const {
          return tuple(*this, row);
      }


      /** \brief Return tuple at \a row */
      const tuple at (size_type row) const {
          if (row >= size())
              cool::raise<cool::limit_error>(COOL_LOCUS);
          return tuple(*this, row);
      }


  protected:

      /** \brief Release current result set if we are owner */
      void release (const char *loc) {
          if (db_) {
              // result not received from connection yet but result already
              // goes out of scope
              // lazy bastard, burn in assertion :)
              //XXX assert(result_ != 0);
              if (result_ != 0)
                  // cool::raise<cool::permission_error>(loc);
                  PQclear(result_);
              db_ = 0;
          }
          result_ = 0;
      }


  private:

      friend class connection;
      friend class pool;

      connection *db_;
      PGresult *result_;
      const char *sql_;
};


}   // namespace pgclient


#endif  // pgclient_result_hpp_included
