#include <pool/pgcpoold.hpp>
#include <pool/listener.hpp>
#include <pool/config.hpp>
#include <pool/partitions.hpp>
#include <cool/thread/mutex.hpp>


namespace pgcpool
{


client::state
client::on_readable ()
{
  try {
      // read as much as we can
      int r = recv((char *)proto.end(), proto.free());
      if (r == 0)
          // readable but not data -- broken
          return broken;

      if (proto.recvd(r))
          // have full request
          return ready;

      if (proto.full())
          // overflow, assume attack and drop client
          return overflow;
  }
  catch (const cool::connection_error &e) {
      return broken;
  }
  catch (const std::exception &e) {
      warning() << peer() << ": recv failed: " << e.what();
      return broken;
  }

  // still no full request
  return need_more;
}


client::state
client::on_writable ()
{
  try {
      // write as much as we can
      int w = send((char *)proto.begin(), proto.more());
      if (w == 0)
          // writable but can't write. Whazzup?
          return broken;

      if (proto.sent(w))
          // everything in buffer is sent
          return ready;
  }
  catch (const cool::connection_error &e) {
      return broken;
  }

  // still not all sent
  return need_more;
}


//
// XXX: for optimization this function bypasses protocol stack :(
//
// XXX: actually there might be point in optimizing also case for very big
// (1x1) result set -- can save 1 big memcpy...
//
void
client::on_ready () throw()
{
  try {
      // get result size and send to proxydb
      size_type r, rc = size();
      tuple::size_type c, cc = fields();
      proto.send_result(protocol::rsp_ok, rc, cc);
      proto.flush();

      pgcpool::debug() << peer() << ": result(" << rc << 'x' << cc << ')';

      if (rc == 1 && cc == 1) {
          // specific case for small result set (to prevent memcpy)
          field f = (*this)[0][0];
          if (!f.is_null())
              proto.send(f.c_str(), f.size());
          else
              proto.send_null();
      }
      else {
          // buffered send
          proto.bptr = 0;
          for (r = 0;  r < rc;  ++r) {
              for (c = 0;  c < cc;  ++c) {
                  field f = (*this)[r][c];

                  bool is_null = f.is_null();
                  uint32_t sz = sizeof(uint32_t) + (is_null ? 0 : f.size());

                  // will next chunk fit to remaining gap?
                  if (proto.size_max - proto.bptr < sz) {
                      // no, flush
                      if (proto.bptr > 0) {
                          proto.channel.send((char *)proto.data, proto.bptr);
                          proto.bptr = 0;
                      }

                      // does it fit at all?
                      // for optimizing we will also count for next minimal size
                      // field also (5 bytes = sizeof(uint32_t) + sizeof(char))
                      if (proto.size_max <= sz + 5) {
                          proto.send(f.c_str(), f.size());
                          continue;
                      }
                  }

                  // push data to buffer
                  if (!is_null) {
                      // decrement by length field
                      sz -= sizeof(uint32_t);

                      // put length
                      *(uint32_t *)(proto.data + proto.bptr) = htonl(sz);
                      proto.bptr += sizeof(uint32_t);

                      // put field
                      memcpy(proto.data + proto.bptr, f.c_str(), sz);
                  }
                  else {
                      *(uint32_t *)(proto.data + proto.bptr) =
                          htonl(std::numeric_limits<uint32_t>::max());
                  }

                  // advance pointer
                  proto.bptr += sz;
              }
          }

          // final flush (if needed)
          if (proto.bptr) {
              proto.channel.send((char *)proto.data, proto.bptr);
              proto.bptr = 0;
          }
      }
      proto.flush();
  }
  catch (const cool::connection_error &e) {
      // connection broken
  }
  catch (...) {
      // no errors can escape from this function
      exception::global_handler();
  }

  // ready for next request
  proto.wait_readable();
  release(COOL_LOCUS);
  owner->ready(this);
}


void
client::on_error (result::error_type code, const pgclient::error &e) throw()
{
  try {
      switch (code) {
          case no_connection:
              warning() << peer() << ": db_error(no available connection)";
              response(protocol::rsp_enores);
              break;

          case db_error:
              warning() << peer() << ": db_error(" << e.what() << ')';
              response(protocol::rsp_invalid);
              break;
      }
  }
  catch (const cool::connection_error &e) {
      // connection broken
  }
  catch (...) {
      // no errors can escape from this function
      exception::global_handler();
  }

  // ready for next request
  release(COOL_LOCUS);
  owner->ready(this);
}


void
listener::wait_and_dispatch ()
{
  // poll incoming client connections
  for (poller::iterator i = poller_.begin();  i != poller_.end();  ++i) {
      if (i == this)
          // new connection
          new_client();

      else if (i == &pair_.first) {
          // request handled, reregister client
          // we get the address through socketpair. It is possible because we
          // have common address-space with threads
          client *c;
          size_t l = pair_.first.recv((char *)&c, sizeof(c));
          assert(l == sizeof(c));
          l = 0;
          try {
              poller_.modify(c, c->wait_events());
          }
          catch (const cool::invalid_value_error &e) {
              delete_client(c);
          }
      }

      else if (i->readable()) {
          // socket become readable
          client *c = static_cast<client *>(i.operator->());
          switch (c->on_readable()) {
              case client::need_more:
                  // nothing to do, wait for more data
                  break;

              case client::ready:
                  // stop listening until request is fulfilled
                  try {
                      poller_.modify(c, c->wait_events());
                  }
                  catch (const cool::invalid_value_error &e) {
                      delete_client(c);
                  }

                  if (!handle_request(c))
                      // failure, close client
                      delete_client(c);

                  break;

              case client::overflow:
                  warning() << c->peer() << ": overflow";
                  // fall-through

              case client::broken:
                  // die, bastard, die
                  delete_client(c);
                  break;
          }
      }

      else if (i->writable()) {
          // socket become writable
          client *c = static_cast<client *>(i.operator->());
          switch (c->on_writable()) {
              case client::need_more:
                  // more to write
                  break;

              case client::ready:
                  // request fullfilled, start listening for readable again
                  try {
                      poller_.modify(c, c->wait_events());
                  }
                  catch (const cool::invalid_value_error &e) {
                      delete_client(c);
                  }
                  break;

              default:
                  // there should be no other statuses during writable...
                  assert(!"UNKNOWN STATUS");
          }
      }
  }
}


void
listener::new_client ()
{
  // accept client
  common::net_space::address a;
  common::net_space::socket::socket_type s;
  try {
      s = accept(a);
  }
  catch (const std::exception &e) {
      warning() << "accept: " << e.what();
      return;
  }

  // find free slot
  client_list::iterator i = clients_.end();
  if (!full()) {
      for (i = clients_.begin();  i != clients_.end();  ++i) {
          if ((*i)->handle() == -1)
              // found
              break;
      }
  }

  if (i == clients_.end()) {
      // no free slot
      warning() << "client list full, reject";
      ::close(s);
      return;
  }

  // update client data structure to reflect that it is active now
  (*i)->operator=(s);
  (*i)->peer(a, i - clients_.begin());

  // register for listening
  try {
      poller_.insert(*i);
  }
  catch (const cool::invalid_value_error &e) {
      // registering failed, drop connection
      (*i)->close();
      return;
  }
  catch (const std::exception &e) {
      warning() << (*i)->peer() << ": failed to listen: " << e.what();
      (*i)->close();
      return;
  }
  catch (...) {
      warning() << (*i)->peer() << ": failed to listen: unknown";
      (*i)->close();
      return;
  }

  info() << (*i)->peer() << ": connect";
}


bool
listener::handle_request (client *c)
{
  try {
      protocol::command_type cmd = c->command();
      switch (cmd) {
          case protocol::cmd_set_status:
              // TODO: cmd_set_status
              info() << c->peer() <<
                  ": cmd_set_status(" << c->proto.new_status() << ')';
              c->response(protocol::rsp_enotimpl);
              break;

          case protocol::cmd_get_info:
              // TODO: cmd_get_info
              info() << c->peer() << ": cmd_get_info";
              c->send_statistics("(1,yks)", 7);
              break;

          case protocol::cmd_exec_query: {
              protocol::query_type query(c->query, c->cluster
                  , c->hash_key.conf_map, c->hash_key.hash_nr);
              c->proto.recv_query(query);
              partitions::queue_query(query, c);

              // no need to update poller flags (already done)
              return true;
          }

          default:
              // bad command, drop client
              warning() << c->peer() << ": invalid(" << cmd << ')';
              return false;
      }
  }
  catch (const cool::connection_error &e) {
      warning() << c->peer() << ": connection broken";
      return false;
  }

  // change waitable event type for client
  try {
      poller_.modify(c, c->wait_events());
  }
  catch (const cool::invalid_value_error &e) {
      warning() << c->peer() << ": poller.modify: failure";
      return false;
  }

  return true;
}


}   // namespace pgcpool
