#ifndef pool_listener_hpp_included
#define pool_listener_hpp_included

/**
 * \file pool/listener.hpp
 * \brief PG connection pooler listener and active clients collection
 */


#include <pool/pgcpoold.hpp>
#include <pool/config.hpp>
#include <pool/logger.hpp>
#include <pool/partitions.hpp>
#include <common/protocol.hpp>
#include <cool/net/poller.hpp>
#include <cool/net/socket_pair.hpp>
#include <vector>


namespace pgcpool
{


class listener;


/** \brief Class representing communication channel between client and daemon */
class client
    : public common::net_space::socket
    , public pgclient::pool::result
{
  public:

      /** \brief State of communication buffer */
      enum state {
          need_more     /**< Need more data to gather full request */
          , ready       /**< Have full request. Start handling */
          , overflow    /**< Bastard overflood channel */
          , broken      /**< Connection broken */
      };


      /** \brief Construct new, inactive client */
      client (listener *owner, size_t max_request)
          : common::net_space::socket(-1)
          , pgclient::pool::result(0)
          , proto(*this, max_request)
          , owner(owner)
      { }


      /** \brief Destroy client, closing socket if neccessary */
      virtual ~client () throw() {
          if (!operator!())
              close();
      }


      /** \brief Set new low-level socket. If already active, throw assertion */
      client &operator= (common::net_space::socket::socket_type s) {
          assert(operator!() == true);
          common::net_space::socket::operator=(s);
          proto.wait_readable();
          return *this;
      }


      /** \brief Close socket and mark client as invactive */
      client &close () {
          assert(operator!() == false);
          common::net_space::socket::close();
          return *this;
      }


      /** \brief Return address of other end-point */
      const char *peer () const throw() {
          return peer_;
      }

      // ** listener callbacks **

      /**
       * \brief Act whenever client becomes readable. Return new state based
       * on received data
       */
      virtual state on_readable ();


      /**
       * \brief Act whenever client becomes writable. Return new state based
       * on sent data
       */
      virtual state on_writable ();


      // ** pgclient::result callbacks **

      /**
       * \brief Called asynchronously whenever database query result has
       * arrived
       */
      void on_ready () throw();


      /**
       * \brief Asynchronous callback invoked whenever database query can't be
       * successfully finished
       */
      void on_error (result::error_type code, const pgclient::error &e) throw();


      /** \brief Return events for what listener is waiting */
      event_type wait_events () const {
          return proto.wait_events;
      }


      /** \brief Return id of current request */
      protocol::command_type command () const {
          return proto.last_command();
      }


      /** \brief Set response code for last command */
      void response (protocol::response_type r) {
          proto.response(r);
      }


      /**
       * \brief Send statistics back to client. \a info must be formatted as
       * tuple ret_pool_statistics
       */
      void send_statistics (const char *info, uint32_t length) {
          proto.send_statistics(info, length);
      }


      /** \brief Last received query */
      std::string query;
      std::string cluster;
      partitions::cluster_map_key_type hash_key;


  private:

      friend class listener;


      /** \brief Other communication end-point address as text */
      char peer_[sizeof("255.255.255.255:65535")];


      /** \brief Set new peer's address */
      template <class T> void peer (T &a, int index);


      /** \brief Communication protocol buffer */
      common::protocol proto;


      /** \brief Owner collection */
      listener *owner;
};


template <> inline void
client::peer<cool::inet::address> (cool::inet::address &a, int index)
{
  snprintf(peer_, sizeof(peer_), "%s:%d", a.host().c_str(), a.port());
}


template <> inline void
client::peer<cool::local::address> (cool::local::address &a, int index)
{
  snprintf(peer_, sizeof(peer_), "%d", index);
}



/**
 * \brief Listener class that accepts incoming connections and polls already
 * connected clients for activity.
 */
class listener: public common::net_space::listener
{
  public:

      /** \brief Start listening at rendevouz point \a a */
      listener (common::net_space::address &a, unsigned short max_clients);


      /** \brief Stop listening and close listener socket */
      virtual ~listener () throw();


      /** \brief Listen for client activity and act accordingly */
      void wait_and_dispatch ();


      /** \brief Return true if max_clients have already connected */
      bool full () const {
          // -1 for listener socket
          return poller_.size() - 1 == clients_.size();
      }


      /** \brief Mark \a client as ready for next request */
      void ready (client *c) {
          // we have to wake up listener, so we will send client address
          // through socketpair. It works because we have common address-space
          // nasty but neat :)
          size_t l = pair_.second.send((char *)&c, sizeof(c));
          assert(l == sizeof(c));
          l = 0;
      }


  protected:

      /** \brief Accept and register new client for listening */
      void new_client ();


      /** \brief Unregister client and close connection */
      void delete_client (client *c) {
          info() << c->peer() << ": disconnect";
          poller_.erase(c);
          c->close();
      }


      /** \brief Handle request sent by client */
      bool handle_request (client *c);


      /** \brief In case of local socket make it world R/W */
      template <class T>
      void chmod_listener (common::net_space::address &a);


      /**
       * \brief In case of local socket, remove rendevouz point during
       * shutdown
       */
      template <class T>
      void remove_listener_socket ();


  private:

      common::net_space::address address_;

      typedef STD_ vector<client *> client_list;
      client_list clients_;

      typedef cool::poller poller;
      poller poller_;

      // socket-pair trick to communicate between listener and client from
      // different threads
      cool::socket_pair<cool::local> pair_;
};


template <> inline void
listener::chmod_listener<cool::inet::listener> (common::net_space::address &a)
{
  // do nothing if AF_INET socket
}


template <> inline void
listener::chmod_listener<cool::local::listener> (common::net_space::address &a)
{
  // make socket rw-rw-rw-
  static const int all_rw = S_IREAD|S_IWRITE|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH;
  chmod(a.c_str(), all_rw);
}


inline
listener::listener (common::net_space::address &a, unsigned short max_clients)
    : common::net_space::listener(a)
    , address_(a)
    , pair_()
{
  chmod_listener<common::net_space::listener>(address_);

  // register itself as listener
  poller_.insert(this);

  // register socket pair first as receiver in communication between listener
  // and partition handler thread
  poller_.insert(&pair_.first);

  // pre-create array of clients
  while (max_clients--)
      clients_.push_back(new client(this, config::max_request));
}


template <> inline void
listener::remove_listener_socket<cool::inet::listener> ()
{
  // do nothing if AF_INET socket
}


template <> inline void
listener::remove_listener_socket<cool::local::listener> ()
{
  // remove socket inode if AF_UNIX socket
  unlink(address_.c_str());
}


inline
listener::~listener () throw()
{
  common::net_space::listener::close();
  remove_listener_socket<common::net_space::listener>();
}


}   // namespace pgcpool


#endif  // pool_listener_hpp_included
