#include <pool/logger.hpp>
#include <pool/config.hpp>
#include <iostream>
#include <ctime>
#include <cstring>
#include <pthread.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <syslog.h>


namespace pgcpool
{


// auxiliary stream buffer that does nothing
class null_streambuf: public std::streambuf
{
  protected: 

      virtual int sync () { return 0; }
};


// create null stream that does nothing
static null_streambuf nsb;
static std::ostream null_stream(&nsb);


std::string log_stream::log_destination_;
bool log_stream::log_is_stdout_ = true;
log_level::value log_stream::log_level_ = log_level::debug;
std::ostream &log_stream::log_null_ = null_stream;


static inline void
drop_dead (const char *function, const std::string &info)
{
  std::string m = std::string(function) + '(' + info + "): " + strerror(errno);
  ::openlog(config::app.c_str()
      , LOG_CONS | LOG_NDELAY | LOG_PID | LOG_PERROR
      , LOG_USER);
  ::syslog(LOG_ALERT, m.c_str());
  ::closelog();
  ::abort();
}


log_stream::~log_stream () throw()
{
  if (level_ == log_level::off)
      // no logging at all
      return;

  // create log string with timestamp and level
  char timebuf[36];
  time_t t = time(0);
  strftime(timebuf, sizeof(timebuf), "%Y.%m.%d %H:%M:%S ", localtime(&t));

  static char level_char[] = "FEWID";
  std::string ev = std::string(timebuf) + level_char[level_] + '\t'
      + str() + '\n';

  if (log_is_stdout_) {
      // to console
      std::cout.write(ev.c_str(), ev.size()).flush();
      return;
  }

  // to file -- we are opening/writing/closing log for each event
  // open
  int f = ::open(log_destination_.c_str()
      , O_WRONLY | O_APPEND | O_CREAT | O_LARGEFILE
      , S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
  if (f == -1)
      drop_dead("open", log_destination_);

  // write
  ssize_t rc = ::write(f, ev.data(), ev.size());
  if (rc == -1 || rc != (ssize_t)ev.size()) {
      ::close(f);
      if (rc >= 0)
          errno = ENOSPC;
      drop_dead("write", log_destination_);
  }

  // close
  if (::close(f) == -1)
      drop_dead("close", log_destination_);
}


}   // namespace pgcpool
