#ifndef pool_logger_hpp_included
#define pool_logger_hpp_included


/**
 * \file pool/logger.hpp
 * \brief Logging facilities.
 */

#include <pool/pgcpoold.hpp>
#include <pool/error.hpp>
#include <string>
#include <ostream>
#include <sstream>


namespace pgcpool
{


/** \brief Logging level values */
struct log_level
{
  enum value {
      fatal
      , error
      , warning
      , info
      , debug
      , off
  };
};


/** \brief Base class for loggers */
class log_stream: public std::ostringstream
{
  public:

      /**
       * \brief Configure logging system
       *
       * \a events are logged into file \a destination. If destination is \c
       * "stdout" then events are written to screen.
       */
      static void
      configure (const std::string &destination, log_level::value level) {
          log_destination_ = destination;
          log_is_stdout_ = log_destination_ == "stdout";
          log_level_ = level;
      }


      /** \brief Create temporary logger stream. Event is logged at \a level */
      log_stream (log_level::value level)
          : level_(level)
      {
          if (level_ > log_level_)
              level_ = log_level::off;
      }


      /** \brief Dtor -- write accumulated log events into \e destination */
      ~log_stream () throw();


      /**
       * \brief Workaround for inserters that require \c "std::ostream &"
       * instead of \c "const std::ostream &"
       */
      template <typename T>
      std::ostream &
      operator<< (const T &t) {
          if (level_ == log_level::off)
              return log_null_;
          return static_cast<std::ostream &>(*this) << t;
      }


  private:

      static std::string log_destination_;
      static bool log_is_stdout_;
      static log_level::value log_level_;
      static std::ostream &log_null_;

      log_level::value level_;
};


/** \brief Convenience class: temporary logger at fatal level */
class fatal: public log_stream
{
  public:

      fatal (): log_stream(log_level::fatal) {}
};


/** \brief Convenience class: temporary logger at error level */
class error: public log_stream
{
  public:

      error (): log_stream(log_level::error) {}
};


/** \brief Convenience class: temporary logger at warning level */
class warning: public log_stream
{
  public:

      warning (): log_stream(log_level::warning) {}
};


/** \brief Convenience class: temporary logger at info level */
class info: public log_stream
{
  public:

      info (): log_stream(log_level::info) {}
};


/** \brief Convenience class: temporary logger at debug level */
class debug: public log_stream
{
  public:

      debug (): log_stream(log_level::debug) {}
};


}  // namespace pgcpool


#endif  // pool_logger_hpp_included
