#include <pool/partitions.hpp>
#include <pool/config.hpp>
#include <pool/logger.hpp>
#include <pool/listener.hpp>
#include <pool/signal.hpp>
#include <cool/thread/thread_list.hpp>
#include <cool/thread/barrier.hpp>
#include <boost/lexical_cast.hpp>
#include <sstream>
#include <list>
#include <fstream>


namespace pgcpool
{


partitions::cluster_map_type partitions::cluster_map;
partitions::proxy_connection_map_type partitions::proxy_connection;


// collection of internal threads (one per partition)
typedef std::list<partitions::handler *> handler_list;
static handler_list handlers;
static cool::thread_list threads;
static cool::barrier *barrier = 0;

// handler threads status
static enum {
    init        // initialize subsystem
    , stopped   // threads not started
    , running   // ready to dispatch queries
    , reconf    // stop handlers, read new configuration and start again
} handlers_state = init;


inline void
wait_until_ready (pgclient::connection &c)
{
  while (!c.wait(1).ready())
      ;
}


bool
partitions::load_conf (const std::string &connect_string
    , const std::string &self)
{
  cluster_map.clear();
  proxy_connection.clear();
  debug() << "partitions::load_conf(" << self << "): '"
      << connect_string << "'";

  // connect database
  pgclient::connection proxydb(connect_string.c_str());
  wait_until_ready(proxydb);

  // connections to partitions
  std::stringstream confstr;
  std::string q = "SELECT "
      "id_partition, conn_count_min, conn_count_max"
      ", connect_string, partition_name "
      "FROM plproxy.get_proxy_connections('" + self + "')";
  pgclient::result r(q.c_str());
  wait_until_ready(proxydb.exec(&r));
  if (r.empty()) {
      fatal() << "plproxy.get_proxy_connections: empty";
      return false;
  }
  confstr << "proxy_connection(" << r.size() << "): ";
  for (int i = 0;  i < r.size();  ++i) {
      id_type partition_id = cool::as<uint32_t>(r[i]["id_partition"].c_str());

      proxy_connection_type conn_data;
      conn_data.min_count = cool::as<uint16_t>(r[i]["conn_count_min"].c_str());
      conn_data.max_count = cool::as<uint16_t>(r[i]["conn_count_max"].c_str());
      conn_data.conn_string = r[i]["connect_string"].c_str();
      conn_data.partition_name = r[i]["partition_name"].c_str();
      conn_data.connection = 0;

      proxy_connection[partition_id] = conn_data;

      confstr << "\n\t* " << conn_data.partition_name
          << '[' << partition_id << "] -> '"
          << conn_data.conn_string << "'; "
          << conn_data.min_count << ".." << conn_data.max_count;
  }
  debug() << confstr.str();
  confstr.str("");

  // active cluster configurations
  q = "SELECT cluster_name FROM plproxy.get_active_confs('" + self + "')";
  r.sql(q.c_str());
  wait_until_ready(proxydb.exec(&r));
  if (r.empty()) {
      fatal() << "plproxy.get_active_confs: empty";
      return false;
  }
  confstr << "cluster_map(" << r.size() << "):";
  for (int i = 0;  i < r.size();  ++i) {
      std::string cluster_name = r[i]["cluster_name"].c_str();
      confstr << "\n\t* '" << cluster_name << "': ";

      // load cluster mapping
      q = "SELECT conf_map, hash_nr, key_partition"
          " FROM plproxy.get_cluster_maps('" + cluster_name + "')";
      pgclient::result maps(q.c_str());
      wait_until_ready(proxydb.exec(&maps));
      if (maps.empty()) {
          confstr << "empty";
          continue;
      }

      for (int j = 0;  j < maps.size();  ++j) {
          cluster_map_key_type cmk;
          cmk.conf_map = maps[j]["conf_map"].c_str();
          cmk.hash_nr = cool::as<uint16_t>(maps[j]["hash_nr"].c_str());

          id_type partition_id =
              cool::as<id_type>(maps[j]["key_partition"].c_str());
          if (proxy_connection.find(partition_id) == proxy_connection.end()) {
              fatal() << "integrity_error";
              return false;
          }

          confstr << "\n\t  ('" << cmk.conf_map << "'; "
              << cmk.hash_nr << ") -> " << partition_id;

          cluster_map[cluster_name][cmk] = partition_id;
      }

      debug() << confstr.str();
      confstr.str("");
  }
  handlers_state = reconf;

  return true;
}


struct delete_handler
{
  void operator() (partitions::handler *ph) {
      delete ph;
  }
};


static char infofmt[] =
"%s, "__DATE__" "__TIME__"\n"
"\n"
"Startup time: %s (%ld sec ago)\n"
"Last statistics gathered since: %s (%ld sec ago)\n"
"\n"
"Last: Activity duration: %lld.%06dsec (avg %.2fusec/query incl connect & reset)\n"
"Last: Query count: %ld\n"
"Last: Failure count: %ld\n"
"\n"
"Total: Number of connections to partition (working/total): %ld/%ld (%d..%d)\n"
"Total: Activity duration: %lld.%06dsec (avg %.2fusec/query incl connect & reset)\n"
"Total: Query count: %ld\n"
"Total: Failure count: %ld\n";


void
partitions::handler::operator() ()
{
  try {
      assert(barrier != 0);

      // set statistics file name
      std::string statistics_file = config::statistics_prefix +
          boost::lexical_cast<std::string>(conf.second.partition_name);

      // signal caller
      barrier->wait();

      // running
      time_t next_stats = 0;

      debug() << "part[" << conf.second.partition_name << "] ready";
      while (!signal_handler::shutdown && handlers_state == running) {
          try {
              pool_.wait(1);
          }
          catch (const pgclient::error &e) {
              error() << e;
          }
          catch (const cool::interrupted &e) {
              continue;
          }
          catch (...) {
              warning() << "partitions::handler()";
              throw;
          }

          if (config::statistics_interval > 0) {
              // show statistics if configured so
              time_t now = time(0);
              if (next_stats <= now) {
                  next_stats = now + config::statistics_interval;

                  // get statistics
                  pgclient::pool::statistics_type total, last;
                  total = pool_.total_stats();
                  last = pool_.last_stats();

                  struct tm totaltm;
                  localtime_r(&total.startup, &totaltm);
                  char totalb[sizeof("YYYYMMDD HHMMSS")];
                  strftime(totalb, sizeof(totalb), "%Y%m%d %H%M%S", &totaltm);

                  struct tm lasttm;
                  localtime_r(&last.startup, &lasttm);
                  char lastb[sizeof("YYYYMMDD HHMMSS")];
                  strftime(lastb, sizeof(lastb), "%Y%m%d %H%M%S", &lasttm);

                  long lqc = last.query_count ? last.query_count : 1;
                  long tqc = total.query_count ? total.query_count : 1;
                  long count = pool_.size();
                  long working = count - pool_.idle();

                  char info[sizeof(infofmt) + 1024];
                  size_t len = snprintf(info, sizeof(info), infofmt
                          , config::app.c_str()
                          , totalb
                          , now - total.startup
                          , lastb
                          , now - last.startup
                          , last.activity_duration_sec
                          , last.activity_duration_usec
                          , (float)(last.activity_duration_usec
                                    + last.activity_duration_sec*1000000)
                              /lqc
                          , last.query_count
                          , last.failure_count
                          , working
                          , count
                          , conf.second.min_count
                          , conf.second.max_count
                          , total.activity_duration_sec
                          , total.activity_duration_usec
                          , (float)(total.activity_duration_usec
                                    + total.activity_duration_sec*1000000)
                              /tqc
                          , total.query_count
                          , total.failure_count
                          );

                  // write to file
                  FILE *f = fopen(statistics_file.c_str(), "w");
                  if (f) {
                      if (fwrite(info, len, 1, f) != 1)
                          warning() << statistics_file <<
                              ": error writing: " << strerror(errno);
                      fclose(f);
                  }
                  else
                      warning() << statistics_file <<
                          ": error opening: " << strerror(errno);
              }
          }
      }

      // stopped
      debug() << "part[" << conf.second.partition_name << "] stopped";
  }
  catch (...) {
      // do not let exceptions to escape from thread
      // log and exit
      exception::global_handler();
  }
}


void
partitions::start_handlers ()
{
  if (handlers_state == init) {
      // initialize handlers (do nothing because conf not arrived yet) */
      return;
  }

  debug() << "partitions::start_handlers";

  // create and start handlers
  barrier = new cool::barrier(proxy_connection.size() + 1);
  proxy_connection_map_type::iterator pc;
  for (pc = proxy_connection.begin()  ;  pc != proxy_connection.end();  ++pc) {
      pc->second.connection = new handler(*pc);
      handlers.push_back(pc->second.connection);
      threads.create(pc->second.connection);
  }

  // wait until all partition threads are running
  handlers_state = running;
  barrier->wait();

  // give threads chance to finish their barrier->wait() first
  // XXX This isn't theoretically correct -- it is still possible to start
  // some query before partition handler is ready to accept queries.
  cool::thread::yield();
}


void 
partitions::stop_handlers ()
{
  if (handlers_state == stopped)
      // not started yet
      return;

  if (handlers_state == running)
      debug() << "partitions::stop_handlers";

  // wait until all handler threads are stopped
  threads.join_all();

  // now it is safe to delete all of them
  std::for_each(handlers.begin(), handlers.end(), delete_handler());

  handlers_state = stopped;
}


void
partitions::queue_query (const protocol::query_type &query, client *c)
{
  // get cluster
  partitions::cluster_map_type::const_iterator cluster =
      partitions::cluster_map.find(c->cluster);
  if (cluster == partitions::cluster_map.end()) {
      // requested cluster not found
      warning() << c->cluster << ": cluster not found";
      c->response(protocol::rsp_enocluster);
      return;
  }

  // get partition
  partitions::cluster_type::const_iterator partition =
      cluster->second.find(c->hash_key);
  if (partition == cluster->second.end()) {
      warning()
          << '(' << c->hash_key.conf_map << ';' << c->hash_key.hash_nr << ')'
          << ": partition not found";
      c->response(protocol::rsp_enopartition);
      return;
  }

  // get connection for partition
  partitions::proxy_connection_map_type::const_iterator connection =
      partitions::proxy_connection.find(partition->second);
  if (connection == partitions::proxy_connection.end()) {
      warning() << partition->second << ": connection not found";
      c->response(protocol::rsp_enoconn);
      return;
  }

  debug() << c->peer() << ": cmd_exec_query("
      << c->query << ") @"
      << c->cluster << '('
      << c->hash_key.conf_map << ';'
      << c->hash_key.hash_nr << "): "
      << connection->second.partition_name;

  // start query in partition handler thread
  c->sql(c->query.c_str());
  connection->second.connection->exec(c);
}


}   // namespace pgcpool
