#ifndef pool_partitions_hpp_included
#define pool_partitions_hpp_included

/**
 * \file pool/partitions.hpp
 * \brief Collections of connections to partitions.
 */


#include <pool/pgcpoold.hpp>
#include <pool/config.hpp>
#include <common/protocol.hpp>
#include <pool/logger.hpp>
#include <pgclient/pool.hpp>
#include <cool/thread/thread.hpp>
#include <string>
#include <map>


namespace pgcpool
{

using namespace common;

class client;


/** \brief Partitions manager */
class partitions
{
  public:


      /** \brief Load partitions configuration from database */
      static bool load_conf (const std::string &connect_string
          , const std::string &self);


      /** \brief Start partition threads. */
      static void start_handlers ();


      /** \brief Stop partition threads. */
      static void stop_handlers ();


      // beware! hairy hash mapping follows:

      /** \brief ID binding mapping to partition connection */
      typedef uint32_t id_type;


      class handler;


      /** \brief Connection data */
      struct proxy_connection_type
      {
          uint16_t min_count;
          uint16_t max_count;
          std::string conn_string;
          std::string partition_name;
          handler *connection;
      };


      /** \brief Mappings between ID and connections */
      typedef std::map<
          id_type
          , proxy_connection_type
      > proxy_connection_map_type;


      /** \brief Collection of partition connections */
      static proxy_connection_map_type proxy_connection;


      /** \brief Connection selector key */
      struct cluster_map_key_type
      {
          std::string conf_map;
          uint16_t hash_nr;
      };


      struct less_cluster_map_key_type
          : public std::binary_function<
              cluster_map_key_type, cluster_map_key_type, bool>
      {
          bool operator() (const cluster_map_key_type &k1
                  , const cluster_map_key_type &k2) const {
              signed int r = k1.hash_nr - k2.hash_nr;
              if (r == 0)
                  r = k1.conf_map < k2.conf_map ? -1 : 0;
              return r < 0;
          }
      };


      /** \brief Bindings between selector and connection id */
      typedef std::map<
          cluster_map_key_type
          , id_type
          , less_cluster_map_key_type
      > cluster_type;


      /** \brief Map cluster name to list of cluster maps */
      typedef std::map<
          std::string
          , cluster_type
      > cluster_map_type;


      /** \brief Collection of bindings between selector and connection id */
      static cluster_map_type cluster_map;


      static void queue_query (const protocol::query_type &query, client *c);


      /** \brief Partition handler (thread per partition) */
      class handler: public cool::thread::handler
      {
        public:

            /** \brief Instantiate handler for connection \a conn_info */
            handler (const proxy_connection_map_type::value_type &conn_info)
                : conf(conn_info)
                , pool_(conf.second.conn_string.c_str()
                        , conf.second.min_count, conf.second.max_count
                        , config::reconnect_interval
                        , config::lifetime)
            { }


            virtual ~handler () { }


            /** \brief Thread worker */
            void operator() ();


            /** \brief Start executing request in handler's thread context */
            void exec (pgclient::pool::result *c) {
                pool_.exec_thread(static_cast<pgclient::pool::result *>(c));
            }


        private:

            /** \brief Reference to proxy connection */
            const proxy_connection_map_type::value_type &conf;

            /** \brief Connection pooler to PostgreSQL */
            pgclient::pool pool_;
      };
};


}   // namespace pgcpool


#endif  // pool_partitions_hpp_included
