CREATE OR REPLACE FUNCTION plproxy.alter_cluster_split(
    i_cluster_name text, 
    i_conf_name text,   
    i_partition_list text[]) -- list of the partitions that must be split
    RETURNS text AS $$
-- used by:: admin when new partitions must be added into cluster
--   first active configuration must be copied and 
--   then it can be split into more partitons
-- br:: newly created partitions stay in the same map as old oens were
-- br:: new partitions are assigned first available free names in given cluster
-- returns:: 200 Ok
-- example::
--   select * from plproxy.alter_cluster_split('eventdb','eventdb_V2', array['eventdb_p3','eventdb_p4']);
DECLARE
    r_partition record;
    r record;
    v_id_cluster int;
    v_id_partition int;
    v_partition_name text;
    v_count int = 1;
    v_id_proxy int;
    v_min_conn int;
    v_max_conn int;
BEGIN	
    -- check cluster
    select id_cluster into v_id_cluster from plproxy.db_cluster where cluster_name = lower(i_cluster_name);
    if NOT found  then 
        raise exception '400 Cluster with given name="%" not found', i_cluster_name;
    end if;
    -- check that conf is not active
    if exists (
        select 1 from plproxy.db_conf
	where cluster_name = lower(i_cluster_name)
	and conf_name = lower(i_conf_name) )
    then
        raise exception '400 Active configuration can NOT be modified.!';	
    end if;
    -- check that conf actualy exists and belongs to given cluster
    if not exists (
        select 1 from plproxy.db_cluster_conf 
	where conf_name = lower(i_conf_name)
	and key_cluster = v_id_cluster )
    then 
        raise exception '400 Configuration with givent name not found or belongs to another cluster. Check names!';
    end if;
    -- 
    for i in array_lower(i_partition_list,1) .. array_upper(i_partition_list,1) loop
        r_partition = plproxy.fetch_partition_by_name(i_partition_list[i]);
        v_partition_name = plproxy.fetch_free_partition_name(i_cluster_name);  
        v_id_partition = nextval('plproxy.proxy_seq');
        insert into plproxy.db_partition(id_partition, key_cluster, partition_name)
        values(v_id_partition, v_id_cluster, v_partition_name);

        -- also create proxy connections here
        select key_proxy, conn_count_min, conn_count_max into v_id_proxy, v_min_conn, v_max_conn 
        from plproxy.db_proxy_connection
        where key_partition = r_partition.id_partition;

        insert into plproxy.db_proxy_connection (key_proxy, key_partition, conn_count_min, conn_count_max)
        values ( v_id_proxy, v_id_partition, v_min_conn, v_max_conn);
     
        for r in 
            select id_conf_part from plproxy.db_conf_partition 
            where key_cluster_conf in (
                select id_cluster_conf from plproxy.db_cluster_conf 
                where conf_name = lower(i_conf_name) )
            and key_partition = r_partition.id_partition
        loop
            -- move every second record into new partition
            if v_count % 2 = 0 then
                update plproxy.db_conf_partition set
                    key_partition = v_id_partition 
                where id_conf_part  = r.id_conf_part;
            end if;
            v_count  = v_count + 1;
        end loop;
        if not found then
            raise exception '400 No records found for partition %.', r_partition.partition_name;
        end if;
    end loop;
    return '200 Ok';
END;
$$ LANGUAGE 'plpgsql' VOLATILE SECURITY DEFINER;
