CREATE SCHEMA plproxy;

SET client_encoding = 'SQL_ASCII';
SET check_function_bodies = false;
SET client_min_messages = warning;


SET search_path = plproxy, pg_catalog;


CREATE TYPE ret_get_active_confs AS (
	cluster_name text,
	conf_name text
);




CREATE TYPE ret_get_cluster_map AS (
	conf_map text,
	hash_nr smallint,
	key_partition integer
);




CREATE TYPE ret_get_cluster_maps AS (
	conf_map text,
	hash_nr smallint,
	key_partition integer
);




CREATE TYPE ret_get_cluster_partitions AS (
	id_partition integer,
	connect_string text
);




CREATE TYPE ret_get_cluster_proxies AS (
	proxy_name text,
	proxy_status text,
	connect_string text
);




CREATE TYPE ret_get_conf_partitions AS (
	partition_name text,
	hash_count bigint,
	connect_string text
);




CREATE TYPE ret_get_hash AS (
	conf_map text,
	hash_nr smallint
);




CREATE TYPE ret_get_proxy_config AS (
	name text,
	value text
);




CREATE TYPE ret_get_proxy_connections AS (
	id_partition integer,
	conn_count_min smallint,
	conn_count_max smallint,
	connect_string text,
	partition_name text
);




CREATE TYPE ret_view_cluster_proxies AS (
	proxy_name text,
	proxy_status text,
	connect_string text
);




CREATE TYPE ret_view_conf_partitions AS (
	partition_name text,
	hash_count bigint,
	connect_string text
);




CREATE TYPE ret_view_partiton_changes AS (
	source_name text,
	source_connect text,
	target_name text,
	target_connect text
);

CREATE OR REPLACE FUNCTION plproxy.alter_cluster_copy_conf(
    i_cluster_name text, 
    i_conf_name text) 
RETURNS text AS $$
-- used by:: admins to start working on new version of the cluster configuration
--   creates copy of existing configuration (all maps are copied also)
-- exceptions:
--   400 Cluster with given name NOT found
--   400 Configuration NOT found
-- returns:: 200 Ok
-- example:: select * from plproxy.copy_conf('eventdb','eventdb_V1');  
DECLARE
    v_id_cluster int;
    r_conf record;
    r_cluster record;
    v_id_cluster_conf int;
    v_conf_name text;
    v_version int;
BEGIN	
    select id_cluster into v_id_cluster from plproxy.db_cluster where cluster_name = lower(i_cluster_name);
    if v_id_cluster is null then 
        raise exception '400 Cluster with given name="%" not found', i_cluster_name;
    end if;

    for r_conf in 
        select * from plproxy.db_cluster_conf where conf_name = i_conf_name and key_cluster = v_id_cluster
    loop
        -- create new configuration
        v_id_cluster_conf = nextval('plproxy.proxy_seq');
        select coalesce( max(version) , 0 ) + 1 into v_version from plproxy.db_cluster_conf where key_cluster = v_id_cluster;
        v_conf_name = i_cluster_name || '_v' || v_version::text;
        insert into plproxy.db_cluster_conf(id_cluster_conf, key_cluster, conf_name, version)
        values(v_id_cluster_conf, r_conf.key_cluster, v_conf_name, v_version);
    
        -- copy database partitioning information
        insert into plproxy.db_conf_partition(key_cluster_conf, key_partition, hash_nr)
        select v_id_cluster_conf, key_partition, hash_nr
        from plproxy.db_conf_partition
        where key_cluster_conf = r_conf.id_cluster_conf;
    end loop;
    
    return '200 Ok';
END;
$$ LANGUAGE 'plpgsql' VOLATILE SECURITY DEFINER;

CREATE FUNCTION alter_cluster_drop_conf(i_cluster_name text, i_conf_name text, i_conf_map text) RETURNS text
    AS $$
-- used by:: admin to remove all data for given cluster configuration
-- br:: active cluster configurations can not be deleted just for safety reasons
-- br:: proxy status must be OFFLINE
-- exceptions:
-- 400 Cluster with given name not found
-- 400 Active cluster configurations must NOT be deleted
-- returns:: 200 Ok
-- example:: select * from plproxy.alter_cluster_drop_conf('eventdb_v1');  
DECLARE
    r_cluster record;
    r_conf record;
BEGIN	
    r_cluster = plproxy.fetch_cluster_by_name(i_cluster_name);
    r_conf = plproxy.fetch_cluster_conf_by_name(i_conf_name, i_conf_map);
    if exists (
        select 1 from plproxy.db_conf 
	where cluster_name = lower(i_cluster_name)
	and conf_name = lower(i_conf_name) )
    then
        raise exception '400 Active configuration can NOT be modified!';
    end if;

    -- delete mappings
    delete from plproxy.db_conf_partition where key_cluster_conf = r_conf.id_cluster_conf;
    -- delete configuration
    delete from plproxy.db_cluster_conf where id_cluster_conf = r_conf.id_cluster_conf;

    return '200 ok';
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;

CREATE OR REPLACE FUNCTION plproxy.alter_cluster_split(
    i_cluster_name text, 
    i_conf_name text,   
    i_partition_list text[]) -- list of the partitions that must be split
    RETURNS text AS $$
-- used by:: admin when new partitions must be added into cluster
--   first active configuration must be copied and 
--   then it can be split into more partitons
-- br:: newly created partitions stay in the same map as old oens were
-- br:: new partitions are assigned first available free names in given cluster
-- returns:: 200 Ok
-- example::
--   select * from plproxy.alter_cluster_split('eventdb','eventdb_V2', array['eventdb_p3','eventdb_p4']);
DECLARE
    r_partition record;
    r record;
    v_id_cluster int;
    v_id_partition int;
    v_partition_name text;
    v_count int = 1;
    v_id_proxy int;
    v_min_conn int;
    v_max_conn int;
BEGIN	
    -- check cluster
    select id_cluster into v_id_cluster from plproxy.db_cluster where cluster_name = lower(i_cluster_name);
    if NOT found  then 
        raise exception '400 Cluster with given name="%" not found', i_cluster_name;
    end if;
    -- check that conf is not active
    if exists (
        select 1 from plproxy.db_conf
	where cluster_name = lower(i_cluster_name)
	and conf_name = lower(i_conf_name) )
    then
        raise exception '400 Active configuration can NOT be modified.!';	
    end if;
    -- check that conf actualy exists and belongs to given cluster
    if not exists (
        select 1 from plproxy.db_cluster_conf 
	where conf_name = lower(i_conf_name)
	and key_cluster = v_id_cluster )
    then 
        raise exception '400 Configuration with givent name not found or belongs to another cluster. Check names!';
    end if;
    -- 
    for i in array_lower(i_partition_list,1) .. array_upper(i_partition_list,1) loop
        r_partition = plproxy.fetch_partition_by_name(i_partition_list[i]);
        v_partition_name = plproxy.fetch_free_partition_name(i_cluster_name);  
        v_id_partition = nextval('plproxy.proxy_seq');
        insert into plproxy.db_partition(id_partition, key_cluster, partition_name)
        values(v_id_partition, v_id_cluster, v_partition_name);

        -- also create proxy connections here
        select key_proxy, conn_count_min, conn_count_max into v_id_proxy, v_min_conn, v_max_conn 
        from plproxy.db_proxy_connection
        where key_partition = r_partition.id_partition;

        insert into plproxy.db_proxy_connection (key_proxy, key_partition, conn_count_min, conn_count_max)
        values ( v_id_proxy, v_id_partition, v_min_conn, v_max_conn);
     
        for r in 
            select id_conf_part from plproxy.db_conf_partition 
            where key_cluster_conf in (
                select id_cluster_conf from plproxy.db_cluster_conf 
                where conf_name = lower(i_conf_name) )
            and key_partition = r_partition.id_partition
        loop
            -- move every second record into new partition
            if v_count % 2 = 0 then
                update plproxy.db_conf_partition set
                    key_partition = v_id_partition 
                where id_conf_part  = r.id_conf_part;
            end if;
            v_count  = v_count + 1;
        end loop;
        if not found then
            raise exception '400 No records found for partition %.', r_partition.partition_name;
        end if;
    end loop;
    return '200 Ok';
END;
$$ LANGUAGE 'plpgsql' VOLATILE SECURITY DEFINER;


CREATE FUNCTION alter_partition(i_partition_name text, i_connect_string text, i_partition_descr text) RETURNS text
    AS $$
-- USED BY: admin to chnage partition connect string or description
-- RETURNS: 200 Ok
-- EXAMPLE: select * from plproxy.alter_partition('userdb_p1', 'dbname=userdb_p1', null);  
DECLARE
    r_partition record;
BEGIN	
    -- get partition record
    r_partition = plproxy.fetch_partition_by_name(i_partition_name);  

    if i_partition_descr is NOT null then
        r_partition.partition_descr = i_partition_descr;
    end if;

    if i_connect_string is NOT null then
        r_partition.connect_string  = i_connect_string;
    end if;

    update plproxy.db_partition set
        partition_descr = r_partition.partition_descr ,
        connect_string  = r_partition.connect_string  
    where id_partition = r_partition.id_partition;
    
    return '200 Ok';
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;




CREATE FUNCTION alter_proxy(i_proxy_name text, i_name text, i_value text) RETURNS text
    AS $$
-- USED BY: admin to chnage proxy connect string or description
-- RETURNS: 200 Ok
-- EXAMPLE: select * from plproxy.alter_proxy('proxydb_p1', 'dbname=proxydb_p1', null);  
DECLARE
    r_proxy record;
BEGIN	
    -- get proxy record
    r_proxy = plproxy.fetch_proxy_by_name(i_proxy_name);  

    update plproxy.db_proxy_config
       set value = i_value
     where key_proxy = r_proxy.id_proxy
       and name = i_name;
    
    if NOT found then 
        insert into plproxy.db_proxy_config (key_proxy, name, value) values (r_proxy.id_proxy, i_name, i_value);
    end if;
    
    return '200 Ok';
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;



CREATE OR REPLACE FUNCTION plproxy.create_cluster(
    i_cluster_name text,
    i_cluster_description text,
    i_db_count int)
    RETURNS text AS $$
-- used by:: admins to create new clusters
--   creates configuration records for new database cluster set
-- exceptions:
-- returns:: 200 Ok
-- example:: select * from plproxy.create_cluster('userdb','Skype user database',3);  
DECLARE
    C_BITS constant int = 8;
    C_PARTITIONS constant int = 256;
    
    v_id_cluster int;
    v_id_cluster_conf int;
    v_id_partition int;
    r record;
BEGIN	
    -- cluster
    v_id_cluster = nextval('plproxy.proxy_seq');
    insert into plproxy.db_cluster(id_cluster, cluster_name, cluster_descr, cluster_bits, cluster_partitions)
    values(v_id_cluster, lower(i_cluster_name), i_cluster_description, C_BITS, C_PARTITIONS);

    -- cluster conf
    v_id_cluster_conf = nextval('plproxy.proxy_seq');
    insert into plproxy.db_cluster_conf(id_cluster_conf, key_cluster, conf_name, version)
    values(v_id_cluster_conf, v_id_cluster, lower(i_cluster_name) || '_v1' , 1);

    for i in 1 .. i_db_count loop
        -- database
        v_id_partition = nextval('plproxy.proxy_seq');
        insert into plproxy.db_partition(id_partition, key_cluster, partition_name)
        values(v_id_partition, v_id_cluster, plproxy.fetch_free_partition_name(i_cluster_name));
        
        -- database mapping
        for r in select generate_series(i,C_PARTITIONS,i_db_count) as partnr loop
            insert into plproxy.db_conf_partition(key_cluster_conf, key_partition, hash_nr)
            values(v_id_cluster_conf, v_id_partition, r.partnr-1);
        end loop;
    end loop;
    return '200 Ok';
END;
$$ LANGUAGE 'plpgsql' VOLATILE SECURITY DEFINER;




CREATE FUNCTION create_proxy(i_proxy_name text, i_proxy_descr text, i_cluster_list text[], i_conn_count_min integer, i_conn_count_max integer) RETURNS text
    AS $$
-- USED BY: admins when new proxy database and pool server is created
--   creates configuration records for new proxy server
-- RETURNS: 200 Ok
-- EXAMPLE: select * from plproxy.create_proxy('eventfe','EventDB and ContactDB',ARRAY['eventdb','contactdb','userdb'],1,100);  
DECLARE
    v_id_proxy int;
    r_cluster record;
    r_part record;
BEGIN	
    select id_proxy from plproxy.db_proxy where proxy_name = i_proxy_name into v_id_proxy;
    if not found then
        -- create proxy record
        v_id_proxy = nextval('plproxy.proxy_seq');
        insert into plproxy.db_proxy(id_proxy, proxy_name, proxy_descr)
        values(v_id_proxy, i_proxy_name, i_proxy_descr);
    end if;

    -- for each cluster in array add configuration records to connection pool
    for i in array_lower(i_cluster_list,1) .. array_upper(i_cluster_list,1) loop
        r_cluster = plproxy.fetch_cluster_by_name(i_cluster_list[i]);  
        for r_part in
            select id_partition from plproxy.db_partition where key_cluster = r_cluster.id_cluster
        loop
            insert into plproxy.db_proxy_connection(key_proxy, key_partition, conn_count_min, conn_count_max)
            values (v_id_proxy, r_part.id_partition, i_conn_count_min, i_conn_count_max);
        end loop;
    end loop;
    
    perform plproxy.alter_proxy(i_proxy_name, 'log-file', 'stdout');
    perform plproxy.alter_proxy(i_proxy_name, 'log-level', '3');
    perform plproxy.alter_proxy(i_proxy_name, 'daemon', 'false');
    perform plproxy.alter_proxy(i_proxy_name, 'listen', '/tmp/.plproxy.sock');
    perform plproxy.alter_proxy(i_proxy_name, 'max-clients', '500');
    perform plproxy.alter_proxy(i_proxy_name, 'max-request', '128');
    perform plproxy.alter_proxy(i_proxy_name, 'reconnect-interval', '1');
    perform plproxy.alter_proxy(i_proxy_name, 'lifetime', '30');
    perform plproxy.alter_proxy(i_proxy_name, 'statistics-interval', '60');
    perform plproxy.alter_proxy(i_proxy_name, 'statistics-prefix', '/tmp/');

    return '200 Ok';
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;


CREATE OR REPLACE FUNCTION plproxy.deactivate_cluster_conf(
    i_cluster_name text , 
    i_conf_name text ) 
RETURNS text AS $$
-- used by:: admins when when cluster goes out of use
--   deactivates configuration so it can be deleted
-- exceptions:
--   400 Cluster with given name not found
--   400 Active cluster configurations must NOT be deleted
-- returns:: 200 Ok
-- example:: select * from plproxy.deactivate_cluster_conf(13, 4);  
DECLARE
    r_cluster record;
    r_conf record;
BEGIN	
    select * into r_cluster from plproxy.fetch_cluster_by_name(i_cluster_name);
    
    if not exists(
        select * from plproxy.db_cluster_conf 
        where conf_name = lower(i_conf_name) 
        and  key_cluster = r_cluster.id_cluster)
    then
        raise exception '400 Configuration with name="%" not found for cluster="%"!', i_conf_name, i_cluster_name;
    end if;

    delete from plproxy.db_conf where cluster_name = lower(i_cluster_name) and conf_name = lower(i_conf_name);

    return '200 Ok';
END;
$$ LANGUAGE 'plpgsql' VOLATILE SECURITY DEFINER;

CREATE OR REPLACE FUNCTION plproxy.delete_proxy(
    i_proxy_name text)
    RETURNS text AS $$
-- used by:: admin to deletes proxy configuration records when proxy is no longer needed
-- br:: also connection records must be deleted
-- br:: proxy status must be OFFLINE
-- returns:: 200 Ok
-- example:: select * from plproxy.delete_proxy('central_p1');  
DECLARE
    r_proxy record;
BEGIN	
    -- get proxy record
    r_proxy = plproxy.fetch_proxy_by_name(i_proxy_name);

    if exists (select 1 from plproxy.db_proxy_connection where key_proxy=r_proxy.id_proxy) then
        raise exception '400 Only proxies without connections can be deleted.';
    end if;

    delete from plproxy.db_proxy_connection where key_proxy = r_proxy.id_proxy;

    delete from plproxy.db_proxy_config where key_proxy = r_proxy.id_proxy;

    update plproxy.db_server set key_proxy = null where key_proxy = r_proxy.id_proxy;

    delete from plproxy.db_proxy where id_proxy = r_proxy.id_proxy;
    
    return '200 Ok';
END;
$$ LANGUAGE 'plpgsql' VOLATILE SECURITY DEFINER;


CREATE OR REPLACE FUNCTION plproxy.drop_cluster(
    i_cluster_name text ) 
RETURNS text AS $$
-- used by:: admins to remove out of use clusters
--   removes all data for given cluster
-- exceptions:
--   400 Cluster with given name not found
--   400 Active cluster configurations must NOT be deleted
-- returns:: 200 Ok
-- example:: select * from plproxy.delete_cluster('userdb');  
DECLARE
    r_cluster record;
    r_conf record;
BEGIN	
    r_cluster = plproxy.fetch_cluster_by_name(i_cluster_name);
    if exists(select 1 from plproxy.db_conf where cluster_name = lower(i_cluster_name)) then
        raise exception '400 Cluster with active configuration should not be dropped!';
    end if;

    -- remove cluster configurations
    for r_conf in 
        select id_cluster_conf 
        from plproxy.db_cluster_conf 
        where key_cluster = r_cluster.id_cluster 
    loop
        -- delete mappings
        delete from plproxy.db_conf_partition where key_cluster_conf = r_conf.id_cluster_conf;
        -- delete configuration
        delete from plproxy.db_cluster_conf where id_cluster_conf = r_conf.id_cluster_conf;
    end loop;

    -- delete cluster databases
    delete from plproxy.db_proxy_connection where key_partition in (
        select id_partition from plproxy.db_partition where key_cluster = r_cluster.id_cluster);

    -- delete cluster databases
    delete from plproxy.db_partition where key_cluster = r_cluster.id_cluster;

    -- cluster
    delete from plproxy.db_cluster where id_cluster = r_cluster.id_cluster;

    return '200 Ok';
END;
$$ LANGUAGE 'plpgsql' VOLATILE SECURITY DEFINER;


SET default_tablespace = '';

SET default_with_oids = false;


CREATE TABLE db_cluster (
    id_cluster integer DEFAULT nextval('plproxy.proxy_seq'::text) NOT NULL,
    cluster_name text NOT NULL,
    cluster_descr text NOT NULL,
    cluster_bits smallint NOT NULL,
    cluster_partitions smallint NOT NULL
);




COMMENT ON TABLE db_cluster IS 'Database cluster - set of partitioned database that are visible as one to db clients';



CREATE FUNCTION fetch_cluster_by_name(i_lookup_key text) RETURNS db_cluster
    AS $$
-- find cluster record by name or raises exception if not found
-- select * from plproxy.get_clsuter_by_name('contactdb');  
DECLARE
    r_result record;
BEGIN
    select * from plproxy.db_cluster where cluster_name = lower(i_lookup_key) into r_result;
    if not found then
        raise exception '400 Cluster with given name="%" NOT found', i_lookup_key;
    end if;
    return r_result;
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;




CREATE TABLE db_cluster_conf (
    id_cluster_conf integer DEFAULT nextval('plproxy.proxy_seq'::text) NOT NULL,
    key_cluster integer NOT NULL,
    conf_name text NOT NULL,
    conf_map text DEFAULT '-'::text NOT NULL,
    version integer DEFAULT 1 NOT NULL,
    conf_descr text,
    creation_time timestamp without time zone DEFAULT now()
);




COMMENT ON TABLE db_cluster_conf IS 'Cluster configuration new configuration is created and replicated to all proxies and only then it is made active.';



CREATE FUNCTION fetch_cluster_conf_by_name(i_conf_name text, i_conf_map text) RETURNS db_cluster_conf
    AS $$
-- find cluster configuration record by coniguration name and map
-- used by other functions to vaqlidate input paramteres and 
-- get the configuration record
-- select * from plproxy.get_cluster_conf_by_name('eventdb_V1','-');  
DECLARE
    r_result record;
BEGIN
    select * from plproxy.db_cluster_conf 
    where conf_name = lower(i_conf_name) 
    and conf_map = lower(i_conf_map)
    into r_result;
    if not found then 
        raise exception '400 Configuration with given name="%" and map = "%" NOT found', 
            i_conf_name, i_conf_map;
    end if;
    return r_result;
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;

CREATE OR REPLACE FUNCTION plproxy.fetch_free_partition_name(
  i_cluster_name text ) 
RETURNS text AS $$
-- find first free partition name for given cluster
-- select * from plproxy.get_free_partition_name('eventdb');  
select min(part_name) from (
    select $1 || '_p' || lpad(generate_series::text, 3, '0') as part_name 
    from generate_series(0, (
        select coalesce(count(1),0)
	from plproxy.db_partition 
        where key_cluster = (
	    select id_cluster from plproxy.db_cluster where cluster_name = lower($1))
    )) 
) names 
where part_name not in (select partition_name from plproxy.db_partition);
$$ LANGUAGE 'sql' VOLATILE SECURITY DEFINER;

CREATE TABLE db_partition (
    id_partition integer DEFAULT nextval('plproxy.proxy_seq'::text) NOT NULL,
    key_cluster integer NOT NULL,
    partition_name text NOT NULL,
    connect_string text,
    partition_descr text
);




COMMENT ON TABLE db_partition IS 'Description of database partition that is part of some cluster';



COMMENT ON COLUMN db_partition.partition_name IS 'Name of the partition . Created from cluster nam by adding partition number';



COMMENT ON COLUMN db_partition.partition_descr IS 'If some comment is needed';



CREATE FUNCTION fetch_partition_by_name(i_partition_name text) RETURNS db_partition
    AS $$
-- find partition record by partition name 
-- select * from plproxy.get_partition('eventdb_p1');  
DECLARE
    r_result record;
BEGIN
    select * from plproxy.db_partition where partition_name = lower(i_partition_name) into r_result;
    if not found then 
        raise exception '400 Partition with given name="%" NOT found', i_partition_name;
    end if;
    return r_result;
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;




CREATE TABLE db_proxy (
    id_proxy integer DEFAULT nextval('plproxy.proxy_seq'::text) NOT NULL,
    proxy_name text NOT NULL,
    proxy_descr text
);




COMMENT ON COLUMN db_proxy.id_proxy IS 'Unique number for machines';



COMMENT ON COLUMN db_proxy.proxy_name IS 'Unique name of the proxy for humans and other life forms';



COMMENT ON COLUMN db_proxy.proxy_descr IS 'Place for admins to write commenst';



CREATE FUNCTION fetch_proxy_by_name(i_lookup_key text) RETURNS db_proxy
    AS $$
-- find proxy record by name
-- select * from plproxy.get_proxy('central_p1');  
DECLARE
    r_result record;
BEGIN
    select * from plproxy.db_proxy where proxy_name = lower(i_lookup_key) into r_result;
    if not found then
        raise exception '400 Proxy with given name="%" NOT found', i_lookup_key;
    end if;
    return r_result;
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;




CREATE FUNCTION get_active_confs(i_proxy_name text) RETURNS SETOF ret_get_active_confs
    AS $_$
-- used by:: proxy pool to initialize internal list of served clusters
--   needed when pool is started or when configuration is changed or reloaded
--   result is used for fetching internal maps for each cluster
-- br:: only clusters served by given proxy are returned
-- returns:: set of records
--   cluster_name text - language knows cluster
--   conf_name text - just for reference
-- example::
--   select * from plproxy.get_active_confs('central_p1');  
select cluster_name, conf_name from plproxy.db_conf
where cluster_name in (
  select cluster_name from plproxy.db_cluster where id_cluster in (
    select key_cluster from plproxy.db_partition where id_partition in (
      select key_partition from plproxy.db_proxy_connection where key_proxy in (
        select id_proxy from plproxy.db_proxy where proxy_name = lower($1) ))))
order by cluster_name;
$_$
    LANGUAGE sql SECURITY DEFINER;




CREATE FUNCTION get_cluster_maps(i_cluster_name text) RETURNS SETOF ret_get_cluster_map
    AS $_$
-- used by:: proxy pool to initialize internal maps for given cluster
--   needed when pool is started or when configuration is changed or reloaded
--   result is used for mapping conf_map and hash_nr received from hash function
--   to one partition and then result from get_cluster_partition is used to
--   get actual connect string
-- br:: maps only for active configuration are loaded
-- returns:: set of records
--   conf_map text -- map_code if several maps used otherwise '-'
--   hash_nr int2 - hash number calculated from username or sth else
--   key_partition int4 - partiton that holds this data
-- example::
--   select * from plproxy.get_cluster_maps('eventdb');  
select conf.conf_map, map.hash_nr, map.key_partition
  from plproxy.db_conf_partition map
       inner join plproxy.db_cluster_conf conf 
             on map.key_cluster_conf = conf.id_cluster_conf 
             and conf.conf_name = (
	         select conf_name from plproxy.db_conf where cluster_name = lower($1))
 order by conf.conf_map, map.hash_nr, map.key_partition;
$_$
    LANGUAGE sql SECURITY DEFINER;




CREATE FUNCTION get_cluster_partitions(i_cluster_name text) RETURNS SETOF ret_get_cluster_partitions
    AS $_$
-- used by:: proxy pool to initialize internal maps
--   needed when run on all command is used in language 
--   and when partition number is found out from hash_map
-- br:: only partitons from active configuration are returned
--   db_proxy -- db_proxy_connection -- db_partition -- db_cluster
-- br:: proxy status must be either LOCKED, OFFLINE or ONLINE
-- if proxy is LOCKED ten exception is raised and script is stopped
--   these must be handled before script is run again
-- returns:: set of records
--    id_partition int - partition number 
--    connect_string text - connect string for that partition
-- example::
--   select * from plproxy.get_cluster_partitions('eventdb');
select p.id_partition, connect_string
from plproxy.db_partition p 
where key_cluster in (
    select key_cluster from plproxy.db_cluster_conf 
    where conf_name = (
        select conf_name from plproxy.db_conf 
	where cluster_name = lower($1)))
$_$
    LANGUAGE sql SECURITY DEFINER;

CREATE OR REPLACE FUNCTION plproxy.get_hash(
    i_cluster_name text ,
    i_key_user text )
RETURNS plproxy.ret_get_hash AS $$
-- USED BY: Language to calulate hash values passed to pool
-- calculates map and hash number form username
-- EXAMPLE: select * from plproxy.get_hash('userdb','askoja');
DECLARE
    retval plproxy.ret_get_hash%ROWTYPE;
BEGIN	
    retval.conf_map = '-';
    retval.hash_nr = hashtext(i_key_user) & 255;
    return retval;
END;
$$ LANGUAGE 'plpgsql' VOLATILE SECURITY DEFINER;


CREATE FUNCTION get_proxy_config(i_proxy_name text) RETURNS SETOF ret_get_proxy_config
    AS $_$
-- used by:: pgcpoold to initialize 
--   needed when pool is started or when configuration is changed or reloaded
-- br:: configuration for one proxy is returned
-- returns:: set of records
--   name text
--   value text
-- example::
--   select * from plproxy.get_proxy_config('central');  
    select name, value
    from plproxy.db_proxy prox 
        inner join plproxy.db_proxy_config conf on prox.id_proxy = conf.key_proxy
    where prox.proxy_name = $1;
$_$
    LANGUAGE sql SECURITY DEFINER;




CREATE FUNCTION get_proxy_connections(i_proxy_name text) RETURNS SETOF ret_get_proxy_connections
    AS $_$
-- used by:: proxy pool to initialize internal maps
--   needed when pool is started or when configuration is changed or reloaded
-- br:: connections only for one proxy are returned
-- returns:: set of records
--   id_partition int4
--   conn_count_min int2
--   conn_count_max int2
--   connect_string text
-- example::
--   select * from plproxy.get_proxy_connections('central');  
    select part.id_partition, conn.conn_count_min, conn.conn_count_max, part.connect_string, partition_name
    from plproxy.db_proxy prox 
        inner join plproxy.db_proxy_connection conn on prox.id_proxy = conn.key_proxy
        inner join plproxy.db_partition part on conn.key_partition = part.id_partition
    where prox.proxy_name = $1;
$_$
    LANGUAGE sql SECURITY DEFINER;



CREATE FUNCTION set_active_conf(i_cluster_name text, i_conf_name text) RETURNS text
    AS $$
-- used by:: configuration change script to update active configuration
--    it is called once in each proxy database when configuration is changed
--    or when configuraton must be restored then once more
-- br:: if cluster has no record in active configuration table then create ir
-- returns:: 200 Ok -- when successful
-- example::
--   select * from plproxy.set_activa_conf('userdb', 'userdb_v2');  
BEGIN	
    update plproxy.db_conf 
    set conf_name = lower(i_conf_name) , start_time = now()
    where cluster_name = lower(i_cluster_name);
    if not found then
        insert into plproxy.db_conf ( cluster_name , conf_name , start_time )
        values ( lower(i_cluster_name) , lower(i_conf_name) , now() );
    end if;
    return '200 Ok';
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;




CREATE FUNCTION set_master_active_conf(i_cluster_name text, i_conf_name text) RETURNS text
    AS $$
-- used by:: configuration change script to update configuration status in master database
--   in the end we use same function to change state but using different funtion gives 
--   possigility to do more on master side if logging or someting else is needed
-- br:: check that configuration with given keys exists
-- returns:: 200 Ok -- when successful
-- exceptions:: 400 Configuration (%) not found in cluster (%), Cannot activate
-- example::
--   select * from plproxy.set_master_activate_conf('userdb', 'userdb_v2');  
BEGIN	
    -- are names correct just to be safe
    if not exists (
       select 1 from plproxy.bp_cluster_conf 
       where conf_name = lower(i_conf_name)
       and key_cluster = (
           select id_cluster from plproxy.db_cluster where cluster_name = lower(i_cluster_name)))
    then
        raise exception '400 Configuration (%) not found in cluster (%), Cannot activate', 
            i_conf_name, i_cluster_name;
    end if;
    
    PERFORM plproxy.set_active_conf(i_cluster_name, i_conf_name);
    
    return '200 Ok';
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;




CREATE FUNCTION set_pool_status(i_cluster_name text, i_status text) RETURNS text
    AS $$
-- used by:: configuration change scipt to sends status change command to pool
-- br:: possible status changes
--    OFFLINE to ONLINE - when proxy has been manually put offline
--    ONLINE to LOCKED - when proxy is locked for configuration change
--    LOCKED to ONLINE - when configuration change is finished
--    LOCKED to OFFLINE - when major fuckup hits during configuration change
--    ONLINE to OFFLINE - when proxy must be taken offilen for whatever reasons
--    ONLINE to RELOAD - when configuration has changed but timing is not critical
-- br:: when status is changed from LOCKED to ONLINE then proxy must check
--   if configuration has been changed and if so then reload and recalculate 
-- br:: RELOAD can be used compatible changes are done to the configuration
--   for example partiton is added for which data will arrive in the future
--   or added partition will read data from old partiton over remote procedure call
--   after configuration is reloaded status goes back to ONLINE
--   hash numbers
-- returns:: pool status text
-- example::
--   select * from plproxy.set_pool_status('userdb','LOCKED');  
DECLARE
    v_result text;
BEGIN	
    
    -- TODO -- call aprpriate pool function 
    -- v_result = pool.lock_cluster(i_cluster_name);

    return i_status;
END;
$$
    LANGUAGE plpgsql SECURITY DEFINER;






CREATE FUNCTION view_cluster_proxies(i_cluster_name text) RETURNS SETOF ret_view_cluster_proxies
    AS $_$
-- used by:: configuration change scipt to get list of proxy servers
--   that are affected by configuration change ingiven cluster
--   list is used to connect to all these proxies and lock them
--   for configuration change
-- br:: it must contain all the proxies that service given cluster
--   db_proxy -- db_proxy_connection -- db_partition -- db_cluster
-- br:: proxy status must be either LOCKED, ONLINE
-- if proxy is LOCKED then calling script should not proceed
-- returns:: set of records
--   proxy_name text - used to alter proxy state later in script
--   proxy_status text - current status of the proxy 
--   connect_string text - used to connect to proxy
-- example::
--   select * from plproxy.get_cluster_proxies('userdb');  
  select proxy_name, null::text, null::text
  from plproxy.db_proxy prox 
  where id_proxy in (
    select key_proxy from plproxy.db_proxy_connection
    where key_partition in (
      select key_partition from plproxy.db_partition 
      where key_cluster = (
        select id_cluster from plproxy.db_cluster where cluster_name = lower($1))));
$_$
    LANGUAGE sql SECURITY DEFINER;




CREATE FUNCTION view_conf_partitions(text) RETURNS SETOF ret_view_conf_partitions
    AS $_$
-- good for validating that conf is correct
-- select * from plproxy.get_conf_partitions('eventdb_v1');
select p.partition_name, 
 (select count(1) from plproxy.db_conf_partition 
  where key_partition = p.id_partition 
  and key_cluster_conf = (
    select id_cluster_conf from plproxy.db_cluster_conf where conf_name = LOWER($1))
 ) as hash_count,
 connect_string
from plproxy.db_partition p 
where key_cluster = (
    select key_cluster from plproxy.db_cluster_conf where conf_name = LOWER($1));
$_$
    LANGUAGE sql SECURITY DEFINER;




CREATE FUNCTION view_partiton_changes(i_cluster_name text, i_target_conf_name text) RETURNS SETOF ret_view_partiton_changes
    AS $_$
-- used by:: configuration change script to get list of partitions
--   whose replication status must be checked
-- br:: we get all needed partitons from db_conf_partition where hash_nr
--   im new configuration are in different partition than in old one
--   this implies also that old partition must be replicated to new one
-- br:: currently active configuration is used as source configuration
-- returns:: set of records
--   source_name text -- replication master used for meaningful messages
--   source_connect text -- used to check if replication is ok 
--      and to end it when configuration gets updated
--   target_name text -- replication slave used for meaningful messages
--   target_connect text -- used to check if replication is 
-- example::
--   select * from plproxy.get_cluster_proxies('userdb');  
select src.partition_name as source_name
, src.connect_string as source_connect
, tgt.partition_name as target_name
, tgt.connect_string as target_connect
from 
( select distinct 
  source.key_partition as source_partition
  , target.key_partition as target_partition
  from
  ( select sp.key_partition, sp.hash_nr 
    from plproxy.db_conf_partition sp
    where sp.key_cluster_conf in (
      select id_cluster_conf from plproxy.db_cluster_conf 
      where conf_name = (
          select conf_name from plproxy.db_conf
	  where cluster_name = lower($1)))
  ) source
  left outer join
  ( select tp.key_partition, tp.hash_nr 
    from plproxy.db_conf_partition tp
    where tp.key_cluster_conf in (
      select id_cluster_conf from plproxy.db_cluster_conf where conf_name = lower($2))
  ) target
  on source.hash_nr = target.hash_nr
) chg
left outer join plproxy.db_partition src on chg.source_partition = src.id_partition
left outer join plproxy.db_partition tgt on chg.target_partition = tgt.id_partition
$_$
    LANGUAGE sql SECURITY DEFINER;




CREATE TABLE db_conf (
    cluster_name text NOT NULL,
    conf_name text NOT NULL,
    start_time timestamp with time zone NOT NULL
);




COMMENT ON TABLE db_conf IS 'This table is managed in each instance separately. It is used to store reference to active configuration. ';



COMMENT ON COLUMN db_conf.conf_name IS 'name of currently active configuration';



CREATE TABLE db_conf_partition (
    id_conf_part integer DEFAULT nextval('plproxy.proxy_seq'::text) NOT NULL,
    key_cluster_conf integer NOT NULL,
    hash_nr smallint NOT NULL,
    key_partition integer NOT NULL
);



SET default_with_oids = true;


CREATE TABLE db_proxy_config (
    id_proxy_config integer DEFAULT nextval('plproxy.proxy_seq'::text) NOT NULL,
    key_proxy integer NOT NULL,
    name text NOT NULL,
    value text
);



SET default_with_oids = false;


CREATE TABLE db_proxy_connection (
    id_proxy_connection integer DEFAULT nextval('plproxy.proxy_seq'::text) NOT NULL,
    key_proxy integer NOT NULL,
    key_partition integer NOT NULL,
    conn_count_min smallint NOT NULL,
    conn_count_max smallint NOT NULL
);



SET default_with_oids = true;


CREATE TABLE db_server (
    id_server integer DEFAULT nextval('plproxy.proxy_seq'::text) NOT NULL,
    key_proxy integer,
    server_name text NOT NULL,
    server_descr text,
    connect_string text,
    status text DEFAULT 'OFFLINE'::text NOT NULL,
    status_time timestamp with time zone DEFAULT now() NOT NULL
);




COMMENT ON COLUMN db_server.server_name IS 'Mnemonic server name like dbpooler3';



COMMENT ON COLUMN db_server.server_descr IS 'Just free text description for the server';



COMMENT ON COLUMN db_server.connect_string IS 'Connect string used by scripts to administer proxies';



COMMENT ON COLUMN db_server.status IS 'Proxy status
- ONLINE proxy is servicing clients,
- OFFLINE proxy is offline but ready for use by next configuration change,
- UNUSED proxy is not used, either prepared or not needed anymore';



CREATE SEQUENCE proxy_seq
    START WITH 1
    INCREMENT BY 1
    NO MAXVALUE
    NO MINVALUE
    CACHE 1;




ALTER TABLE ONLY db_cluster_conf
    ADD CONSTRAINT db_cluster_conf_pkey PRIMARY KEY (id_cluster_conf);




ALTER TABLE ONLY db_cluster_conf
    ADD CONSTRAINT db_cluster_conf_ukey UNIQUE (conf_name, conf_map);




ALTER TABLE ONLY db_cluster
    ADD CONSTRAINT db_cluster_pkey PRIMARY KEY (id_cluster);




ALTER TABLE ONLY db_cluster
    ADD CONSTRAINT db_cluster_ukey UNIQUE (cluster_name);




ALTER TABLE ONLY db_conf_partition
    ADD CONSTRAINT db_conf_partition_pkey PRIMARY KEY (id_conf_part);




ALTER TABLE ONLY db_conf_partition
    ADD CONSTRAINT db_conf_partition_ukey UNIQUE (key_cluster_conf, hash_nr);




ALTER TABLE ONLY db_conf
    ADD CONSTRAINT db_conf_ukey UNIQUE (cluster_name);




ALTER TABLE ONLY db_partition
    ADD CONSTRAINT db_partition_pkey PRIMARY KEY (id_partition);




ALTER TABLE ONLY db_partition
    ADD CONSTRAINT db_partition_ukey UNIQUE (partition_name);




ALTER TABLE ONLY db_proxy_config
    ADD CONSTRAINT db_proxy_config_pkey PRIMARY KEY (id_proxy_config);




ALTER TABLE ONLY db_proxy_config
    ADD CONSTRAINT db_proxy_config_ukey UNIQUE (key_proxy, name);




ALTER TABLE ONLY db_proxy_connection
    ADD CONSTRAINT db_proxy_connection_pkey PRIMARY KEY (id_proxy_connection);




ALTER TABLE ONLY db_proxy_connection
    ADD CONSTRAINT db_proxy_connection_ukey UNIQUE (key_proxy, key_partition);




ALTER TABLE ONLY db_proxy
    ADD CONSTRAINT db_proxy_pkey PRIMARY KEY (id_proxy);




ALTER TABLE ONLY db_proxy
    ADD CONSTRAINT db_proxy_ukey UNIQUE (proxy_name);




ALTER TABLE ONLY db_server
    ADD CONSTRAINT db_server_pkey PRIMARY KEY (id_server);




ALTER TABLE ONLY db_server
    ADD CONSTRAINT db_server_ukey UNIQUE (key_proxy, server_name);




ALTER TABLE ONLY db_cluster_conf
    ADD CONSTRAINT db_cluster_conf_key_cluster_fkey FOREIGN KEY (key_cluster) REFERENCES db_cluster(id_cluster);



ALTER TABLE ONLY db_conf_partition
    ADD CONSTRAINT db_conf_partition_key_cluster_conf_fkey FOREIGN KEY (key_cluster_conf) REFERENCES db_cluster_conf(id_cluster_conf);



ALTER TABLE ONLY db_conf_partition
    ADD CONSTRAINT db_conf_partition_key_partition_fkey FOREIGN KEY (key_partition) REFERENCES db_partition(id_partition);



ALTER TABLE ONLY db_partition
    ADD CONSTRAINT db_partition_key_cluster_fkey FOREIGN KEY (key_cluster) REFERENCES db_cluster(id_cluster);



ALTER TABLE ONLY db_proxy_connection
    ADD CONSTRAINT db_proxy_connection_key_partition_fkey FOREIGN KEY (key_partition) REFERENCES db_partition(id_partition);



ALTER TABLE ONLY db_proxy_connection
    ADD CONSTRAINT db_proxy_connection_key_proxy_fkey FOREIGN KEY (key_proxy) REFERENCES db_proxy(id_proxy);





